#=======================================================================
# this is where everything that is site-specific gets defined
# $Id: AbsenceConfig.pm 91 2009-07-14 15:58:06Z urban $
#=======================================================================

package AbsenceConfig;

use FileHandle;

my $VERSION	= '1.10';
my $DEBUG		= 0;

#-------------------------------------------------------------
# INSTANCES
# (are not currently supported, tested, or working)
#
# If you want multiple instances of absence, you MUST add
# an instance-name and corresponding path to the config-file
# below.  Do not include the config-file in the path.  The
# config-file is ALWAYS called "absence.conf".
#
# Additionally you must edit your top-html file (the one which
# defines the frames) and add "?instance=<your-instance>" to
# the end of both frame target-URLs.  For example:
#
#	SRC="/cgi-bin/absence-control.pl?instance=foo"
#	SRC="/cgi-bin/absence.pl?instance=foo"
#
# for an instance you call "foo"
#-------------------------------------------------------------
my %INSTANCES = (
	# example
	#2	=> '/home/httpd/urban/www.spielwiese.de/html/absence2',
);

#-------------------------------------------------------------
# DATA and CGI directories
# ACHTUNG.  You will probably want to set these
#-------------------------------------------------------------

# the following variables are declared for convenience

#-------------------------------------------------------------
# this is the relative path to the data-dir.
#-------------------------------------------------------------
my $data_dir_rel	= '/absence';					# as used in a URL

#-------------------------------------------------------------
# This is the absolute path to the data directory, for the
# use of a perl script looking for data-files.
#
# users of windoze/IIS will probably need to shove some
# absolute path into "$data_dir_abs". For example:
#	$data_dir_abs = "c:/inetpub/wwwroot$data_dir_rel";
#-------------------------------------------------------------
my $data_dir_abs	= "$ENV{DOCUMENT_ROOT}$data_dir_rel";	# as used by Perl

#-------------------------------------------------------------
# the relative path to the cgi-bin, i.e., what is needed to
# create a URL for a cgi-script.
#-------------------------------------------------------------
my $cgi_dir_rel	= "/cgi-bin";
my $cgi_dir_abs	= "$ENV{DOCUMENT_ROOT}/..$cgi_dir_rel";

my $js_dir_abs	= "$data_dir_abs/JavaScript";
my $js_dir_rel	= "$data_dir_rel/JavaScript";

my %VARS = (
	data_dir_rel	=> $data_dir_rel,
	data_dir_abs	=> $data_dir_abs,
	cgi_dir_rel		=> $cgi_dir_rel,
	cgi_dir_abs		=> $cgi_dir_abs,
	js_dir_abs		=> $js_dir_abs,
	js_dir_rel		=> $js_dir_rel,

	#-------------------------------------------------------------
	# where are the images kept.  Probably OK as is.
	#-------------------------------------------------------------
	image_dir_rel	=> "$data_dir_rel/img",
	image_dir_abs	=> "$data_dir_abs/img",

	#-------------------------------------------------------------
	# access to database
	#-------------------------------------------------------------
	database_name	=> 'absence',
	database_host	=> 'localhost',
	database_port	=> undef,
	database_user	=> 'absence',
	database_pass	=> 'absence',

	#-------------------------------------------------------------
	# where is log file.  Probably OK as is.
	#-------------------------------------------------------------
	log_file		=> "$data_dir_abs/log",

	#-------------------------------------------------------------
	# where to put backup copies of things
	#-------------------------------------------------------------
	backup_dir		=> "$data_dir_abs/backup",

	#-------------------------------------------------------------
	# marking current month and day
	#-------------------------------------------------------------
	mark_curr_month	=> 'yes',
	mark_curr_day	=> 'yes',
	curr_day_style	=> 'box',	# can be "fill" or "box"
	curr_day_begin_row	=> 2,
	auto_refresh	=> 'yes',

	#-------------------------------------------------------------
	# paths to various CGI-scripts
	#-------------------------------------------------------------
	main_script		=> "$cgi_dir_rel/absence.pl",
	manage_script	=> "$cgi_dir_rel/absence-manage.pl",
	control_script	=> "$cgi_dir_rel/absence-control.pl",
	js_acl_script	=> "$js_dir_rel/acl.js",
	nph_lo_script	=> "$cgi_dir_rel/nph-absence-logout.pl",

	#-------------------------------------------------------------
	# this is the URL to the html page containing the frameset.
	# If your main page is called "index.html" and you are using
	# Apache, this can simply be set to "$data_dir_rel".
	#-------------------------------------------------------------
	top_page		=> "$data_dir_rel/index.html",

	#-------------------------------------------------------------
	# legend-style.  can either be "constant:<num>" or "fit"
	# if set to constant, all legend-labels will be <num> pixels wide
	# if "fit" labels will be fit to text
	#-------------------------------------------------------------

	#legend_style	=> 'constant:60',
	legend_style	=> 'fit',

	#-------------------------------------------------------------
	# select the image-type, png, jpg, or gif.  'gif' is only
	# supported on ancient versions of gd.  'png' is probably
	# the best choice if you have it.
	#-------------------------------------------------------------
	image_type		=> 'png',	# one of 'png', 'jpg', or 'gif'

	#-------------------------------------------------------------
	# map_type determines whether client or server-side image-maps
	# are used.  The value 'client' causes a map to be generated
	# and integrated into the main web-page for each month.  This
	# can cause thousands of map-areas to be defined, depending
	# on how many months are displayed and how many people are
	# in a group.
	# The value 'server' causes the "ismap" option to be added to
	# the month-image, which additionally is a hyperlink. This
	# results in the coordinates of the click being passed to
	# the server for processing.
	# The advantage of the client-side option is tooltips are used
	# to display the contents of a reservation.
	# CAUTION: "server" maps *cannot* be used when multiple
	# reservations per day are configured.
	# In fact, "server" maps are not well tested, so I recommend
	# sticking to "client" maps. Bandwidth is free, right?
	#
	# legal values: "client" or "server"
	#-------------------------------------------------------------
	map_type		=> 'client',

	#-------------------------------------------------------------
	# draw separation-lines between day-columns, "yes" or "no"
	#-------------------------------------------------------------
	vertical_grid   => 'yes',

	#-------------------------------------------------------------
	# multiple reservations per day
	#-------------------------------------------------------------
	multi_res		=> 'yes',
	max_multi		=> 3,
	variable_height	=> 'yes',
	min_height		=> 1,

	#-------------------------------------------------------------
	# holiday configuration
	# ---------------------
	# "holiday_label" is the label used in the legend to indicate
	# holidays (bank-holidays in brittain).
	# "holiday_mark" determines how a holiday is indicated on
	# the month-image.  "small" means the square underneath
	# the day-number is colored using the color indicated by
	# "mi_bankholiday".  "large" means the whole column is
	# colored, starting at the day-number and going to the bottom
	# of the month-image.
	#
	# "holiday_scheme" can have two values, "basic", and "advanced"
	# basic means the old (v1.8) holiday style is used.  all holidays
	# in the DB will be displayed in the header, i.e., the country
	# and region are ignored.
	# "advanced" means the holidays must be associated with a country
	# and/or a region. "header_holiday_country" and "header_holiday_region"
	# determine which holidays are displayed in the header
	#-------------------------------------------------------------
	holiday_label	=> 'Bank Holiday',
	holiday_scheme	=> 'advanced',

	# the following only applies when "holiday_scheme" is "basic"
	holiday_mark	=> 'large',		# can be 'large' or 'small'

	# the following are only relevant if holiday_scheme is 'advanced'
	show_holidays_in_header	=> 'yes',
	header_holiday_country	=> 'de',	# ISO country-code
	header_holiday_region	=> 'Bayern',

	#-------------------------------------------------------------
	# SKIP stuff
	#
	# if "skip_we_hol" is set to "yes", an absence including
	# weekend days and holidays will be displayed as a sequence
	# of individual blocks that only cover the non-holiday
	# weekdays.  That is, weekends and holidays are skipped.
	# "skip_mark_skipped" causes a frame to be drawn in the
	# day-boxes of all skipped days.  The color used is the same
	# as for the blocks.
	# To use skipping, you must also check the corresponding box
	# for a particular reservation-type using the managment web page.
	#-------------------------------------------------------------
	skip_we_hol		=> 'yes',
	skip_mark_skipped	=> 'yes',

	#-------------------------------------------------------------
	# this determines whether a user can be a member of multiple
	# groups, or only a single one.  values: "single", "multiple"
	#-------------------------------------------------------------
	group_policy	=> 'multiple',	# 'single' or 'multiple'

	#-------------------------------------------------------------
	# various authentication/authorization parameters
	#
	# "partial_auth_hack" deserves perhaps some explanation.
	# It is a mechanism for restricting administrative access (in
	# a trivial and easily defeated manner) to certain admin-operations.
	# Currently they include all reservation-type operations.
	# partial_auth_hack exists simply to deter non-hostile
	# reservation-type changes...
	# if you want any real protection from hostile users, you must
	# turn on real authentication/authorization
	#-------------------------------------------------------------
	authentication	=> 'yes',		# 'yes' or 'no'
	auth_type		=> 'simple',	# currently can be "http" or "simple"
	secret			=> 'qwerty',	# a random string
	credential_src	=> 'absence',	# 'absence' or 'htaccess'
	session_timeout	=> '30m',		# use suffixs m,h,d,w for units, 0 for none
	manage_password	=> 'yes',		# 'yes' or 'no'
	pw_hash_format	=> 'md5',		# "cleartext" or "md5"
	partial_auth_hack			=> 0,
	partial_auth_hack_pw		=> 'secret',

	# the following are for http authentication
	htaccess_path	=> "$data_dir_abs/password.file",
	auth_realm		=> 'Absence',	# for 'http' authentication

	#-------------------------------------------------------------
	# access control defaults
	# The format for "rules" is "<permission>:<object>"
	# <permission> can be "w", or "r" for "write", or "read" respectively.
	# <object> can only be "self" for people,
	# or "self", "all" groups.
	# for people, "self" refers to the object associated with the
	# authenticated user. Note that "r:self" for either people or
	# group defaults makes no sense.
	#
	# for groups, "self" refers to any group which has the person
	# as a member who corresponds to the authenticate user.
	# To give all users all rights, set group defaults to "a:all";
	# The group rule "r:self" is implicit and cannot be disabled.
	# Some examples:
	#
	# 	# all users should be able to modify any people in groups
	# 	# to which they belong:
	# 	pacl_default	=> undef,
	# 	gacl_default	=> 'w:self',
	#
	#	# users should only be able to modify their own objects:
	# 	pacl_default	=> 'w:self',
	# 	gacl_default	=> undef,
	#
	#	# users should only be able to read groups to which they belong
	# 	pacl_default	=> undef,
	# 	gacl_default	=> undef,
	#-------------------------------------------------------------
	pacl_default	=> undef,
	gacl_default	=> 'w:self',

	#-------------------------------------------------------------
	# if the managed objects represent people, there must be
	# some association between the managed objects and authenticated
	# users.
	#-------------------------------------------------------------
	objects_are_people	=> 'yes',

	#-------------------------------------------------------------
	# set the following if you want alternating background colors
	# for people rows.  The first color used is "mi_altbg",
	# the second is "mi_background"
	#-------------------------------------------------------------
	alternate_colors	=> 'no',

	#-------------------------------------------------------------
	# %% COLORS %%
	#-------------------------------------------------------------

	#-------------------------------------------------------------
	# MONTH-IMAGE colors
	#
	# color-names that begin with "mi_" belong to "month-images".
	# colors may be specified in four formats:
	#
	#  1: '#rrggbb'
	#  2: '0xrrggbb'
	#  3: 'rr gg bb'
	#  4: '0xrr 0xgg 0xbb'
	#
	# where 'rr', 'gg', and 'bb' are the red, green, and blue
	# components, respectively.  In forms 1 and 2, the components
	# *must* be hexadecimal numbers.  In forms 3 and 4, they *may*
	# be decimal or hexadecimal.  If hex, each component must
	# be preceeded by "0x"
	#-------------------------------------------------------------
	mi_background	=> '#b0a0b0',	# normal background
	mi_altbg		=> '#c0b0c0',	# 2nd color for alternating bg
	mi_weekend		=> '#908090',	# weekend background
	mi_bankholiday	=> '#a020f0',	# guess?
	mi_text_dark	=> '#000000',	# color for text
	mi_text_light	=> '#ffffff',	# not currently used
	#mi_grid		=> '#808080',	# grid color (absence blocks jump out)
	mi_grid			=> '#404040',	# grid color
	mi_cwbox		=> '#aa80aa',	# the calendar-week boxes
	mi_padding		=> '#605060',	# to pad months < 31 days
	mi_res_border	=> '#000000',	# reservation border when multi_res enabled

	#-------------------------------------------------------------
	# colors used to highlight current month and/or day
	#-------------------------------------------------------------
	mi_curr_day		=> '#ffffff',	# frame around current day
	mi_cm_bg		=> '#c0b0c0',
	mi_cm_altbg		=> '#d0c0d0',
	mi_cm_weekend	=> '#a090a0',
	mi_cm_grid		=> '#808080',	# grid color
	mi_cm_monthlab	=> '#2020ff',
	mi_cm_mi_frame	=> '#ffffff',	# the frame around the image
	mi_cm_cwbox		=> '#ba90ba',

	#-------------------------------------------------------------
	# web-page colors
	# These are specified according to HTML rules: either
	# "#rrggbb" (each component  a 2-digit hex values, as above),
	# or a color-name can be specified.
	#-------------------------------------------------------------
	wp_main			=> '#acacac',		# bg col for display frame
	wp_control		=> '#ff9081',		# bg col for control frame
	wp_nopriv		=> 'teal',			# bg col for auth errors
	wp_manage		=> '#aaa0b8',		# bg col for management
	wp_aflabel		=> 'papayawhip',	# bg col for labels in absence form
	wp_afbg			=> '#acacac',		# bg col for absence form
);
# %%END_MARKER%%

if (exists($ENV{INSTANCE_NAME})) {
	# an instance has been specified
	my $inst = $ENV{INSTANCE_NAME};
	print "** instance specified: [$inst] **<BR>\n";
	if (exists($INSTANCES{$inst})) {
		my $cf = "$INSTANCES{$inst}/absence.conf";
		if (! -e $cf) {
			die "no configfile found for instance [$inst] ($cf)\n";
		}
		parse($cf);
	} else {
		die "instance [$inst] not defined\n";
	}
}

sub fetch
{
	my $val = shift;

	if (exists($VARS{$val})) {
		return undef unless(defined($VARS{$val}));
		if ($VARS{$val} =~ /^(yes|true)$/i) { return 1; }
		if ($VARS{$val} =~ /^(no|false)$/i) { return 0; }
		return $VARS{$val};
	}

	return undef;
}

sub parse
{
	my ($file) = @_;

	my $fh = FileHandle->new($file);
	defined($fh) || die "open [$file] for reading";
	while(<$fh>) {
		$DEBUG && print "CONF> $_";
		if (/^\s*$/) { next; }		# skip blank lines
		if (/^\s*#$/) { next; }		# skip comment lines
		if (/^\s*(\S+)\s*=\s*(\S+)/) {
			$varname = $1;
			$value = $2;
			$value =~ s/\${([^}]+)}/expand($1)/ge;
			$DEBUG && print "  setting [$varname] = [$value]\n";
			$VARS{$varname} = $value;
		}
	}
	$fh->close;
}

sub expand
{
	my $vname = shift;

	if (exists($VARS{$vname})) {
		return $VARS{$vname};
	} else {
		die "error in config-file\n";
	}
}

sub dumpVars
{
	foreach my $var (sort keys(%VARS)) {
		printf("%-18s %s\n", $var, $VARS{$var});
	}
}

1;

__END__

=head1 NAME

AbsenceConfig.pm - handles configuration for web-absence

=head1 SYNOPSIS

  use AbsenceConfig;

  AbsenceConfig::fetch('config-param');

=head1 DESCRIPTION

C<AbsenceConfig.pm> handles the configuration information that
belongs to a particular instance.  The module itself defines
'default' values for all configuration parameters.  If no
instance is explicitly configured, these 'default' values
are used.  If an instance is configured, the instance-specific
parameter has precedence, then the default value.

Currently understood parameters are:

=over 4

=item data_dir_rel

The relative directory path where data-files reside
(C<absence.db>, C<month_mod_times.db>, C<holiday.db>, C<absence-types.db>).
That is, the path as used in HTML source.

=item data_dir_abs

The absolute directory path where data-files reside.
That is, the path as used in CGI scripts.

=item cgi_dir_rel

The relative directory path where CGI-scripts and modules reside.

=item image_dir_rel

The relative directory path where the month-images and legend reside.

=item image_dir_abs

The absolute directory path where the month-images and legend reside.

=item database_file

The absolute path to the database file.

=item modtime_db_file

The absolute path to the modification time file.  This is used to hold
the modification times for month-images.  The modification time allows
URLs for the month-images to be constructed such that the browser
automatically reloads a month-image when it has been modified.

=item holiday_db_file

The absolute path to the holiday database file.

=item log_file

The absolute path to the log file.

=item type_file

The absolute path to the absence-type file.  This is where the various
kinds of absences can be defined along with a color.

=item main_script

The relative path to C<absence.pl>.

=item manage_script

The relative path to C<absence-manage.pl>.

=item control_script

The relative path to C<absence-control.pl>.

=item top_page

The relative path to the startinge-page.  This is usually C<index.html>
on UNIX systems.  If you are using Apache, you can simply set this
to C<$data_dir_rel>.

=item legend_style

The method used to arrive at the width for the legend boxes.  There are
two possibilities, C<fit> and C<contant:N> (where N is the width in pixels).
C<fit> uses the text to determine the width of the box. C<contant:N>
simply causes all boxes to have a width of N pixels.

=item backup_copies

Before the database is modified, a backup copy named C<absence.db.savN>
is made, where N is incremented each time.  This determines how many copies
are kept.

=item image_type

The format of the images used (legend and month-images).  The choices are
C<png>, C<jpg>, and C<gif>.  The version of GD that you installed must
support the format you choose here.

=item map_type

This determines what method to use for manipulating absences.  There
are two choices, C<server> (server-side image maps) and C<client>
(client-side image maps).  Client-side image maps allow the display
of so-called "tool-tips", but the the HTML generated is very large.
Server-side image maps result in much smaller HTML, but you must
click on absences and holidays to get more information about them.
Server-side image maps won't work on windoze/IIS.

=item cookie_domain

The domain information to be used when setting cookies.

=item holiday_label

The label that should be used to mark (bank) holidays, as opposed
to vacation (as understood in america :).

=item group_policy

This determines whether a user can be a member of multiple groups,
or only of a single group.  Values are C<single> and C<multiple>.

=back

=head1 INSTANCES

absence now supports multiple "instances".  That is, completely
independent databases which all use the same Perl scripts and
modules.

=cut
