package AbsenceImage;

# $Id: AbsenceImage.pm 91 2009-07-14 15:58:06Z urban $

use FileHandle;
use GD;
use Carp;
use Data::Dumper;
use Encode qw(from_to decode encode decode_utf8);

use AbsenceConfig;
use AbsenceDB;
use AbsenceDate;
#use AbsenceHolidays;
#use AbsenceTypes;
use AbsenceLog;

use strict;

# colors

my %COLOR_VALUES = (
	bg			=> getColorTriplet('mi_background'),
	altbg		=> getColorTriplet('mi_altbg'),
	we			=> getColorTriplet('mi_weekend'),
	bholiday	=> getColorTriplet('mi_bankholiday'),
	text_dark	=> getColorTriplet('mi_text_dark'),
	text_light	=> getColorTriplet('mi_text_light'),
	grid		=> getColorTriplet('mi_grid'),
	cw			=> getColorTriplet('mi_cwbox'),
	padding		=> getColorTriplet('mi_padding'),
	res_border	=> getColorTriplet('mi_res_border'),
	white		=> [ 0xff, 0xff, 0xff ],
	black		=> [ 0x00, 0x00, 0x00 ],
);

my $FULL_HEIGHT		= 12;
my $FULL_SLOTS		= 4;
my $SLOT_PIXELS		= int($FULL_HEIGHT / $FULL_SLOTS);
my $COL_HEIGHT		= 20;
my $MULTI_RES		= AbsenceConfig::fetch('multi_res');
my $MAX_MULTI		= AbsenceConfig::fetch('max_multi');
my $MIN_HEIGHT		= AbsenceConfig::fetch('min_height');
my $VARIABLE_HEIGHT	= AbsenceConfig::fetch('variable_height');
my $NUM_SLOTS		= $MAX_MULTI * $FULL_SLOTS;
my %COLOR			= ();
my $ROW_HEIGHT		= $MULTI_RES
	? $MAX_MULTI * $FULL_HEIGHT + 1
	: $COL_HEIGHT;
my $HEADER_ROW_HEIGHT	= 20;
my $COL_WIDTH		= 20;
my $HGRID_WIDTH		= $MULTI_RES ? 2 : 1;
my $ROW_LABEL_WIDTH	= 100;
my $HEADER_LINES	= 2;
my $CGI_DIR			= AbsenceConfig::fetch('cgi_dir_rel');
my $MARK_CM			= AbsenceConfig::fetch('mark_curr_month');
my $MARK_CD			= AbsenceConfig::fetch('mark_curr_day');
my $HOLIDAY_MARK	= AbsenceConfig::fetch('holiday_mark');
my $HOLIDAY_SCHEME	= AbsenceConfig::fetch('holiday_scheme');
my $IMAGE_DIR_ABS	= AbsenceConfig::fetch('image_dir_abs');
my $IMAGE_DIR_REL	= AbsenceConfig::fetch('image_dir_rel');
my $SKIP_WE_HOL		= AbsenceConfig::fetch('skip_we_hol');
my $SKIP_MARK_SK	= AbsenceConfig::fetch('skip_mark_skipped');
my $NON_BLOCK_PRIO	= 100;

my $HOL_SHOW_HEAD;
my $HOL_HEAD_REGION;
my $HOL_HEAD_COUNTRY;

if ($HOLIDAY_SCHEME eq 'advanced') {
	$HOL_SHOW_HEAD		= AbsenceConfig::fetch('show_holidays_in_header');
	$HOL_HEAD_REGION	= AbsenceConfig::fetch('header_holiday_region');
	$HOL_HEAD_COUNTRY	= AbsenceConfig::fetch('header_holiday_country');
}

my %HEIGHT_MAP = (
	'full'		=> { slots => $FULL_SLOTS,		pixels => $FULL_HEIGHT	},
	'half'		=> { slots => $FULL_SLOTS/2,	pixels => $FULL_HEIGHT/2 },
	'quarter'	=> { slots => $FULL_SLOTS/4,	pixels => $FULL_HEIGHT/4 },
);

#------------------------------------------------------------
# only retrieve these if I need them
#------------------------------------------------------------

if ($MARK_CM || $MARK_CD) {
	$COLOR_VALUES{cm_bg}		= getColorTriplet('mi_cm_bg');
	$COLOR_VALUES{cm_altbg}		= getColorTriplet('mi_cm_altbg');
	$COLOR_VALUES{cm_we}		= getColorTriplet('mi_cm_weekend');
	$COLOR_VALUES{cm_grid}		= getColorTriplet('mi_cm_grid');
	$COLOR_VALUES{cm_monthlab}	= getColorTriplet('mi_cm_monthlab');
	$COLOR_VALUES{cm_mi_frame}	= getColorTriplet('mi_cm_mi_frame');
	$COLOR_VALUES{cm_cw}		= getColorTriplet('mi_cm_cwbox');
	$COLOR_VALUES{currday}		= getColorTriplet('mi_curr_day'),
}

my $DEBUG			= 0;
my $HDEBUG			= 0;
my $VERSION			= '1.9';

my $I_HEIGHT;

sub getColorTriplet
{
	my $name = shift;

	my $str = AbsenceConfig::fetch($name);

	my @t;
	if (@t = ($str =~ /^(?:#|0x)?([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$/))
	{
		return [map(hex, @t)];
	}

	if ($str =~ /^((0x[0-9a-fA-F]+|\d+)\s+){2}(0x[0-9a-fA-F]+|\d+)$/) {
		my $ref;
		foreach my $val (split(' ', $str)) {
			($val =~ /0x/) ? push(@{$ref}, hex($val)) : push(@{$ref}, $val);
		}

		return $ref;
	}

	die "getColorTriplet: failed to match [$str] from [$name]";
}

sub create
{
	my ($month, $year, $group_id, $image_outfile, $map_outfile, $mapname) = @_;

	my $debug = 0;

	$debug && abslog("\n%%% create() month=$month, year=$year");

	if (!$month || !$year) {
		confess "missing people, month, or year";
	}

	my $map_fh = FileHandle->new($map_outfile, 'w');
	defined($map_fh) || die "failed to open [$map_outfile] to write: $!";
	print $map_fh qq[<MAP NAME="$mapname">\n];

	my $numdays = AbsenceDate::daysInMonth($month, $year);

	#----------------------------------------------------------
	# get people
	#----------------------------------------------------------
	my @people = AbsenceDB::getPeople($group_id);

	#----------------------------------------------------------
	# calculate height of person-rows, if necessary
	# $pr_info is a reference to an array of object-ids, i.e.:
	#	$pr_info = [1] = { stuff for OID 1 },
	#	$pr_info = [5] = { stuff for OID 5 },
	#----------------------------------------------------------
	my $pr_info = layoutObjectRows($month, $year, \@people);

	#----------------------------------------------------------
	# calculate width of image
	#----------------------------------------------------------
	my $use_width = $ROW_LABEL_WIDTH + $numdays * $COL_WIDTH;
	my $i_width = $ROW_LABEL_WIDTH + 31 * $COL_WIDTH + 1;

	#----------------------------------------------------------
	# calculate height of image
	#----------------------------------------------------------
	my $pr_height = 0;
	foreach my $pid (@people) {
		#abslog(" HEIGHT[$pid] = $pr_info->[$pid]->{height}");
		$pr_height += $pr_info->[$pid]->{height};
	}
	#$I_HEIGHT = @people * $ROW_HEIGHT + $HEADER_LINES * $HEADER_ROW_HEIGHT + 1;
	$I_HEIGHT = (@people + 1) * $HGRID_WIDTH
		+ $pr_height
		+ $HEADER_LINES * $HEADER_ROW_HEIGHT;

	#----------------------------------------------------------
	# save dimensions in DB
	#----------------------------------------------------------
	AbsenceDB::updateImageDimensions(
		$group_id, $month, $year,
		$i_width, $I_HEIGHT,
	);

	#----------------------------------------------------------
	# create new image
	#----------------------------------------------------------
	my $im = new GD::Image($i_width, $I_HEIGHT, 1);

	#----------------------------------------------------------
	# allocate... you guessed it, colors
	#----------------------------------------------------------
	allocateColors($im);
	#$im->transparent($COLOR{bg});

	#----------------------------------------------------------
	# do background
	# I don't need to do this.  Because the background color
	# is allocated first (in "allocateColors()"), GD automagically
	# uses it for the background
	#----------------------------------------------------------
	$im->filledRectangle(0, 0, $i_width, $I_HEIGHT, $COLOR{bg});

	my ($month_label_color, $grid_color, $alt_bg);

	#----------------------------------------------------------
	# set colors depending on whether this is the current month
	#----------------------------------------------------------
	if ($MARK_CM && currentMonth($month, $year)) {
		$im->filledRectangle(0, 0, $i_width, $I_HEIGHT, $COLOR{cm_bg});
		$grid_color			= $COLOR{cm_grid};
		$month_label_color	= $COLOR{cm_monthlab};
		$alt_bg				= $COLOR{cm_altbg};
	} else {
		$grid_color			= $COLOR{grid};
		$month_label_color	= $COLOR{text_dark};
		$alt_bg				= $COLOR{altbg};
	}

	my $cur_y = $HEADER_LINES * $HEADER_ROW_HEIGHT + $HGRID_WIDTH;

	#----------------------------------------------------------
	# optionally add alternating bars in the people rows...
	#----------------------------------------------------------
	if (AbsenceConfig::fetch('alternate_colors')) {
		my $row = 0;
		my $y = $cur_y;
		foreach my $pid (@people) {
			my $row_height = $pr_info->[$pid]->{height};
			if (($row % 2) != 1) {
				$im->filledRectangle(
					0, $y,								# (x1, y1)
					$use_width, $y + $row_height - 1,	# (x2, y2)
					$alt_bg,
				);
			}
			$row++;
			$y += $row_height + $HGRID_WIDTH;
		}
	}

	#----------------------------------------------------------
	# if numdays != 31, add padding color
	#----------------------------------------------------------
	if ($numdays != 31) {
		my $x1 = $ROW_LABEL_WIDTH + $numdays * $COL_WIDTH + 1;
		#print "<BR>X1 = [$x1], X2 = [$i_width]<BR>\n";
		$im->filledRectangle(
			$x1, 0,
			$i_width, $I_HEIGHT,
			$COLOR{padding});
	}

	my $sav;
	$cur_y = $sav = $HEADER_LINES * $HEADER_ROW_HEIGHT;

	#----------------------------------------------------------
	# mark weekends
	#----------------------------------------------------------
	markWeekends($im, $month, $year);
	
	#----------------------------------------------------------
	# mark holidays
	#----------------------------------------------------------
	markHeaderHolidays($im, $month, $year, $map_fh);

	#----------------------------------------------------------
	# add month and year
	#----------------------------------------------------------
	$im->string(gdGiantFont, 10, 5, AbsenceDate::monthName($month),
		$month_label_color);
	$im->string(gdGiantFont, 10, 20, $year, $month_label_color);

	#----------------------------------------------------------
	# add CW bars
	#----------------------------------------------------------
	drawCwBars($im, $month, $year);

	#----------------------------------------------------------
	# mark current day, if configured to do so
	#----------------------------------------------------------
	if ($MARK_CD && currentMonth($month, $year)) {
		markCurrentDay($im, $month, $year);
	}

	#----------------------------------------------------------
	# draw vertical grid lines
	# horizontal lines must wait, except for first one
	#----------------------------------------------------------

	my $y2 = $cur_y + $HGRID_WIDTH - 1;
	$debug && abslog("first grid line. (0,$cur_y) ($use_width,$y2)");

	$im->filledRectangle(
		0, $cur_y,							# (x1, y1)
		$use_width, $cur_y + $HGRID_WIDTH - 1,	# (x2, y2)
		$grid_color,
	);
	
	my $cur_x = $ROW_LABEL_WIDTH;
	$im->line($cur_x, $sav, $cur_x, $I_HEIGHT, $grid_color);
	$cur_x += $COL_WIDTH;

	if (AbsenceConfig::fetch('vertical_grid')) {
		for(my $i = 1; $i <= $numdays; $i++) {
			#print "line: $cur_x, $sav, $cur_x, $I_HEIGHT\n";
			$im->line($cur_x, $sav, $cur_x, $I_HEIGHT, $grid_color);
			$cur_x += $COL_WIDTH;
		}
	}

	$cur_y += $HGRID_WIDTH;

	#----------------------------------------------------------
	# add monthdays
	#----------------------------------------------------------
	$cur_x = $ROW_LABEL_WIDTH;
	for(my $i = 1; $i <= $numdays; $i++) {
		$im->string(gdSmallFont, $cur_x + 5, $HEADER_ROW_HEIGHT + 2, $i, $COLOR{text_dark});
		$cur_x += $COL_WIDTH;
	}

	#----------------------------------------------------------
	# build person-rows
	#----------------------------------------------------------

	my $x_offset = 5;
	my $fit_chars = int(($ROW_LABEL_WIDTH - $x_offset) / 6.0);

	my $row = 0;
	foreach my $pid (@people) {
		my $row_height = $pr_info->[$pid]->{height};
		#$debug && abslog("row-height = [$row_height]");
		my $pref = AbsenceDB::getPerson($pid);

		#----------------------------
		# add holidays, if configured
		#----------------------------
		if ($HOLIDAY_SCHEME eq 'advanced') {
			my @params;
			my $found = 0;
			if (defined($pref->{country_id})) {
				push(@params, country => $pref->{country_id});
				$found = 1;
			}
			if (defined($pref->{region_id})) {
				push(@params, region => $pref->{region_id});
				$found = 1;
			}
			
			if ($found) {
				my @holidays = AbsenceDB::getHolidayList($month, $year, @params);
				drawHolidays($im, \@holidays, $cur_y, $cur_y + $row_height - 1, 0, $numdays);
			}
		}

		#----------------------------
		# add people
		#----------------------------
		my $name = $pref->{name};
		from_to($name, 'utf8', 'iso-8859-1') || die "from_to failed: $!";
		if (length($pref->{name}) > 15) {
			$name = substr($name, 0, $fit_chars);
		}
		my $y_offset = $cur_y + int($row_height / 2) - 7;
		#my $name_latin = decode_utf8($name);
		$im->string(gdSmallFont, $x_offset, $y_offset, $name, $COLOR{text_dark});

		#----------------------------
		# add absences
		#----------------------------
		$DEBUG && print "adding absences..\n";
		drawObjectRow($im, $pr_info->[$pid], $pid,
			$group_id, $cur_y, $month, $year, $map_fh);
		$cur_y += $row_height;

		#----------------------------
		# add horizontal grid line
		#----------------------------
		$y2 = $cur_y + $HGRID_WIDTH - 1;
		#$debug && abslog("hgrid-line: (0,$cur_y) ($use_width,$y2)");
		$im->filledRectangle(
			0, $cur_y,							# (x1, y1)
			$use_width, $cur_y + $HGRID_WIDTH - 1,	# (x2, y2)
			$grid_color,
		);
		$cur_y += $HGRID_WIDTH;
	}

	#----------------------------------------------------------
	# mark current month, if configured
	#----------------------------------------------------------
	if ($MARK_CM && currentMonth($month, $year)) {
		$im->rectangle(0, 0, $use_width, $I_HEIGHT - 1, $COLOR{cm_mi_frame});
		$im->rectangle(1, 1, $use_width-1, $I_HEIGHT - 2, $COLOR{cm_mi_frame});
	}

	#----------------------------------------------------------
	# write image
	#----------------------------------------------------------
	open(OUT, '>', $image_outfile) || die "open";
	binmode(OUT);	# stoopid windoze
	my $it = AbsenceConfig::fetch('image_type');
	if ($it eq 'png') {
		print OUT $im->png;
	} elsif ($it eq 'gif') {
		print OUT $im->gif;
	} elsif ($it eq 'jpg') {
		print OUT $im->jpg;
	} else {
		die "unknown image_type [$it]";
	}
	close(OUT);

	print $map_fh "</MAP>\n";
	$map_fh->close;

	AbsenceDB::updateMonthModTime($group_id, $month, $year);
}

sub layoutObjectRows
{
	my ($month, $year, $people_ref) = @_;

	my $pr_info = [];
	my $sum_heights = 0;

	foreach my $oid (@{ $people_ref }) {
		$DEBUG && abslog("\nlayoutObjectRows OID=$oid");
		my @res_list = AbsenceDB::getMonthReservations($oid, $month, $year);
		if ($MULTI_RES) {
			my $layout_ref = layoutObjectRow(\@res_list, $month, $year);
			$pr_info->[$oid] = {
				height	=> ($layout_ref->{last_slot} + 1) * $SLOT_PIXELS,
				rects	=> $layout_ref->{rects},
			};
		} else {
			$pr_info->[$oid] = {
				height		=> $COL_HEIGHT,
				res_list	=> \@res_list,
			};
		}
	}

	return $pr_info;
}

sub drawObjectRow
{
	my ($im, $pr_info, $pid, $gid, $cur_y, $month, $year, $map_fh) = @_;

	if ($MULTI_RES) {
		my $rects_ref = $pr_info->{rects};
		defined($rects_ref) || die "drawObjectRow failed";
		foreach my $rect_ref (sort { $a->{prio} <=> $b->{prio} } @{ $rects_ref })
		{
			my $y1 = $cur_y + $rect_ref->{start_slot} * $SLOT_PIXELS;
			my $y2 = $cur_y + ($rect_ref->{start_slot} + $rect_ref->{num_slots})
					* $SLOT_PIXELS - 1;
			#abslog("  rect: id=$rect_ref->{mres}->{res}->{id}, y1=$y1, y2=$y2");
			my $y_bounds = [ $y1, $y2 ];
			$rect_ref->{y_bounds} = $y_bounds;
			drawReservation($im, $y_bounds, $rect_ref->{mres}, $month, $year);
		}
		writeMultiMapAreas(
			$pid, $gid,
			[ $cur_y, $cur_y + $pr_info->{height} - 1 ],
			$month, $year, $rects_ref, $map_fh,
		);
	} else {
		my $res_list = $pr_info->{res_list};
		foreach my $ref (@{ $res_list }) {
			my $y1 = $cur_y;
			my $y2 = $cur_y + $ROW_HEIGHT - 1;
			$DEBUG && print "got res=$ref->{res}->{id}, start=$ref->{bounds}->{start}, end=$ref->{bounds}->{end}\n";
			drawReservation($im, [ $y1, $y2 ], $ref, $month, $year);
		}
		writeObjectRowMapAreas($pid, $gid, $month, $year, $cur_y, $map_fh);
	}
}

sub layoutObjectRow
{
	my ($res_list, $month, $year) = @_;

	my $debug = 0;
	#$debug && abslog("**layoutObjectRow**");

	#---------------------------------------------------------------
	# if @res_list is empty, short-circuit
	#---------------------------------------------------------------
	if (!@{ $res_list }) {
		my $last_slot = $VARIABLE_HEIGHT
			? $MIN_HEIGHT * $FULL_SLOTS - 1
			: $MAX_MULTI * $FULL_SLOTS - 1;
		return {
			last_slot	=> $last_slot,
			rects		=> [],
			error		=> undef,
		};
	}

	my @blocks;
	my $numdays = AbsenceDate::daysInMonth($month, $year);
	my @day_info;
	my @day_info_b;
	for(my $d = 1; $d <= $numdays; $d++) {
		$day_info[$d] = { slots => [], types => [] };
		$day_info_b[$d] = { slots => [], types => [] };
	}

	#---------------------------------------------------------------
	# create a sortable list based on priority
	# non-block type reservations get sort_prio = priority
	# block type reservations get sort_prio > 100, where prio=0,
	#	sort_prio = 200, prio=99, sort_prio = 101
	#---------------------------------------------------------------
	my (@top_unsorted, @bottom_unsorted, @block_unsorted);
	foreach my $mres (@{ $res_list }) {
		my $type_id = $mres->{res}->{type_id};
		my $type_ref = AbsenceDB::getType($type_id);
		my $height = defined($type_ref->{height})
			? $type_ref->{height}
			: 'full';
		my $prio = defined($type_ref->{priority})
			? $type_ref->{priority}
			: 10;
		if ($height eq 'block') {
			$debug && abslog(" % rid=$mres->{res}->{id} --> block");
			push(@block_unsorted, {
				prio	=> $prio,
				slots	=> $NUM_SLOTS,
				mres	=> $mres,
			});
		} elsif ($prio >= 0) {
			$debug && abslog(" % rid=$mres->{res}->{id} --> top");
			push(@top_unsorted, {
				prio	=> $prio,
				slots	=> $HEIGHT_MAP{$height}->{slots},
				mres	=> $mres,
			});
		} else {
			$debug && abslog(" % rid=$mres->{res}->{id} --> bottom");
			push(@bottom_unsorted, {
				prio	=> $prio,
				slots	=> $HEIGHT_MAP{$height}->{slots},
				mres	=> $mres,
			});
		}
	}

	my @rects;
	my $t2b_lowest_slot;
	#---------------------------------------------------------------
	# top-to-bottom
	# sort according to order in which blocks need to be drawn on image
	# and attempt to fit on object-row
	#---------------------------------------------------------------

	foreach my $prio_group (sort_prio_groups(\@top_unsorted)) {
		foreach my $obj (@{ $prio_group }) {
			$debug && abslog("%% find slot for rid=[$obj->{mres}->{res}->{id}], num_slots=[$obj->{slots}], start=[$obj->{mres}->{bounds}->{start}], end=[$obj->{mres}->{bounds}->{end}]");
			my $starting_slot = findFreeSlotsFromTop(
				$obj->{slots}, \@day_info,
				$obj->{mres}->{bounds}->{start},
				$obj->{mres}->{bounds}->{end},
			);
			my $e_slot = $starting_slot + $obj->{slots} - 1;
			$t2b_lowest_slot = $e_slot if ($t2b_lowest_slot < $e_slot);
			#$debug && abslog("  found starting slot [$starting_slot]");
			occupySlots($starting_slot, $obj->{slots}, \@day_info, $obj->{mres});
			push(@rects, {
				prio		=> $NON_BLOCK_PRIO,
				mres		=> $obj->{mres},
				start_slot	=> $starting_slot,
				num_slots	=> $obj->{slots},
			});
		}
	}

	#---------------------------------------------------------------
	# bottom-to-top
	# sort according to order in which blocks need to be drawn on image
	# and attempt to fit on object-row
	#---------------------------------------------------------------

	#-------------------------------------
	# figure out where to start
	#-------------------------------------
	my $slot_begin = $VARIABLE_HEIGHT ? 0 : $NUM_SLOTS - 1;
	my @b_rects;
	my $highest_slot = 0;
	my $b2t_lowest_slot = 0;

	foreach my $prio_group (sort_prio_groups(\@bottom_unsorted, '-reverse')) {
		foreach my $obj (@{ $prio_group }) {
			my ($start, $end) = (
				$obj->{mres}->{bounds}->{start},
				$obj->{mres}->{bounds}->{end},
			);
			$debug && abslog("%% find b2t slot for rid=[$obj->{mres}->{res}->{id}], num_slots=[$obj->{slots}], start=[$start], end=[$end]");
			my $starting_slot
				= findFreeSlotsFromTop($obj->{slots}, \@day_info_b,
					$start, $end);
			my $e_slot = $starting_slot + $obj->{slots} - 1;
			$b2t_lowest_slot = $e_slot if ($e_slot > $b2t_lowest_slot);
			$highest_slot = $starting_slot if ($starting_slot < $highest_slot);
			$debug && abslog("starting_slot=[$starting_slot], slots=[$obj->{slots}]");
			#defined($starting_slot) || return undef;
			occupySlots($starting_slot, $obj->{slots}, \@day_info_b, $obj->{mres});
			push(@b_rects, {
				prio		=> $NON_BLOCK_PRIO,
				mres		=> $obj->{mres},
				start_slot	=> $starting_slot,
				num_slots	=> $obj->{slots},
			});
		}
	}

	my $lowest_slot = ($t2b_lowest_slot > $b2t_lowest_slot)
		? $t2b_lowest_slot
		: $b2t_lowest_slot;

	if (@b_rects) {
		$debug && abslog("b2t type reservations exist");
		$debug && abslog("min-height=$MIN_HEIGHT, lowest-slot=$b2t_lowest_slot");
		#-------------------------------------
		# figure out how much the b2t reservations have
		# to be shifted down to fit with the t2b reservations
		#-------------------------------------
		my $init_shift;
		if ($VARIABLE_HEIGHT) {
			$init_shift = ($lowest_slot > ($MIN_HEIGHT * $FULL_SLOTS))
				? $lowest_slot
				: $MIN_HEIGHT * $FULL_SLOTS - 1;
		} else {
			$init_shift = ($lowest_slot > ($MAX_MULTI * $FULL_SLOTS))
				? $lowest_slot
				: $MAX_MULTI * $FULL_SLOTS - 1;
		}

		$debug && abslog("init_shift = [$init_shift]");
		my $shift_slots = figure_shift(
			\@day_info,
			\@day_info_b,
			$lowest_slot,
			$init_shift,
			$numdays,
		);
		$debug && abslog("shift_slots = [$shift_slots]");
		#-------------------------------------
		# add b2t rects to the t2b rects with shift...
		#-------------------------------------
		foreach my $rect (@b_rects) {
			$rect->{start_slot}
				= $shift_slots - $rect->{start_slot} - $rect->{num_slots} + 1;
			push(@rects, $rect);
		}
		#-------------------------------------
		# copy the type-coincidence info from the b2t day_info structure
		#-------------------------------------
		for(my $d = 1; $d <= $numdays; $d++) {
			push(@{ $day_info[$d]->{types} }, @{ $day_info_b[$d]->{types} });
		}
		$lowest_slot = $shift_slots;
	}
	
	#---------------------------------------------------------------
	# blocks
	# sort according to order in which blocks need to be drawn on image
	# and attempt to fit on object-row
	#---------------------------------------------------------------

	#--------------------------------------------------
	# first figure out where 
	#--------------------------------------------------
	my $last_slot;
	if ($VARIABLE_HEIGHT) {
		$last_slot = ($lowest_slot >= ($MIN_HEIGHT * $FULL_SLOTS))
			? $lowest_slot
			: $MIN_HEIGHT * $FULL_SLOTS - 1;
	} else {
		$last_slot = ($lowest_slot >= ($MAX_MULTI * $FULL_SLOTS))
			? $lowest_slot
			: $MAX_MULTI * $FULL_SLOTS - 1;
	}

	$debug && abslog("blocks: last-slot=[$last_slot]");

	my @block_info;
	foreach my $obj (sort {$b->{prio} <=> $a->{prio}} @block_unsorted) {
		#my ($rid, $start, $end, $type_id, $desc) = @{ $ojb->{res} };
		my ($start, $end) = (
			$obj->{mres}->{bounds}->{start},
			$obj->{mres}->{bounds}->{end},
		);
		my $type_id = $obj->{mres}->{res}->{type_id};
		for(my $day = $start; $day <= $end; $day++) {
			push(@{ $block_info[$day] }, [$type_id, $obj->{mres}->{res}->{id}]);
		}
		#---------------------------------------------------------
		# what's this prio thingy do?
		# since non-block reservations cannot overlap oneanother,
		# they are all assigned (above) the priority $NON_BLOCK_PRIO.
		# Later, just before the reservations are actually drawn, they 
		# are sorted in ascending order according to this priority.
		# thus, the negative block-types are at the top of the list,
		# then come the non-block types, then the block-types with
		# positive priority. As a list:
		#
		# Negative-priority block type: 0-9
		# Non-block type:				100		($NON_BLOCK_PRIO)
		# Positive-priority block type: 101-110
		#---------------------------------------------------------
		my $prio = ($obj->{prio} >= 0)
			? $obj->{prio} + $NON_BLOCK_PRIO
			: 10 + $obj->{prio};
		push(@rects, {
			prio		=> $prio,
			mres		=> $obj->{mres},
			start_slot	=> 0,
			num_slots	=> $last_slot + 1,
		});
	}

	#---------------------------------------------------------------
	# now I can check if there is invalid coincidence...
	#---------------------------------------------------------------

	my $error;

	# first check non-block reservations
	for(my $day = 1; $day <= $numdays; $day++) {
		exists($day_info[$day]->{types}) || next;
		for(my $i = 0; $i < $#{ $day_info[$day]->{types} }; $i++) {
			my $first = $day_info[$day]->{types}->[$i]->[0];
			for(my $j = $i+1; $j <= $#{ $day_info[$day]->{types} }; $j++) {
				my $type_id = $day_info[$day]->{types}->[$j]->[0];
				if (!AbsenceDB::resTypeCoincidenceAllowed($first, $type_id)) {
					push(@{ $error },
						{
							type	=> 'nb_coincidence',
							date	=> "$day.$month.$year",
							ids		=> [
								$day_info[$day]->{types}->[$i]->[1],
								$day_info[$day]->{types}->[$j]->[1],
							]
						}
					);
				}
			}
		}
	}
	
	# then check block-type reservations
	for(my $day = 1; $day <= $numdays; $day++) {
		defined($block_info[$day]) || next;
		for(my $i = 0; $i < $#{ $block_info[$day] }; $i++) {
			my $first = $block_info[$day]->[$i]->[0];
			for(my $j = $i+1; $j <= $#{ $block_info[$day] }; $j++) {
				my $type_id = $block_info[$day]->[$j]->[0];
				if (!AbsenceDB::resTypeCoincidenceAllowed($first, $type_id)) {
					push(@{ $error },
						{
							type	=> 'b_coincidence',
							date	=> "$day.$month.$year",
							ids		=> [
								$block_info[$day]->[$i]->[1],
								$block_info[$day]->[$j]->[1],
							]
						}
					);
				}
			}
		}
	}

	return {
		last_slot	=> $last_slot,
		rects		=> \@rects,
		error		=> $error,
	};
}

sub figure_shift
{
	my ($di_ref, $di_b_ref, $lowest_slot, $shift, $num_days) = @_;

	my $found = 0;
	OUTER: while(1) {
		for(my $slot = 0; $slot <= $lowest_slot; $slot++) {
			my $xform = $shift - $slot;
			for(my $day = 1; $day <= $num_days; $day++) {
				if (defined($di_b_ref->[$day]->{slots}->[$slot]) &&
					defined($di_ref->[$day]->{slots}->[$xform]))
				{
					next OUTER;
				}
			}
		}
		last;
	} continue {
		$shift++;
	}

	return $shift;
}


sub sort_prio_groups
{
	my ($lref, $arg) = @_;

	my $sort_sub = ($arg eq '-reverse')
		? sub { $a <=> $b }
		: sub { $b <=> $a };
	
	my %prio_blocks;
	foreach my $bref (@{ $lref }) {
		push(@{ $prio_blocks{ $bref->{prio} } }, $bref);
	}

	my @out;
	foreach my $prio (sort $sort_sub keys(%prio_blocks)) {
		push(@out, sort_by_start_day($prio_blocks{$prio}));
	}

	return @out;
}

sub sort_by_start_day
{
	my $lref = shift;

	my $sortfunc = sub {
		$a->{mres}->{bounds}->{start} <=> $b->{mres}->{bounds}->{start};
	};

	return [ sort $sortfunc @{ $lref } ];
}

sub by_start_day
{
	return $a->{mres}->{bounds}->{start} <=> $b->{mres}->{bounds}->{start};
}

sub occupySlots
{
	my ($starting_slot, $num_slots, $di_ref, $mres) = @_;

	#my ($rid, $start, $end, $type_id, $desc) = @$res;
	my ($start, $end, $type_id) = (
		$mres->{bounds}->{start},
		$mres->{bounds}->{end},
		$mres->{res}->{type_id},
	);
	#abslog("  occupySlots: start-slot=[$starting_slot], n_slots=[$num_slots], start=[$start], end=[$end], type_id=[$type_id]");

	for(my $day = $start; $day <= $end; $day++) {
		my $slot = $starting_slot;
		while($slot < ($starting_slot + $num_slots)) {
			defined($di_ref->[$day]->{slots}->[$slot])
				&& die "occupySlots: oops. day=[$day], slot=[$slot], contents=[".$di_ref->[$day]->{slots}->[$slot]."]";
			$di_ref->[$day]->{slots}->[$slot] = $type_id;
			#$di_ref->[$day]->{slots}->[$slot] = $mres;
			#abslog("    occupying day=[$day], slot=[$slot]");
			$slot++;
		}
		push(@{ $di_ref->[$day]->{types} }, [ $type_id, $mres->{res}->{id} ]);
	}
}

sub findFreeSlotsFromTop
{
	my ($num_slots, $day_info_ref, $start, $end) = @_;

	my $found_free = 0;
	my $starting_slot;
	my $found = 0;

	my $slot = 0;

	OUTER:
	while(1) {
		if (!defined($starting_slot)) {
			$starting_slot = $slot;
			#abslog("  resetting starting_slot to [$slot]");
		}
		for(my $day = $start; $day <= $end; $day++) {
			if (defined($day_info_ref->[$day]->{slots}->[$slot])) {
				#abslog("  hit occupied space at day=[$day], slot=[$slot]");
				$found_free = 0;
				$starting_slot = undef;
				next OUTER;
			}
		}
		$found_free++;
		if ($found_free == $num_slots) {
			$found = 1;
			last;
		}
	} continue {
		$slot++;
	}

	#return $found ? $starting_slot : undef;
	# $found should contain something real no matter what...
	$found || die "this can't happen...";
	return $starting_slot;
}

sub currentMonth
{
	my ($month, $year) = @_;
	my ($m, $y) = (localtime)[4,5];
	$m++;
	$y += 1900;

	#print "m=$m, month=$month, y=$y, year=$year<BR>\n";
	if (($y == $year) && ($m == $month)) { return 1; }

	return 0;
}

sub markCurrentDay
{
	my ($im, $month, $year) = @_;

	my ($d) = (localtime)[3];
	if ($DEBUG) {
		abslog("--- CURRENT DAY START ---");
		abslog("current day-num = [$d]");
		abslog("localtime=".localtime());
		abslog("env:");
		foreach my $ev (keys(%ENV)) {
			abslog("$ev=[$ENV{$ev}]");
		}
		abslog("--- CURRENT DAY END ---");
	}

	my $style	= AbsenceConfig::fetch('curr_day_style');
	my $brow	= AbsenceConfig::fetch('curr_day_begin_row');

	if ($brow > 3) { $brow = 3; }

	my $x1 = $ROW_LABEL_WIDTH + ($d - 1) * $COL_WIDTH + 1;
	my $x2 = $ROW_LABEL_WIDTH + $d * $COL_WIDTH - 1;
	#my $y1 = $HEADER_LINES * $ROW_HEIGHT + 1;
	my $y1 = ($brow - 1) * $HEADER_ROW_HEIGHT + 1;
	my $y2 = $I_HEIGHT - 2;

	if ($style eq 'fill') {
		# fill column
		$im->filledRectangle($x1, $y1, $x2, $y2, $COLOR{currday});
	} else {
		# put frame around column...
		$im->rectangle($x1, $y1, $x2, $y2, $COLOR{currday});
		$im->rectangle($x1+1, $y1+1, $x2-1, $y2-1, $COLOR{currday});
	}
}

#---------------------------------------------------------------
# findAbsence()
#
# is passed month, year, and (x,y) coordinates of click
#
# returns array of two elements.  The first can be 'none', 'absence',
# 'holiday' or 'start'.  The second depends on the first.
#
# The values mean:
# 'none'		user clicked on the first line (CW, or on the second
#				line (month-day) but not on a holiday
# 'absence'		user clicked on a marked absence
# 'holiday'		user clicked on a month-day marked as holiday
# 'start'		user clicked on a user-line not marked as absence
#
# The second element is:
# 'none'		blank
# 'absence'		the absence-ID
# 'holiday'		returns ref to a holiday array in format:
#				[<day>, <month>, <year>, <holiday-name>]
# 'start'		a reference to an array containing the person_id and date
#				corresponding to the click, in format
#				[<person_id>,<D>,<M>,<Y>]
#---------------------------------------------------------------
sub findAbsence
{
	my ($month, $year, $gid, $x, $y) = @_;

	my $numdays = AbsenceDate::daysInMonth($month, $year);
	my $sav = $HEADER_LINES * $HEADER_ROW_HEIGHT;

	# sanity check
	if ($x < $ROW_LABEL_WIDTH) {
		return (none => '');
	}

	# check if click in padding
	if ($x > ($ROW_LABEL_WIDTH + $numdays * $COL_WIDTH)) {
		return (none => 'stop-that');
	}

	# get row number from y-coord
	my $row = int($y/$ROW_HEIGHT);		# returns row 0-n

	if ($row == 1) {
		# month-day row
		my $hol = findHoliday($month, $year, $x);
		if (defined($hol)) {
			return (holiday => $hol);
		}
		return (none => 'no-holiday');
	}

	# at this point, coordinates must be in a person row
	if ($y <= $sav) {
		return (none => '');
	}

	# get people
	my @people = AbsenceDB::getPeople($gid);
	my $person_id = $people[$row - $HEADER_LINES];

	my ($start, $end, $short_desc, $x1, $x2);

	# go through absences
	my $rid;
	$DEBUG && print "finding absence: $month/$year, ($x,$y)..\n";
	my @res_list = AbsenceDB::getMonthReservations($person_id, $month, $year);
	foreach my $res (@res_list) {
		($rid, $start, $end, $short_desc) = @$res;
		$x1 = $ROW_LABEL_WIDTH + ($start - 1) * $COL_WIDTH + 1;
		$x2 = $ROW_LABEL_WIDTH + $end * $COL_WIDTH - 1;
		$DEBUG && print "checking id=$rid, x1=$x1, x2=$x2...\n";
		if (($x >= $x1) && ($x < $x2)) {
			return (absence => $rid);
		}
	}

	# must be a click on a person-row but not on absence...
	# figure out date

	$x -= $ROW_LABEL_WIDTH;
	my $day = int($x/$COL_WIDTH) + 1;
	return (start => [$person_id, $day, $month, $year]);
}

sub markWeekends
{
	my ($im, $month, $year) = @_;

	my $numdays = AbsenceDate::daysInMonth($month, $year);

	my ($x1, $y1, $x2, $y2, $start, $end);

	$y1 = 0;
	$y2 = $I_HEIGHT;

	my $we_color = $COLOR{we};
	if ($MARK_CM && currentMonth($month, $year)) {
		$we_color = $COLOR{cm_we};
	}

	# find first sat/sun
	my $i = 1;
	my $wd;
	while(($wd = AbsenceDate::weekDayFromDate($i, $month, $year)) % 6 != 0) {
		$i++;
	}
	$start = $i;
	$end = ($wd == 6) ? $i + 1 : $i;

	do {
		$x1 = $ROW_LABEL_WIDTH + ($start - 1) * $COL_WIDTH + 1;
		$x2 = $ROW_LABEL_WIDTH + $end * $COL_WIDTH - 1;
		$im->filledRectangle($x1, $y1, $x2, $y2, $we_color);
		$start = $end + 6;
		$end = (($start + 1) <= $numdays) ? $start + 1 : $start;
	} while($start <= $numdays);
}

sub findHoliday
{
	my ($month, $year, $x) = @_;

	my $numdays = AbsenceDate::daysInMonth($month, $year);
	#my @holidays = AbsenceHolidays::getList($month, $year);
	my @holidays = AbsenceDB::getHolidayList($month, $year);

	if (!@holidays) { return undef; }

	my ($x1, $x2);

	my $i;
	for($i = 1; $i <= $numdays; $i++) {
		if ($i == $holidays[0]->[0]) {
			$x1 = $ROW_LABEL_WIDTH + ($i - 1) * $COL_WIDTH + 1;
			$x2 = $ROW_LABEL_WIDTH + $i * $COL_WIDTH - 1;
			if (($x >= $x1) && ($x <= $x2)) {
				$holidays[0]->[2] = $year;
				return $holidays[0];
			}
			shift(@holidays);
			if (!@holidays) { last; }
		}
	}

	return undef;
}

sub markHeaderHolidays
{
	my ($im, $month, $year, $map_fh) = @_;

	$HOLIDAY_SCHEME eq 'advanced' && !$HOL_SHOW_HEAD && return;

	my $numdays = AbsenceDate::daysInMonth($month, $year);

	my @params;
	my ($country_id, $region_id);

	if ($HOLIDAY_SCHEME eq 'advanced') {
		if (defined($HOL_HEAD_COUNTRY)) {
			$country_id = AbsenceDB::getCountryId(code => $HOL_HEAD_COUNTRY);
			if (defined($country_id)) {
				push(@params, country => $country_id);
			}
		}
		if (defined($HOL_HEAD_REGION)) {
			$region_id = AbsenceDB::getRegionId($HOL_HEAD_REGION);
			if (defined($region_id)) {
				push(@params, region => $region_id);
			}
		}
	}

	$HDEBUG && abslog("params=[".join(',', @params)."]");
	my @holidays = AbsenceDB::getHolidayList($month, $year, @params);
	$HDEBUG && abslog("got [".scalar(@holidays)."] holidays");

	if (!@holidays) { return; }

	my ($y1, $y2, $start, $end);

	$y1 = $HEADER_ROW_HEIGHT;
	if ($HOLIDAY_MARK eq 'small' || $HOLIDAY_SCHEME eq 'advanced') {
		$y2 = $HEADER_LINES * $HEADER_ROW_HEIGHT;
	} else {
		$y2 = $I_HEIGHT - 1;
	}

	my $map_y2 = $HEADER_LINES * $HEADER_ROW_HEIGHT;
	my $map = drawHolidays($im, \@holidays, $y1, $y2, $map_y2, $numdays);
	print $map_fh $map;
}

sub drawHolidays
{
	my ($im, $hol_ref, $y1, $y2, $map_y2, $numdays) = @_;

	my ($i, $x1, $x2);
	my $map = '';

	for($i = 1; $i <= $numdays && @{ $hol_ref }; $i++) {
		if ($i == $hol_ref->[0]->[0]) {
			$x1 = $ROW_LABEL_WIDTH + ($i - 1) * $COL_WIDTH + 1;
			$x2 = $ROW_LABEL_WIDTH + $i * $COL_WIDTH - 1;
			$im->filledRectangle($x1, $y1, $x2, $y2, $COLOR{bholiday});
			$map .= qq[<AREA SHAPE="rect" COORDS="$x1,$y1,$x2,$map_y2" NOHREF TITLE="$hol_ref->[0]->[3]">\n];
			shift(@{ $hol_ref });
		}
	}

	return $map;
}

sub allocateColors
{
	my $im = shift;

	my @types = AbsenceDB::getTypes('all');

	# allocate colors
	
	# background must be allocated first...
	$COLOR{bg} = $im->colorAllocate(@{$COLOR_VALUES{bg}});

	foreach my $col (keys(%COLOR_VALUES)) {
		if ($col eq 'bg') { next; }
		#print "COL[$col]: [".join(',', @{$COLOR_VALUES{$col}}),"]<BR>\n";
		$COLOR{$col} = $im->colorAllocate(@{$COLOR_VALUES{$col}});
		#$HUHU{$col} = (MAX(@{$COLOR_VALUES{$col}}) < 128) ? 1 : 0;
	}

	foreach my $type_ref (@types) {
		my $trans = defined($type_ref->{transparency})
			? int((127 * $type_ref->{transparency}) / 100)
			: 0;
		$COLOR{ $type_ref->{id} } = $im->colorAllocateAlpha(
			$type_ref->{color_red},
			$type_ref->{color_green},
			$type_ref->{color_blue},
			$trans
		);
		#$HUHU{$type} = (MAX(@{$types->{$type}}) < 128) ? 1 : 0;
	}
}

#void RGBtoHSV( float r, float g, float b, float *h, float *s, float *v )
sub RGBtoHSV
{
	my ($r, $g, $b) = @_;

	my ($min, $max, $h, $s, $v);

	#float min, max, delta;

	$min = MIN($r, $g, $b);
	$max = MAX($r, $g, $b);
	$v = $max;							# v

	my $delta = $max - $min;

	if ($max != 0) {
		$s = $delta / $max;				# s
	} else {
		# r = g = b = 0					# s = 0, v is undefined
		$s = 0;
		$h = -1;
		return($h, $s, $v);
	}

	if ($r == $max ) {
		$h = ($g - $b) / $delta;		# between yellow & magenta
	} elsif ($g == $max) {
		$h = 2 + ($b - $r) / $delta;	# between cyan & yellow
	} else {
		$h = 4 + ($r - $g) / $delta;	# between magenta & cyan
	}

	$h *= 60;							# degrees
	if ($h < 0) {
		$h += 360;
	}

	($h, $s, $v);
}

sub MIN
{
	my $min = 99999999;
	foreach (@_) {
		if ($_ < $min) { $min = $_; }
	}

	$min;
}

sub MAX
{
	my $max = 0;
	foreach (@_) {
		if ($_ > $max) { $max = $_; }
	}

	$max;
}

sub drawCwBars
{
	my ($image, $month, $year) = @_;

	my $monthdays = AbsenceDate::daysInMonth($month, $year);
	my ($cw, $cy, $x1, $y1, $x2, $y2);

	my $wd;
	my $day = 1;
	while(
		(($wd = AbsenceDate::weekDayFromDate($day, $month, $year)) == 0) ||
		($wd == 6)
	)
	{
		$day++;
	}
	$DEBUG && print "day = $day\n";

	my $cw_color = $COLOR{cw};
	if ($MARK_CM && currentMonth($month, $year)) {
		$cw_color = $COLOR{cm_cw};
	}

	# $day is first monday - friday
	while($day <= $monthdays) {
		$x1 = $ROW_LABEL_WIDTH + ($day - 1) * $COL_WIDTH + 1;
		$y1 = 1;
		my $end;
		if (($day > 1) || ($wd == 1)) {
			$end = (($day + 4) <= $monthdays)
				? 4
				: $monthdays - $day;
		} else {
			$end = 5 - $wd;
		}
		$x2 = $ROW_LABEL_WIDTH + ($day + $end) * $COL_WIDTH - 1;
		$y2 = $HEADER_ROW_HEIGHT - 1;
		$image->filledRectangle($x1, $y1, $x2, $y2, $cw_color);
		if ($end == 0) {
			# if CW-bar is only one day long, no room for CW-label
			$day += 3 + $end;
			$wd = 0;
			next;
		}
		($cw, $cy) = AbsenceDate::calendarWeek(
			{
				day		=> $day,
				month	=> $month,
				year	=> $year,
			}
		);
		my $mid = int(($x2 - $x1)/2);
		$image->string(gdSmallFont,
			$x1 + $mid - 13,
			3,
			"CW $cw",
			$COLOR{text_dark});
		if ($wd) {
			$day += 8 - $wd;
			$wd = 0;
		} else {
			$day += 7;
		}
	}
}

#---------------------------------------------------------------
# drawReservation()
#
# marks a reservation (absence).  If the skipping of weekends
# and holidays has been configured, findBlocks() is called to
# find valid blocks, i.e., sub-parts of reservation, that can be
# marked (weekdays without holidays).
#---------------------------------------------------------------
sub drawReservation
{
	#my ($im, $y_offset, $start, $end, $type_id, $desc, $rid, $month, $year) = @_;
	#my ($im, $y_offset, $ref, $month, $year) = @_;
	my ($im, $y_bounds, $ref, $month, $year) = @_;

	my ($type_id, $desc, $rid, $start, $end) = (
		$ref->{res}->{type_id},
		$ref->{res}->{description},
		$ref->{res}->{id},
		$ref->{bounds}->{start},
		$ref->{bounds}->{end},
	);

	my $type_ref = AbsenceDB::getType($type_id);
	defined($type_ref) || die "type_id [$ref->{res}->{id}] not found";

	$DEBUG && abslog("SKIP_WE_HOL=[$SKIP_WE_HOL]");

	# call the regular unless skipping desired
	if (!$SKIP_WE_HOL || !$type_ref->{skip_non_workdays}) {
		_drawReservation($im, $y_bounds, $start, $end, $type_ref, $desc, $rid);
		return;
	}

	my $i = $start;

	#$DEBUG && abslog("\nstart=$ref->{bounds}->{start}, end=$ref->{bounds}->{end}, mon=$month, year=$year, rid=$ref->{res}->{id}");

	my ($blocksref, $invalref) = findBlocks($ref->{bounds}, $month, $year);
	foreach my $blk (@{$blocksref}) {
		_drawReservation($im, $y_bounds, $blk->[0], $blk->[1], $type_ref, $desc, $rid);
	}

	$SKIP_MARK_SK || return;

	foreach my $day (@{$invalref}) {
		#markInvalid($im, $y_offset, $day, $type_id);
		markInvalid($im, $y_bounds, $day, $type_id);
	}
}

#---------------------------------------------------------------
# findBlocks()
#
# find sub-day-ranges within a day-range that consist only of
# weekdays that are not holidays.
#
# returns references to two arrays:
#
# 1: an array of references to arrays of the form (start, end):
# ([startday1, endday1], [startday2, endday2], [startdayN, enddayN])
#
# 2: an array of days that have been skipped
#---------------------------------------------------------------
sub findBlocks
{
	my ($bounds, $month, $year) = @_;

	#my @holidays = AbsenceHolidays::getList($month, $year);
	my @holidays = AbsenceDB::getHolidayList($month, $year);
	my %hols = map { $_->[0], 1} @holidays;

	my ($want, $s, @blocks, @invalid);

	my $day = $bounds->{start};

	my $in_block = 0;
	my $last;
	while($day <= $bounds->{end}) {
		$want = (AbsenceDate::weekDayFromDate($day, $month, $year) % 6 != 0)
			&& !exists($hols{$day});
		if ($in_block && !$want) {
			push(@blocks, [$s, $last]);
			#abslog("pushing block: $s -> $last");
			$in_block = 0;
			push(@invalid, $day);
		} elsif (!$in_block && $want) {
			$in_block = 1;
			$s = $day;
		} elsif (!$in_block) {
			push(@invalid, $day);
		}
		$last = $day;
		$day++;
	}

	if ($in_block) {
		push(@blocks, [$s, $bounds->{end}]);
		#abslog("pushing block: $s -> $bounds->{end}");
	}

	return (\@blocks, \@invalid);
}

#------------------------------------------------------------------
# drawReservation()
#
# $start and $end are the month-day numbers within the month in
# question
#------------------------------------------------------------------
sub _drawReservation
{
	my ($im, $y_bounds, $start, $end, $type_ref, $desc, $rid) = @_;

	my $type_id = $type_ref->{id};

	#abslog("_addRes: rid=$rid, start=$start, end=$end");
	my ($x1, $x2);
	my ($y1, $y2) = @{ $y_bounds };
	$x1 = $ROW_LABEL_WIDTH + ($start - 1) * $COL_WIDTH + 1;
	#$y1 = $y_offset + 1;
	$x2 = $ROW_LABEL_WIDTH + $end * $COL_WIDTH - 1;
	#$y2 = $y_offset + $ROW_HEIGHT - 1;

	$DEBUG && print "  ABS: ($x1, $y1), ($x2, $y2) - $type_id, ID=$rid\n";

	$im->filledRectangle($x1, $y1, $x2, $y2, $COLOR{$type_id});
	
	if ($MULTI_RES && (($y2 - $y1) > 3)) {
		# draw frame around rectangle to differentiate it from neighbor
		$im->rectangle($x1, $y1, $x2, $y2, $COLOR{res_border});
	}

	#-----------------------------------------------------------------
	# experience shows that approximately 3 characters in the chosen
	# label font fit in a column of width 20 ($COL_WIDTH). Yeah, it's
	# a hack.
	#-----------------------------------------------------------------

	# if not enough space for text, return
	(($y2 - $y1) < 11) && return;

	# if no description present, return
	(!defined($desc) || length($desc) == 0) && return;

	my $num_days = $end - $start + 1;
	my $mid = int(($x2 - $x1)/2);

	my ($half, $lab_xpos);

	# need to call from_to() before "substr()" is called so that the string
	# is chopped at proper character boundaries and not at octets

	$DEBUG && abslog("length of desc before conversion: ".length($desc));
	from_to($desc, 'utf8', 'iso-8859-1') || die "from_to failed: $!";
	$DEBUG && abslog("length of desc after conversion: ".length($desc));

	if (int((length($desc)+2)*20/3) >= $num_days * $COL_WIDTH) {
		$lab_xpos = $x1 + 2;
		my $trim_len = ($num_days * $COL_WIDTH * 3)/20;
		$DEBUG && abslog("trim-len=[$trim_len]");
		$desc = substr($desc, 0, $trim_len);
	} else {
		$half = int(length($desc)/6.0 * $COL_WIDTH);
		$lab_xpos = $x1 + $mid - $half;
	}

	#my $offset = $MULTI_RES ? ($y2 - $y1)/2 - 7;
	my $offset = int(($y2 - $y1)/2) - 6;

	my $color = determineLabelColor(
		$type_ref->{color_red},
		$type_ref->{color_green},
		$type_ref->{color_blue},
	);

	$im->string(gdSmallFont,
		$lab_xpos,
		$y1 + $offset,
		$desc,
		$COLOR{$color});
}

sub determineLabelColor
{
	my ($r, $g, $b) = @_;
	#if (0.212671 * ($r/255) + 0.715160 * ($g/255) + 0.072169 * ($b/255) > 0.5) {
	if (0.212671 * $r + 0.715160 * $g + 0.072169 * $b < 127.5) {
		return 'white';
	}
	return 'black';
}

sub markInvalid
{
	#my ($im, $y_offset, $day, $type_id) = @_;
	my ($im, $y_bounds, $day, $type_id) = @_;

	my ($x1, $x2);
	#my ($y1, $y2) = @{ $y_bounds };
	$x1 = $ROW_LABEL_WIDTH + ($day - 1) * $COL_WIDTH + 1;
	#$y1 = $y_offset + 1;
	$x2 = $ROW_LABEL_WIDTH + $day * $COL_WIDTH - 1;
	#$y2 = $y_offset + $ROW_HEIGHT - 1;

	$im->rectangle($x1, $y_bounds->[0], $x2, $y_bounds->[1], $COLOR{$type_id});
}

sub legendFilePath
{
	my $it = AbsenceConfig::fetch('image_type');
	my $legend_abs = "$IMAGE_DIR_ABS/legend.$it";
	my $legend_rel = "$IMAGE_DIR_REL/legend.$it";

	return ($legend_rel, $legend_abs);
}

sub removeLegendFile
{
	my $abs = (legendFilePath())[1];
	$DEBUG && abslog("removing legend file [$abs]");
	unlink($abs) || abslog("error unlinking legend file [$abs]: $!");
}

sub makeLegend
{
	#my $outfile = shift;
	my $outfile = (legendFilePath())[1];

	my $type;
	my $tag_width;
	my $hl_tmp = AbsenceConfig::fetch('holiday_label');
	my $hl_convtest = $hl_tmp;

	# worry about encoding of holiday-label
	my $holiday_label;
	eval {
		$holiday_label = decode('utf8', $hl_convtest, 1);
	};
	if ($@) {
		# decode() failed, so assume charset=Latin1
		$holiday_label = decode('iso-8859-1', $hl_tmp);
	}
	
	my $style = AbsenceConfig::fetch('legend_style');
	if ($style =~ /^constant:(\d+)$/) {
		$tag_width = $1;
		$style = 'constant';
	} elsif ($style ne 'fit') {
		die "unrecognized legend-style [$style]";
	}

	my $legend_mtime;
	# check if legend needs to be produced.
	if (-e $outfile) {
		#my $type_mtime = AbsenceTypes::modTime();
		my $type_mtime = AbsenceDB::getTypeModTime();
		my $legend_mtime = (stat($outfile))[9];
		$DEBUG && abslog("type_mtime=[$type_mtime], legend_mtime=[$legend_mtime]");
		if ($type_mtime < $legend_mtime) {
			# print "OLDER.<BR>\n";
			# print "type file OLDER than legend.<BR>\n";
			return $legend_mtime;
		}
		# younger
		# print "type file YOUNGER than legend.<BR>\n";

		# if the legend has changed, all month-images must be deleted
		abslog("removing month images (makeLegend)");
		removeMonthImages();
	}

	my @types = AbsenceDB::getTypes();
	my $num_types = scalar(@types) + 1;

	my $tag_sep = 20;
	my $i_width = 60;
	my ($top_width, $bottom_width, $block_width);
	my (@full, @block, @short);
	my $label;
	my $col_id;
	my @other;
	my $height;
	my $name;

	foreach $type (@types) {
		$name = decode_utf8($type->{name});
		my $w = ($style eq 'constant')
			? $tag_width
			: length($name) * 6 + 5;
		my $height = defined($type->{height}) ? $type->{height} : 'full';
		$DEBUG && abslog("type name=[$name], height=[$height]");
		if (!$MULTI_RES || ($height eq 'full')) {
			push(@full, [$type, $name, $w]);
			$top_width += $w + $tag_sep;
		} elsif ($height eq 'block') {
			push(@block, [$type, $name, $w]);
			$block_width += $w + $tag_sep;
		} else {
			push(@other, [$type, $name, $w, $height]);
			$bottom_width += $w + $tag_sep;
		}
	}

	my $w = ($style eq 'constant')
		? $tag_width
		: (length($holiday_label) * 6 + 5);

	$top_width += $w + $tag_sep;
	@full = sort { lc($a->[1]) cmp lc($b->[1]) } @full;
	push(@full, [ 'bholiday', $holiday_label, $w ]);

	@block = sort { lc($a->[1]) cmp lc($b->[1]) } @block;
	@other = sort { lc($a->[1]) cmp lc($b->[1]) } @other;

	my $i_height = ($MULTI_RES && (@block || @other)) ? 60 : 25;
	$i_width += ($top_width > $bottom_width) ? $top_width : $bottom_width;
	my $block_start = $i_width;
	$i_width += $block_width;

	my $im = new GD::Image($i_width, $i_height);

	allocateColors($im);

	# do background
	# I don't need to do this.  Because the background color
	# is allocated first (in "allocateColors()"), GD automagically
	# uses it for the background
	#$im->filledRectangle(0, 0, $i_width, $i_height, $COLOR{bg});

	$im->string(gdSmallFont,
		2, 5,
		'Legend:',
		$COLOR{text_dark});

	my $x = 60;

	my $width;
	foreach my $ref (@full) {
		($type, $label, $width) = @$ref;
		$col_id = ref($type) ? $type->{id} : $type;
		$im->filledRectangle($x, 4, $x+$width, 21, $COLOR{$col_id});
		#from_to($label, 'utf8', 'iso-8859-1') || die "from_to failed: $!";
		$label = encode('iso-8859-1', $label);
		drawLegendLabel($im, $type, $label, $x+2, 6);
		$x += $width + $tag_sep;
	}

	if ($MULTI_RES) {
		$x = 60;
		my $base_y = 30;

		foreach my $ref (@other) {
			($type, $label, $width, $height) = @$ref;
			$col_id = $type->{id};
			$im->rectangle($x, $base_y, $x+$width, $base_y + 25, $COLOR{text_dark});
			my $pix_h = ($height eq 'half')
				? 2 * $SLOT_PIXELS
				: $SLOT_PIXELS;
			$im->filledRectangle($x+2, $base_y + 3, $x+$width - 2, $base_y + 3 + $pix_h, $COLOR{$col_id});
			#from_to($label, 'utf8', 'iso-8859-1') || die "from_to failed: $!";
			$label = encode('iso-8859-1', $label);
			$im->string(gdSmallFont,
				$x+2, $base_y + 10,
				$label,
				$COLOR{text_dark});
			$x += $width + $tag_sep;
		}

		$x = $block_start;
		foreach my $ref (@block) {
			($type, $label, $width) = @$ref;
			my $col_id = $type->{id};
			$im->filledRectangle($x, 4, $x+$width, 50, $COLOR{$col_id});
			$im->rectangle($x, 4, $x+$width, 50, $COLOR{text_dark});
			#from_to($label, 'utf8', 'iso-8859-1') || die "from_to failed: $!";
			$label = encode('iso-8859-1', $label);
			drawLegendLabel($im, $type, $label, $x+2, 20);
			$x += $width + $tag_sep;
		}
	}

	my $fh = FileHandle->new($outfile, 'w');
	defined($fh) || die "open $outfile for writing";
	binmode($fh);	# stoopid windoze
	my $it = AbsenceConfig::fetch('image_type');
	if ($it eq 'png') {
		print $fh $im->png;
	} elsif ($it eq 'gif') {
		print $fh $im->gif;
	} elsif ($it eq 'jpg') {
		print $fh $im->jpg;
	} else {
		die "unknown image_type [$it]";
	}
	close($fh);

	return time();
}

sub drawLegendLabel
{
	my ($im, $type, $label, $x, $y) = @_;

	my $color;
	if (ref($type)) {
		$color = determineLabelColor(
			$type->{color_red},
			$type->{color_green},
			$type->{color_blue},
		);
	} else {
		$color = determineLabelColor(@{ $COLOR_VALUES{bholiday} });
	}

	$im->string(gdSmallFont,
		$x, $y,
		$label,
		$COLOR{$color});
}

sub writeObjectRowMapAreas
{
	my ($pid, $gid, $month, $year, $cur_y, $map_fh) = @_;

	my ($x1, $y1, $x2, $y2, $start, $end);

	my $rid;
	my $title;
	my $target;
	my $inst;

	my $numdays = AbsenceDate::daysInMonth($month, $year);

	$y1 = $cur_y + 1;
	$y2 = $cur_y + $ROW_HEIGHT - 1;
	my $read_only = 0;

	my $p = AbsenceDB::getPerson($pid, 'name');
	
	#print "<!-- checking absences for $p, person_id [$pid] -->\n";
	print $map_fh "<!-- start person_id=[$pid] -->\n";

	$DEBUG && print "getting month absences for $p, person_id [$pid]\n";
	my @res_list = AbsenceDB::getMonthReservations($pid, $month, $year);
	#my @sorted = sort {$a->[1] <=> $b->[1]} @res_list;
	my @sorted = sort { $a->{bounds}->{start} <=> $b->{bounds}->{start} } @res_list;

	#---------------------------------------------------------------
	# if $read_only evaluates to 'true', a client-map will be
	# produced with no clickable areas
	#---------------------------------------------------------------
	if ($read_only) {
		foreach my $res (@sorted) {
			my $desc = $res->{res}->{description};
			my ($start, $end) = ($res->{bounds}->{start},$res->{bounds}->{end});
			$x1 = $ROW_LABEL_WIDTH + ($start - 1) * $COL_WIDTH + 1;
			$x2 = $ROW_LABEL_WIDTH + $end * $COL_WIDTH - 1;
			$title = qq{TITLE="$desc"};
			print $map_fh qq[<AREA SHAPE="rect" COORDS="$x1,$y1,$x2,$y2" NOHREF $title>\n];
		}
		$cur_y += $ROW_HEIGHT;
		next;
	}

	my $day = 1;
	while($day <= $numdays) {
		$x1 = $ROW_LABEL_WIDTH + ($day - 1) * $COL_WIDTH + 1;
		if (@sorted && ($sorted[0]->{bounds}->{start} == $day)) {
			my $desc = $sorted[0]->{res}->{description};
			my $rid = $sorted[0]->{res}->{id};
			my ($start, $end) = (
				$sorted[0]->{bounds}->{start},
				$sorted[0]->{bounds}->{end},
			);
			$x2 = $ROW_LABEL_WIDTH + $end * $COL_WIDTH - 1;
			$title = qq{TITLE="$desc"};
			$target = "res,$rid";
			shift(@sorted);
			$day = $end + 1;
		} else {
			$x2 = $ROW_LABEL_WIDTH + $day * $COL_WIDTH - 1;
			$title = '';
			$target = "new,$pid,$gid,$day,$month,$year";
			$day++;
		}
		$target = qq[HREF="$CGI_DIR/absence-click.pl$inst?$target"];
		print $map_fh qq[<AREA SHAPE="rect" COORDS="$x1,$y1,$x2,$y2" $target $title>\n];
	}
	print $map_fh "<!-- end person_id=[$pid] -->\n";
}

sub writeMultiMapAreas
{
	my ($pid, $gid, $y_bounds, $month, $year, $rects_ref, $map_fh) = @_;

	my ($x1, $y1, $x2, $y2);
	my ($sd, $ed, $rid, $desc, $target, $title);
	my $inst;

	#-----------------------------------------------------------------
	# sort the reservations rectangles by descending priority
	#-----------------------------------------------------------------
	foreach my $rect (sort {$b->{prio} <=> $a->{prio}} @{ $rects_ref }) {
		$sd = $rect->{mres}->{bounds}->{start};
		$ed = $rect->{mres}->{bounds}->{end};
		$x1 = $ROW_LABEL_WIDTH + ($sd - 1) * $COL_WIDTH + 1;
		$x2 = $ROW_LABEL_WIDTH + $ed * $COL_WIDTH - 1;
		($y1, $y2) = @{ $rect->{y_bounds} };
		$rid = $rect->{mres}->{res}->{id};
		$desc = $rect->{mres}->{res}->{description};
		$target = qq{HREF="$CGI_DIR/absence-click.pl$inst?res,$rid"};
		$title = qq{TITLE="$desc"};
		print $map_fh qq{<AREA SHAPE="rect" COORDS="$x1,$y1,$x2,$y2" $target $title>\n};
	}

	my $numdays = AbsenceDate::daysInMonth($month, $year);
	#-----------------------------------------------------------------
	# now write the areas for unreserved days
	#-----------------------------------------------------------------
	($y1, $y2) = @{ $y_bounds };

	my $day = 1;

	while($day <= $numdays) {
		$x1 = $ROW_LABEL_WIDTH + ($day - 1) * $COL_WIDTH + 1;
		$x2 = $ROW_LABEL_WIDTH + $day * $COL_WIDTH - 1;
		$target = "new,$pid,$gid,$day,$month,$year";
		$target = qq[HREF="$CGI_DIR/absence-click.pl$inst?$target"];
		print $map_fh qq[<AREA SHAPE="rect" COORDS="$x1,$y1,$x2,$y2" $target>\n];
		$day++;
	}
}

sub removeMonthImages
{
	my (@gid_list) = @_;

	$DEBUG && abslog("removeMonthImages() called with gids=[".join(',', @gid_list)."]");
	local(*DIR);
	opendir(DIR, $IMAGE_DIR_ABS) || die "opendir [$IMAGE_DIR_ABS]";
	
	my $it = AbsenceConfig::fetch('image_type');
	my $file;

	my $re = '('.join('|',@gid_list).')';

	while($_ = readdir(DIR)) {
		if ((@gid_list && /^absence-${re}-.*\.$it$/) ||
			(!@gid_list && /^absence-.*\.$it$/))
		{
			$file = "$IMAGE_DIR_ABS/$_";
			#$DEBUG && abslog("unlinking [$file]");
			$DEBUG && abslog("unlinking [$file]");
			unlink($file) || dbg("failed: $!");
			my $map = $file;
			$map =~ s/\.$it$/.map/;
			-f $map && unlink($map) || dbg("failed to unlink map file: $!");
		}
	}
	closedir(DIR);
}

1;

__END__

=head1 NAME

AbsenceImage.pm - handles manipulation of month-images and related functions

=head1 SYNOPSIS

 AbsenceImage::create($month, $year, $group_id, $image_file, $mapname);
 $arr_ref        = AbsenceImage::findAbsence($month, $year, $group_id, $x, $y);
 $holiday_struct = AbsenceImage::findHoliday($month, $year, $x);
 $legend_mtime   = AbsenceImage::makeLegend();
 ($rel, $abs)    = AbsenceImage::legendFilePath();

=head1 DESCRIPTION

C<AbsenceImage.pm> presents an interface to month-image data.  It
handles the generation of month-images and, if map-type is server (which
is not well tested any more), also the querying of absence-data based on
an (X,Y) coordinate-pair (see C<findAbsence()>, C<findHoliday()>).

=head1 FUNCTIONS

=over 4

=item B<create()>

generates a month-image, which is written to output file specified.

=item B<findAbsence()>

map-type = server

This is used when server-side image maps have been configured.
given a (X,Y) coordinate pair, determines what, if any, absence
entity corresponds to the pair, and returns a corresponding
data-structure as a list.  The possible return values are:

=over 4

=item (none => '')

no entity exists matching the supplied coordinates

=item (none => 'stop-that')

coordinates correspond to padding at end of month with less than 31 days.

=item (none => 'no-holiday')

coordinates correspond to row holding day-numbers, but not to a holiday.

=item (holiday => [$day, $month, $year, $holiday_description])

coordinates correspond to row holding day-numbers and to a holiday.

=item (absence => $absence_id)

coordinates correspond to an existing absence.

=item (start => [$person_id, $day, $month, $year])

coordinates correspond to a valid unreserved person/day combination.

=back

=item B<findHoliday()>

map-type = server

findHoliday() isn't really public.  It's normally used by findAbsence().
Takes a $month, $year, and $x coordinate, and returns an array reference
if the $x coordinate corresponds to a holiday, or B<undef> if not.

=back

=head1 INTERNAL FUNCTIONS

=over 4

=item B<layoutObjectRow()>

layoutObjectRow() is by far the nastiest function in this module. It must
be called with a list of reservations for an object (person) for a particular
month, and does more-or-less the following:

1. goes through reservation list and sorts into 3 groups based on the height
of the reservation type (RT): "block", "non-block" top-to-bottom, and
"non-block" bottom-to-top

2. the non-block height top-to-bottom reservations are then sorted according
to priority and then laid out on the object-row.

3. the non-block bottom-to-top reservations are sorted and laid out on
the object-row.

4. the block reservations are sorted according to priority and laid out
on the object-row.

5. the layout is checked for forbidden combinations of reservation-types or
lack of space and an error-list is created.

6. the result is returned, which consists of the number of the last slot
used for the layout, the list of reservations with beginning and ending
slot numbers, and the error-list.

=back

=head1 AUTHOR

Robert Urban <urban@tru64.org>

Copyright (C) 2003 Robert Urban

=cut
