package AbsenceLog;

use Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(abslog);

use Fcntl (':flock'); # import LOCK_* constants
use FileHandle;

use AbsenceConfig;

our $LOGFILE;

my $AUTH	= AbsenceConfig::fetch('authentication');
my $IPADDR	= $ENV{'REMOTE_ADDR'};
my $AUTH_USER;
my $LOG_METHOD = '';

sub SEEK_END() { 2; }	# not portable, but what the hell

sub import
{
	shift;  # get rid of $class

	#print "IMPORT...\n";
	my $word;
	my @args;
	while ($word = shift) {
		#print "word=[$word]\n";
		if ($word eq 'LOGFILE') {
			#print "parm = [$_[0]], LOGFILE=[$LOGFILE]\n";
			$LOGFILE = shift unless (defined($LOGFILE));
			#print "LOGFILE = [$LOGFILE]\n";
		} elsif ($word eq 'LOG_STDOUT') {
			$LOG_METHOD = 'stdout';
		} else {
			push(@args, $word);
		}
	}
	AbsenceLog->export_to_level(1, @args);

	if (!defined($LOGFILE)) {
		if (exists($ENV{ABSENCE_TEST})) {
			$LOGFILE	= '/tmp/absence.log';
		} else {
			#print "getting logfile path from AbsenceConfig...\n";
			$LOGFILE	= AbsenceConfig::fetch('log_file') ||
				die "no logfile configured.\n";
		}
	}
}

sub setAuthUser
{
	$AUTH_USER = ' '.shift;
}

sub abslog
{
	my $do_ts = 1;
	if ($_[0] eq '-nots') {
		$do_ts = 0;
		shift;
	}
		
	my $msg = shift;

	my $ts = timestamp();

	my $openstr = "+<$LOGFILE";

	if (! -e $LOGFILE) {
		$openstr = ">$LOGFILE";
	}

	if ($LOG_METHOD eq 'stdout') {
		print "LOG: $msg\n";
		return;
	}

	my $fh = FileHandle->new($openstr);
	defined($fh) || die "open of [$LOGFILE] failed";
	flock($fh, LOCK_EX) || die "flock lock_ex on $LOGFILE ($!)";
	# position file-pointer at EOF
	seek($fh, 0, SEEK_END);
	if (ref($msg)) {
		print $fh "$ts $IPADDR$AUTH_USER\n";
		foreach (@$msg) {
			print $fh "\t$_\n";
		}
	} else {
		if ($msg =~ /^debug/i) {
			print $fh "$msg\n";
		} elsif (!$do_ts) {
			print $fh "$msg\n";
		} else {
			print $fh "$ts $IPADDR$AUTH_USER $msg\n";
		}
	}
	flock($fh, LOCK_UN) || die "flock lock_un on $LOGFILE ($!)";
	$fh->close;
}

sub timestamp
{
	my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
	my ($sec,$min,$hour,$mday,$mon,$year,@junk) = localtime(time);

	return sprintf("%d-%s-%d %02d:%02d:%02d",
		$mday, $months[$mon], $year+1900, $hour, $min, $sec);
}

1;

__END__

=head1 NAME

AbsenceLog.pm - primitive logging facility for absence

=head1 SYNOPSIS

  use AbsenceLog;

  abslog('text to be logged, not CR-terminated');

=head1 DESCRIPTION

AbsenceLog.pm provides a simple logging facility.  Messages are
written to the logfile specified in by 'log_file' in the configuration.
There is no provision for not having a logfile, i.e., a C<use AbsenceLog>
will fail if 'log_file' is not configured.  The log file must also exist.

=head1 FUNCTIONS

=over 4

=item abslog()

Takes a single parameter consisting of text to be logged.  The text
should not be carriage-return terminated (that is done automatically).
B<abslog> is exported by default for convenience.

=back

=head1 AUTHOR

Robert Urban <urban@tru64.org>

Copyright (C) Robert Urban 2003

=cut
