#!/usr/bin/perl

# $Id: absence-click.pl 89 2009-07-14 00:12:25Z urban $

#==========================================================================
# Welcome to "absence-click.pl"!
#
# It's a bit spaghetti-ish. Sorry about that.
# This script is a bit complicated because it is called at two
# stages.  First, it is called when a user clicks upon a month-image.
# In this case, the tests for the existence of the show_absence and
# create_absence CGI parameters fail, and processing starts at the
# point marked "First entry-point".
# The click-information is analyzed and then one of
#   badClick()
#   showAbsence()
#   showHoliday()
#   newAbsence()
# is called (depending on whether client-side or server-side image-maps
# have been configured).
#
# These functions display a form whose target is this script.
# When the users clicks on the SUBMIT button in one of the forms,
# the appropriate CGI parameters are set and this script is re-entered.
#
# The script performs the requested operation (adding a new reservation,
# for example), and either causes a browswer redirect to the main
# page (absence.pl) on success, or causes an error message to be
# displayed.
#
# Everything is slightly complicated by the necessity of have the
# "INSTANCE_NAME" environmental-variable set (if applicable) before
# the AbsenceConfig module is loaded. The instance-information must
# be passed to this script, and the method used to pass it is
# different for stages 1 and 2.  In stage 1, the instance-name
# is passed as part of the "extended path information" contained
# in the URL.  In stage 2, the instance-name is passed as a
# CGI parameter.
#==========================================================================

use CGI;
use Carp;

# uncomment the following if you are debugging
#use CGI::Carp qw(fatalsToBrowser);

my $q;
BEGIN {
	$q = new CGI;
	$q->import_names;
	$q->charset('utf-8');

	#system("echo 'PATH_INFO=$ENV{PATH_INFO}' > /tmp/foo.xx");
	if (exists($ENV{PATH_INFO}) && ($ENV{PATH_INFO} =~ m!/in=([^/?]+)!)) {
		$ENV{INSTANCE_NAME} = $1;
		#system("echo 'instance=$1' >> /tmp/foo.xx");
	} elsif (defined($Q::instance)) {
		$ENV{INSTANCE_NAME} = $Q::instance;
	}
}

use AbsenceConfig;
use AbsenceImage;
use AbsenceDB;
use AbsenceLog;
use AbsenceAuthentication;
use AbsenceAuthorization;

my %MONTH_LABELS = (
	1	=> 'January',
	2	=> 'February',
	3	=> 'March',
	4	=> 'April',
	5	=> 'May',
	6	=> 'June',
	7	=> 'July',
	8	=> 'August',
	9	=> 'September',
	10	=> 'October',
	11	=> 'November',
	12	=> 'December',
);
my $DEBUG = 0;
$SIG{__DIE__} = \&handleError;

# various paths to scripts and pages
my $MAIN_SCRIPT		= AbsenceConfig::fetch('main_script');
my $COLOR_NOPRIV	= AbsenceConfig::fetch('wp_nopriv');
my $COLOR_MANAGE	= AbsenceConfig::fetch('wp_manage');
my $COLOR_AFLABEL	= AbsenceConfig::fetch('wp_aflabel');
my $COLOR_AFBG		= AbsenceConfig::fetch('wp_afbg');

#my $TOP_PAGE		= AbsenceConfig::fetch('top_page');

my $VERSION		= '1.7';

$CGI::POST_MAX = 300;
$CGI::DISABLE_UPLOADS = 1;

if (exists($ENV{INSTANCE_NAME})) {
	$MAIN_SCRIPT .= "?instance=$ENV{INSTANCE_NAME}";
}

#-----------------------------------------------------------------
# first, authenticate if necessary
#-----------------------------------------------------------------
my $AUTH_UID = AbsenceAuthentication::checkAuthentication($q);


if (defined($Q::show_absence)) {
	if ($Q::show_absence =~ /^save/i) {
		modifyAbsence($q);
		exit;
	} elsif ($Q::show_absence eq 'Delete') {
		deleteAbsence($q);
		exit;
	} elsif ($Q::show_absence eq 'Cancel') {
		print $q->redirect($MAIN_SCRIPT);
		exit(0);
	} else {
		print "whoops!\n";
	}
}

if (defined($Q::create_absence)) {
	createAbsence($q);
	exit(0);
}

#-------------------------------------------------------------------
# First entry-point.
#
# if I got here, it means the script has been called after a user
# clicked on one of the month-images
#-------------------------------------------------------------------
print $q->header();

my @keywords = $q->keywords;
my $stuff = $keywords[0];
my ($key, @result);

#-------------------------------------------------------------------
# The encoding of absence information depends on whether map_type
# is set to 'client' or 'server'
#-------------------------------------------------------------------
if (AbsenceConfig::fetch('map_type') eq 'server') {
	if ($ENV{'PATH_INFO'} !~ m!^(/in=[^/]+)?/(\d+)-(\d+)-(\d+)$!) {
		die "failed to match [$ENV{'PATH_INFO'}]";
	}
	my $gid = $2;
	my $month = $3;
	my $year = $4;
	my ($x, $y);
	if ($stuff =~ m!^(\d+),(\d+)$!) {
		$x = $1;
		$y = $2;
		#print "month [$month], year [$year], x [$x], y [$y]<BR>\n";
	} else {
		print "map_type = server, unable to parse [$stuff]<BR>\n";
	}
	@result = AbsenceImage::findAbsence($month, $year, $gid, $x, $y);
	$key = $result[0];
} else {
	if ($stuff =~ /res,(\d+)$/) {
		@result = (absence => $1);
		$key = 'absence';
	} elsif ($stuff =~ /new,(\d+),(\d+),(\d+),(\d+),(\d+)$/) {
		($pid,$gid,$day,$month,$year) = ($1, $2, $3, $4, $5);
		$DEBUG && abslog("pid=$pid, gid=$gid, day=$day, mon=$month, year=$year");
		@result = (start => [$pid, $day, $month, $year]);
		$key = 'start';
	} else {
		print "map_type = client, unable to parse [$stuff]\n";
	}
}

if ($key eq 'none') {
	badClick($q, $result[1]);
} elsif ($key eq 'absence') {
	showAbsence($q, \@result);
} elsif ($key eq 'holiday') {
	showHoliday($q, \@result);
} elsif ($key eq 'start') {
	newAbsence($q, \@result);
} else {
	print "findAbsence error. returned: [@result]<BR>\n";
}

#print "<P>ENV:<BR>\n";
#foreach my $key (keys(%ENV)) {
#	print "[$key] $ENV{$key}<BR>\n";
#}

if ($DEBUG) {
	my @names = $q->param;

	print "Parameters:<BR>\n";
	foreach my $p (@names) {
		print " [$p]<BR>\n";
	}

	print "<P>Keywords:<BR>\n";
	foreach my $p (@keywords) {
		print " [$p]<BR>\n";
	}
}

print $q->end_html();

exit;

#====================================================================
# subs
#====================================================================

sub handleError
{
	my $msg = shift;
	confess($msg);
}

#------------------------------------------------------------------
# deleteAbsence()
#
# user has clicked on "Delete" in the show/modify/delete form
#------------------------------------------------------------------

sub deleteAbsence
{
	my $q = shift;
	print "<H1>Delete an Absence:</H1><BR>\n";

	my $msg;

	#abslog("res_id=[$Q::res_id]");
	my $res = AbsenceDB::getReservation($Q::res_id);
	if (!defined($res)) {
		print $q->header();
		print $q->start_html(
			-title		=> 'Absence: no-action',
			-BGCOLOR	=> $COLOR_AFBG,
		);
		print "Absence was deleted by someone else<BR>\n";
		reloadMain($q);
		exit;
	}
	
	my $pid = AbsenceDB::getReservation($Q::res_id)->{object_id};
	if (!AbsenceAuthorization::allowWriteForPerson($AUTH_UID, $pid)) {
		print $q->header();
		print $q->start_html(
			-title		=> 'Absence: delete an absence',
			-target		=> 'display',
			-BGCOLOR	=> $COLOR_NOPRIV,
		);
		print "<H1>no authorization</H1><BR>\n";
		print $q->end_html();
		exit 0;
	}

	my $ret = AbsenceDB::deleteReservation($Q::res_id);

	if ($ret eq 'ok') {
		print $q->redirect($MAIN_SCRIPT);
		#print "Successfully completed.<BR>\n";
		exit(0);
	} elsif ($ret eq 'conflict') {
		$msg = "Looks like someone else got there first.<BR>\n";
	} elsif ($ret eq 'internalerror') {
		$msg = "internal error. (sorry)<BR>\n";
	}

	print $q->header();
	print $q->start_html(
		-title => 'Absence: delete absence',
		-target => 'display',
		-BGCOLOR	=> $COLOR_AFBG,
	);
	print $msg;
	print $q->start_form(-action => $MAIN_SCRIPT, -method => 'POST');
	print $q->submit(action => 'Jump back to Calendar');
	print $q->end_form();
	print $q->end_html();
}

#------------------------------------------------------------------
# modifyAbsence()
#
# user has clicked on "Save Changes" in the show/modify/delete form
#------------------------------------------------------------------

sub modifyAbsence
{
	my $q = shift;

	#my $start = "$Q::sday.$Q::smonth.$Q::syear";
	#my $end = "$Q::eday.$Q::emonth.$Q::eyear";

	my $start = { day => $Q::sday, month => $Q::smonth, year => $Q::syear };
	my $end = { day => $Q::eday, month => $Q::emonth, year => $Q::eyear };

	my $msg;

	if (!AbsenceAuthorization::allowWriteForPerson($AUTH_UID, $Q::person_id)) {
		print $q->header();
		print $q->start_html(
			-title		=> 'Absence: modify an absence',
			-target		=> 'display',
			-bgcolor	=> $COLOR_NOPRIV,
		);
		print "<H1>no authorization</H1><BR>\n";
		print $q->end_html();
		exit 0;
	}

	my ($ret, $addit) = AbsenceDB::addReservation($Q::res_id, $Q::person_id,
		$start, $end, $Q::type_id, $Q::desc);

	#print "AbsenceDB::addReservation returned [$ret]<BR>\n";

	if ($ret eq 'ok') {
		print $q->redirect($MAIN_SCRIPT);
		#print "<P>Absence successfully modified.<BR>\n";
		exit(0);
	} elsif ($ret eq 'disappeared') {
		$msg = "<P>ERROR: The requested absence has been deleted.\n";
	} elsif ($ret eq 'impossible') {
		$msg = "<P>ERROR: The requested absence ends before it starts!\n";
	} elsif ($ret eq 'conflict') {
		$msg = conflictMessage($addit, $Q::res_id);
	}

	print $q->header();
	print $q->start_html(
		-title => 'Absence: modify an existing absence',
		-target => 'display',
		-BGCOLOR	=> $COLOR_AFBG,
	);

	print "<H1>Modify an Absence:</H1><BR>\n";

	print $msg;

	print $q->start_form(-action => $MAIN_SCRIPT, -method => 'POST');
	print $q->submit(action => 'Jump back to Calendar');
	print $q->end_form();
	print $q->end_html();
}

sub conflictMessage
{
	my ($addit, $res_id) = @_;

	my $msg;

	if ($addit->{type} eq 'simple') {
		my $res = $addit->{res};
		my $type = AbsenceDB::getType($res->{type_id});
		$msg = "<P>ERROR: The requested absence would overlap "
			."with an existing one:<BR>\n"
			."Start: $res->{start}<BR>"
			."End: $res->{end}<BR>"
			."Type: $type->{name}<BR>"
			."Description: $res->{description}<BR>\n"
			."Please select a new start/end date.<BR>\n";
	} elsif ($addit->{type} eq 'no_space') {
		my $m = AbsenceDate::monthName($addit->{month});
		$msg = qq{<P>ERROR: Not enough space is available in $m, $addit->{year} for the requested reservation.};
	} else {
		my $id = ($addit->{ids}->[0] eq $res_id)
			? $addit->{ids}->[1]
			: $addit->{ids}->[0];
		my $res = AbsenceDB::getReservation($id);
		my $type = AbsenceDB::getType($res->{type_id});
		$msg = qq/The requested reservation would coincide illegally with the following reservation:<BR>Type: $type->{name}<BR>Description: $res->{description}<BR>Start: $res->{start}<BR>End: $res->{finish}<BR>/;
	}

	return $msg;
}

#------------------------------------------------------------------
# newAbsence()
#
# display a form to allow the create of a new absence
#------------------------------------------------------------------

sub createAbsence
{
	my $q = shift;

	if ($Q::create_absence eq 'Cancel') {
		print $q->redirect($MAIN_SCRIPT);
		exit(0);
	}

	if (!AbsenceAuthorization::allowWriteForPerson($AUTH_UID, $Q::person_id)) {
		print $q->header();
		print $q->start_html(
			-title		=> 'Absence: create a new absence',
			-target		=> 'display',
			-BGCOLOR	=> $COLOR_NOPRIV,
		);
		print "<H1>no authorization</H1><BR>\n";
		print $q->end_html();
		exit 0;

		#print $q->start_form(-action => $MAIN_SCRIPT, -method => 'POST');
		#print $q->submit(action => 'Jump back to Calendar');
		#print $q->end_form();
	}

	my $start = { day => $Q::sday, month => $Q::smonth, year => $Q::syear };
	my $end = { day => $Q::eday, month => $Q::emonth, year => $Q::eyear };

	my $msg;

	my ($ret, $addit) = AbsenceDB::addReservation('new', $Q::person_id,
		$start, $end, $Q::type_id, $Q::desc);

	print "AbsenceDB::addReservation returned [$ret]<BR>\n";
	if ($ret eq 'ok') {
		print $q->redirect($MAIN_SCRIPT);
		#print "<P>Absence successfully created.<BR>\n";
		exit(0);
	} elsif ($ret eq 'impossible') {
		$msg = "<P>ERROR: The requested absence ends before it starts!\n";
	} elsif ($ret eq 'conflict') {
		$msg = conflictMessage($addit, 'new');
	} elsif ($ret eq 'badday') {
		$msg = "<P>ERROR: start or end date invalid (past end of month)\n";
	}

	print $q->header();
	print $q->start_html(
		-title		=> 'Absence: create a new absence',
		-target		=> 'display',
		-BGCOLOR	=> $COLOR_AFBG,
	);
	print "<H1>Create Absence Error</H1><BR>\n";

	print $msg;

	print $q->start_form(-action => $MAIN_SCRIPT, -method => 'POST');
	print $q->submit(action => 'Jump back to Calendar');
	print $q->end_form();
	print $q->end_html();
}

sub newAbsence
{
	my ($q, $rref) = @_;

	my $sref = $rref->[1];

	my $pid = $sref->[0];
	if (!AbsenceAuthorization::allowWriteForPerson($AUTH_UID, $pid)) {
		print $q->start_html(
			-title		=> 'Absence: error: unauthorized',
			-bgcolor	=> $COLOR_NOPRIV,
		);
		my $name = AbsenceDB::getPerson($pid, 'name');
		print "<H3>You are not authorized to create an absence for [$name]</H3>\n";
		print $q->end_html();
		exit;
	}

	print $q->start_html(
		-title		=> 'Absence: create a new absence',
		-bgcolor	=> $COLOR_AFBG,
	);

	print $q->h1("Create a new absence");

	absenceForm(new => $sref);

	print $q->submit(create_absence => 'Save Changes'),
		$q->submit(create_absence => 'Cancel');

	print $q->end_form;
}

sub showAbsence
{
	my ($q, $rref) = @_;

	print $q->start_html(
		-title		=> 'Absence: display/modify an absence',
		-BGCOLOR	=> $COLOR_AFBG,
	);

	print $q->h1("Display or Modify an absence");

	my $res_id = $rref->[1];
	my $res = AbsenceDB::getReservation($res_id);
	if (!defined($res)) {
		#print "error: reservation id [$res_id] not found.<BR>\n";
		print "Absence was deleted by someone else<BR>\n";
		reloadMain($q);
		exit;

	} else {
		absenceForm(show => $res);
		my $pid = $res->{object_id};
		if (AbsenceAuthorization::allowWriteForPerson($AUTH_UID, $pid)) {
			print $q->submit(show_absence => 'Save Changes'),
				$q->submit(show_absence => 'Cancel'),
				$q->submit(show_absence => 'Delete');
		} else {
			print $q->submit(show_absence => 'Cancel');
		}
		print $q->end_form;
	}
}

sub reloadMain
{
	my $q = shift;

	print $q->h1("Click on button below to reload calendar");
	print $q->start_form(
		#-action => $MAIN_PAGE,
		-action => $MAIN_SCRIPT,
		-method => 'POST',
		#-target	=> '_parent',
	);
	print $q->submit(action => 'Jump back to Calendar');
	print $q->end_form();
	print $q->end_html();
}

sub absenceForm
{
	my ($type, $ref) = @_;

	if ($DEBUG) {
		abslog(["absenceForm: res-id=[$ref->{id}], oid=[$ref->{object_id}]",
			"start: $ref->{start}",
			"finish: $ref->{finish}",
			"type-id=[$ref->{type_id}]",
			"desc=[$ref->{description}]",
		]);
	}

	my ($pid);

	# handle default values
	my %def;

	if ($type eq 'show') {
		$pid = $ref->{object_id};
		my $s = AbsenceDB::convToAbsenceDate($ref->{start});
		($def{sday}, $def{smonth}, $def{syear}) =
			($s->{day}, $s->{month}, $s->{year});
			#AbsenceDB::convToAbsenceDate($ref->{start});

		my $e = AbsenceDB::convToAbsenceDate($ref->{finish});
		($def{eday}, $def{emonth}, $def{eyear}) =
			($e->{day}, $e->{month}, $e->{year});
			#AbsenceDB::convToAbsenceDate($ref->{finish});

		$def{type_id}	= $ref->{type_id};
		$def{desc}		= $ref->{description};
		$def{res_id}	= $ref->{id};
	} else {
		my $def_type = AbsenceDB::getDefaultType();
		my $def_type_id;
		if (defined($def_type)) {
			$def_type_id = $def_type->{id};
		}

		$pid = $ref->[0];
		$def{sday}		= $ref->[1];
		$def{smonth}	= $ref->[2];
		$def{syear}		= $ref->[3];
		$def{eday}		= $ref->[1];
		$def{emonth}	= $ref->[2];
		$def{eyear}		= $ref->[3];
		$def{type_id}	= $def_type_id;
		$def{desc}		= '';
		$def{res_id}	= '';
	}

	if ($DEBUG) {
		abslog("defaults:");
		foreach my $key (keys(%def)) {
			abslog("$key=[$def{$key}]");
		}
	}

	#my ($form_start_year, $form_end_year)
	my $max = ($def{syear} > $def{eyear}) ? $def{syear} : $def{eyear};
	my $year = (localtime)[5] + 1900;
	my $end_year = $year + 10;
	my $start_year = $year - 2;
	$end_year = ($max > $end_year) ? $max : $end_year;

	#my $bg = 'papayawhip';
	#my $bg = $COLOR_AFLABEL;

	my $person = AbsenceDB::getPerson($pid, 'name');

	#my $start = "$sref->[1]/$sref->[2]/$sref->[3]";
	#print "Add an absence for [$person] starting on $start.<BR>\n";

	print $q->start_form;
	print $q->hidden(-name => 'person_id', -default => $pid);
	print $q->hidden(-name => 'res_id', -default => $def{res_id});
	if (exists($ENV{INSTANCE_NAME})) {
		print $q->hidden(-name => 'instance', -default => $ENV{INSTANCE_NAME});
	}
	print "<TABLE>\n";
	#------------------------------------------------
	# person row
	#------------------------------------------------
	print "<TR><TH ALIGN=right BGCOLOR=$COLOR_AFLABEL>Person</TH><TD>$person</TD></TR>\n";
	#------------------------------------------------
	# begin row
	#------------------------------------------------
	print "<TR><TH ALIGN=right BGCOLOR=$COLOR_AFLABEL>Begin</TH><TD>";
	print $q->popup_menu(
		-name	=> 'sday',
		-values	=> [1..31],
		-default=> $def{sday},
	);
	print $q->popup_menu(
		-name	=> 'smonth',
		-values	=> [1..12],
		-default=> $def{smonth},
		-labels	=> \%MONTH_LABELS,
	);
	print $q->popup_menu(
		-name	=> 'syear',
		-values	=> [$start_year..$end_year],
		-default=> $def{syear},
	);
	print "</TD></TR>\n";
	#------------------------------------------------
	# ending day row
	#------------------------------------------------
	print "<TR><TH ALIGN=right BGCOLOR=$COLOR_AFLABEL>End</TH><TD>";
	print $q->popup_menu(
		-name	=> 'eday',
		-values	=> [1..31],
		-default=> $def{eday},
	);
	print $q->popup_menu(
		-name	=> 'emonth',
		-values	=> [1..12],
		-default=> $def{emonth},
		-labels	=> \%MONTH_LABELS,
	);
	print $q->popup_menu(
		-name	=> 'eyear',
		-values	=> [$start_year..$end_year],
		-default=> $def{eyear},
	);
	print "</TD></TR>\n";
	#------------------------------------------------
	# what row
	#------------------------------------------------
	my (%type_labels, @type_ids);
	my @types = AbsenceDB::getTypes();
	%type_labels = map { $_->{id} => $_->{name} } @types;
	@type_ids = map { $_->{id} } sort { lc($a->{name}) cmp lc($b->{name}) } @types;
	print "<TR><TH ALIGN=right BGCOLOR=$COLOR_AFLABEL>What</TH><TD>";
	print $q->popup_menu(
		-name	=> 'type_id',
		#-values	=> ['Vacation', 'Training', 'Travel', 'CV', 'Other'],
		#-values	=> [AbsenceTypes::getTypes()],
		-values		=> \@type_ids,
		-labels		=> \%type_labels,
		-default	=> $def{type_id},
	);
	print "</TD></TR>\n";
	#------------------------------------------------
	# long-desc row
	#------------------------------------------------
	print "<TR><TH ALIGN=right BGCOLOR=$COLOR_AFLABEL>Description</TH><TD>";
	print $q->textfield(
		-name		=> 'desc',
		-size		=> 50,
		-maxlength	=> 100,
		-default	=> $def{desc},
	);
	print "</TD></TR>\n";
	print "</TABLE>\n";
}

sub showHoliday
{
	my ($q, $rref) = @_;

	print $q->start_html(
		-title		=> 'Absence: Holiday',
		-BGCOLOR	=> $COLOR_AFBG,
	);

	my $href = $rref->[1];
	my ($d, $m, $y, $holiday) = @{$href};
	my $mon = AbsenceDate::monthName($m, '-short');
	my $wd = AbsenceDate::weekDayFromDate($d, $m, $y, '-name');
	print $q->h1("$wd, $d-$mon-$y, is $holiday");
}

sub badClick
{
	my ($q, $more) = @_;

	print $q->start_html(
		-title		=> 'Absence: no-action',
		-BGCOLOR	=> $COLOR_AFBG,
	);
	if ($more eq 'stop-that') {
		print "Don't click there again.<BR>\n";
		return;
	}
	print <<_EOF_;
<P>Instructions:<BR>
<UL>
	<LI>month-days coloured purple are holidays.  You can click
		on them to get the holiday.
	<LI>absences are blocks of red.  Clicking on them allows you to
		display, modify, and delete them.
	<LI>To create a new absence, click on the day in the person-row
		when the absence should start.
</UL>
_EOF_
}
