#!/usr/bin/perl

# $Id: absence-manage.pl 91 2009-07-14 15:58:06Z urban $

#=====================================================================
# Welcome to "absence-manage.pl"!
#
# absence-manage.pl is somewhat complicated because it is called at
# two stages (see also the comments at the top of absence-click.pl).
#
# Stage 1:
#
# user clicks on the "manage" URL in the control frame and is presented
# with a number of forms for managing people and groups.
#
# Stage 2:
#
# user clicks on one of the submit buttons belonging to one of the
# forms above and the script is re-entered with some CGI parameters
# set.
#=====================================================================

use CGI qw/:standard -debug/;
# uncomment the following if you are debugging
use CGI::Carp qw(fatalsToBrowser);
use FileHandle;
use Carp;

BEGIN {
	if ($ENV{QUERY_STRING} =~ /^instance=(\S+)$/) {
		# an instance has been specified
		$ENV{INSTANCE_NAME} = $1;
	}
}

use AbsenceDB;
use AbsenceConfig;
use AbsenceLog;
use AbsenceAuthentication;
use AbsenceAuthorization;
use AbsenceImage;
use AbsenceInput;
use Data::Dumper;

#--------------------------------------------------------------------
# global variables
#--------------------------------------------------------------------
my $IMAGE_DIR		= AbsenceConfig::fetch('image_dir_abs');
my $TOP_PAGE		= AbsenceConfig::fetch('top_page');
my $MAIN_SCRIPT		= AbsenceConfig::fetch('main_script');
my $GROUP_POLICY	= AbsenceConfig::fetch('group_policy');
my $OAP				= AbsenceConfig::fetch('objects_are_people');
my $AUTH			= AbsenceConfig::fetch('authentication');
#my $CRED_SRC		= AbsenceConfig::fetch('credential_src');
my $MANAGE_PASSWORD	= AbsenceConfig::fetch('manage_password');
my $COLOR_NOPRIV	= AbsenceConfig::fetch('wp_nopriv');
my $COLOR_MANAGE	= AbsenceConfig::fetch('wp_manage');
my $MULTI_RES		= AbsenceConfig::fetch('multi_res');
my $THING			= $OAP ? 'person' : 'object';
my $UCF_THING		= ucfirst($THING);
my $JS_COLOR_PATH	= jsColorScriptPath();

my $JS_COLOR		= {
	-src	=> $JS_COLOR_PATH,
	-type	=> "text/javascript",
};

#my $DEBUG_OUT	= '/tmp/manage.log';
my $VERSION		= '1.6';
my $DEBUG		= 0;

$CGI::POST_MAX = 2000;
$CGI::DISABLE_UPLOADS = 1;

#--------------------------------------------------------------------
# main()
#--------------------------------------------------------------------
$SIG{__DIE__} = \&handleErrors;

if (exists($ENV{INSTANCE_NAME})) {
	$MAIN_SCRIPT .= "?instance=$ENV{INSTANCE_NAME}";
}

my $q = new CGI;
$q->import_names;
$q->charset('utf-8');

#-----------------------------------------------------------------
# first, authenticate if necessary
#-----------------------------------------------------------------
my $AUTH_UID		= AbsenceAuthentication::checkAuthentication($q);
my $SUPER_USER		= AbsenceAuthorization::checkSuperuser($AUTH_UID);
my @ADMIN_GROUPS	= AbsenceAuthorization::getAdminGroups($AUTH_UID);

#-----------------------------------------------------------------
# check if $AUTH_UID is allowed to perform any admin tasks. If not,
# complain.
#-----------------------------------------------------------------
if ($AUTH
	&& (!$SUPER_USER && !@ADMIN_GROUPS)
	&& (defined($Q::action)
		|| defined($Q::action_add_object)
		|| defined($Q::action_delete_object)
		|| defined($Q::action_modify_object))
	&& ($Q::action ne 'Change Password'))
{
	print $q->header();
	print $q->start_html(
		-title		=> 'Absence: Manage People',
		-target		=> 'display',
		-BGCOLOR	=> $COLOR_NOPRIV,
	),
	"<H3>You do not have any admin priviledges.</H3>\n",
	$q->end_html();
	exit;
}


# DEBUG
if ($DEBUG) {
	my @keywords = $q->keywords;
	my @names = $q->param;
	dbg('Parameters:');
	foreach my $p (@names) {
		dbg(" [$p] ".$q->param($p));
	}

	dbg("Keywords:");
	foreach my $p (@keywords) {
		dbg(" [$p] ");
	}
}


if (defined($Q::action_add_object)) {
	addPerson($q);
} elsif (defined($Q::action_delete_object)) {
	deletePerson($q);
} elsif (defined($Q::action_modify_object)) {
	modifyPersonForm($q);
} elsif ($Q::action eq 'Add User') {
	addUser($q);
} elsif ($Q::action eq 'Modify User') {
	modifyUserForm($q);
} elsif ($Q::action eq 'Delete User') {
	deleteUser($q);
} elsif ($Q::action eq 'Modify Group') {
	modifyGroupForm($q);
} elsif ($Q::action eq 'Delete Group') {
	deleteGroup($q);
} elsif ($Q::action eq 'Add Group') {
	addGroup($q);
} elsif ($Q::action eq 'Submit Modification') {
	dbg("got [Submit Modification]");
	modifyPerson($q);
} elsif ($Q::action eq 'Submit Group Modification') {
	modifyGroup($q);
} elsif ($Q::action eq 'Submit User Modification') {
	modifyUser($q);
} elsif ($Q::action eq 'Change Password') {
	changePassword($q);
} elsif ($Q::action eq 'Add Type') {
	addType($q);
} elsif ($Q::action eq 'Modify Type') {
	modifyTypeForm($q);
} elsif ($Q::action eq 'Deactivate Type') {
	deactivateType($q);
} elsif ($Q::action eq 'Reactivate Type') {
	reactivateType($q);
} elsif ($Q::action eq 'Submit Type Modification') {
	modifyType($q);
}

print $q->header(-expires => '-1d');
print $q->start_html(
	-title		=> 'Absence: Manage People',
	-target		=> 'display',
	-BGCOLOR	=> $COLOR_MANAGE,
	-script		=> $JS_COLOR,
);

# include JavaScript for ACL-managment if necessary
if ($AUTH) {
	includeJsAclScript();
}

#print "<H2>AUTH = [$AUTH]</H2>\n";

($AUTH && $MANAGE_PASSWORD) && changePasswordForm();
if ($AUTH && !$SUPER_USER && !@ADMIN_GROUPS) {
	$q->end_html();
	exit;
}

print "<HR><P>\n";
print $q->h1({-align => 'center'},'Manage People, Groups &amp; Users');

print "<HR>\n<P>\n";

mainAddPersonForm($q);
mainModifyPersonForm($q);
mainDeletePersonForm($q);

my $hack = AbsenceConfig::fetch('partial_auth_hack');
my $show = ($AUTH && $SUPER_USER)
	|| (!$AUTH && (!$hack || ($hack && AbsenceAuthentication::isAuthed())));

if (!$AUTH || $SUPER_USER) {
	mainAddGroupForm($q);
	mainModifyGroupForm($q);
	mainDeleteGroupForm($q);
	if ($show) {
		mainAddTypeForm($q);
		mainModifyTypeForm($q);
		mainDeactivateTypeForm($q);
		mainReactivateTypeForm($q);
	}
}

if ($AUTH) {
	mainAddUserForm($q);
	mainModifyUserForm($q);
	mainDeleteUserForm($q);
}


print $q->start_form(-action => $MAIN_SCRIPT, -method => 'POST');
print $q->submit(action => 'Cancel');
print $q->end_form();
print $q->end_html();

exit;

sub handleErrors
{
	my $msg = shift;

	confess($msg);
}

#=================================================================
#=================================================================
# first level forms
#=================================================================
#=================================================================

sub changePasswordForm
{
	my $username = AbsenceDB::getUser(id => $AUTH_UID, 'username');
	print $q->h1("Change Password for $username");
	print $q->start_form(
		-name		=> 'change_password',
	);
	print "<TABLE>\n";
	print "<TR><TH ALIGN=right>New Password</TH><TD>\n";
	print $q->password_field(
		-name		=> 'password1',
		-size		=> 30,
		-maxlength	=> 30,
	);
	print "</TD></TR>\n";
	print "<TR><TH ALIGN=right>New Password (again)</TH><TD>\n";
	print $q->password_field(
		-name		=> 'password2',
		-size		=> 30,
		-maxlength	=> 30,
	);
	print "</TD></TR>\n</TABLE>\n";
	print $q->submit(action => 'Change Password');
	print $q->end_form();
}

sub changePassword
{
	my $q = shift;

	if (!length($Q::password1) || !length($Q::password2)) {
		dbg("looks bad");
		reloadDisplay($q, 'Error changing password',
			'one or both password fields blank');
	}

	if ($Q::password1 ne $Q::password2) {
		dbg("looks bad");
		reloadDisplay($q, 'Error changing password', 'password mismatch');
	}

	my $ret = AbsenceDB::setPassword($AUTH_UID, $Q::password1);
	dbg("ret = [$ret]");

	if ($ret eq 'ok') {
		dbg("looks good");
		print $q->redirect($MAIN_SCRIPT);
	} elsif ($ret eq 'disappeared') {
		dbg("looks bad");
		reloadDisplay($q, 'Error changing password',
			"user [$AUTH_UID] disappeared");
	} elsif ($ret =~ /^baddata:\s(.*)$/) {
		dbg("looks bad");
		reloadDisplay($q, 'Error changing password',
			$1);
	} elsif ($ret eq 'error') {
		dbg("looks bad");
		reloadDisplay($q, 'Error changing password', 'error changing password');
	}

	exit;
}

#---------------------------------------------------------------
# ADD PERSON
#---------------------------------------------------------------

sub mainAddPersonForm
{
	my $q = shift;

	print $q->h1("Add a $UCF_THING");
	print qq[<P>A "$UCF_THING" is an visible object which can have absences
		assigned to it.<P>];
	my @onsubmit = ($OAP && $AUTH)
		? (-onSubmit   => 'return validate(document.add_person);') 
		: ();
	print $q->start_form(
		-name		=> 'add_person',
		@onsubmit,
	);
	print "<TABLE>\n";

	#------------------------
	# Name
	#------------------------
	print "<TR><TH ALIGN=right>Name (Last, First)</TH><TD>\n";
	print $q->textfield(
		-name		=> 'name',
		-size		=> 20,
		-maxlength	=> 20,
	);
	print "</TD><TD>[This will appear in the month-images]</TD></TR>\n";

	if ($AUTH && $OAP) {
		#------------------------
		# User-ID
		#------------------------
		print "<TR><TH ALIGN=right>Username</TH><TD>\n";
		print $q->textfield(
			-name		=> 'username',
			-size		=> 20,
			-maxlength	=> 20,
		);
		print "</TD><TD>[This is the login-name]</TD></TR>\n";
		#------------------------
		# password
		#------------------------
		if ($MANAGE_PASSWORD) {
			print "<TR><TH ALIGN=right>Password</TH><TD>\n";
			print $q->textfield(
				-name		=> 'password',
				-size		=> 30,
				-maxlength	=> 30,
			);
			print "</TD></TR>\n";
		}
	}

	#------------------------
	# E-mail
	#------------------------
	print "<TR><TH ALIGN=right>E-Mail Address</TH><TD>\n";
	print $q->textfield(
		-name		=> 'email',
		-size		=> 30,
		-maxlength	=> 50,
	);

	#------------------------
	# groups
	#------------------------
	print "<TR><TH ALIGN=right VALIGN=top>Group</TH><TD>\n";
	($GROUP_POLICY eq 'single')
		? makeGroupPopup($q)
		: makeGroupCheckboxes($q, []);

	print "</TD>
		<TD VALIGN=top>
			[choose which groups this person should belong to]
		</TD></TR>\n";

	($AUTH && $OAP) && aclForm($q, 'add_person');

	print "</TABLE>\n";
	print $q->submit(action_add_object => 'Add Person');
	print $q->end_form();
	print "<HR>\n";
}

#---------------------------------------------------------------
# MODIFY PERSON
#---------------------------------------------------------------

sub mainModifyPersonForm
{
	my $q = shift;

	print $q->h1("Modify a $UCF_THING");
	print $q->start_form;
	print "$UCF_THING";
	makePeoplePopup($q);
	print "&nbsp;";
	print $q->submit(action_modify_object => "Modify $UCF_THING");
	print $q->end_form();
	print "<HR>\n";
}

#---------------------------------------------------------------
# DELETE PERSON
#---------------------------------------------------------------

sub mainDeletePersonForm
{
	my $q = shift;

	print $q->h1("Delete a $UCF_THING");
	print $q->start_form;
	print "$UCF_THING";
	makePeoplePopup($q, 'delete');

	print "&nbsp;";
	print $q->submit(action_delete_object => "Delete $UCF_THING");
	#print $q->submit(
	#	-name	=> 'action',
	#	-label	=> "Delete huhu",
	#	-action	=> "Delete $UCF_THING",
	#);
	print $q->end_form();

	print "<HR>\n";
}

#---------------------------------------------------------------
# ADD GROUP
#---------------------------------------------------------------

sub mainAddGroupForm
{
	my $q = shift;

	print $q->h1('Add a Group');
	print qq[<P>A "group" is a collection of "${THING}s".<P>];
	print $q->start_form;
	print "Name\n";
	print $q->textfield(
			-name		=> 'name',
			-size		=> 20,
			-maxlength	=> 20,
		), '<BR>',
		'Description',
		$q->textfield(
			-name		=> 'description',
			-size		=> 40,
			-maxlength	=> 80,
		);
	print "&nbsp;";

	print $q->submit(action => 'Add Group');
	print $q->end_form();
}

#---------------------------------------------------------------
# MODIFY GROUP
#---------------------------------------------------------------

sub mainModifyGroupForm
{
	my $q = shift;

	print "<HR>\n";

	print $q->h1('Modify a group');
	print $q->start_form(-name => 'ModifyGroup');
	print "Group";
	makeGroupPopup($q);
	print "&nbsp;";
	print $q->submit(action => 'Modify Group');
	print $q->end_form();

	print "<HR>\n";
}

#---------------------------------------------------------------
# DELETE GROUP
#---------------------------------------------------------------

sub mainDeleteGroupForm
{
	my $q = shift;

	print $q->h1('Delete a group');
	if ($GROUP_POLICY eq 'single') {
		print "(this will also delete <EM>all</EM> members of the group)<BR>\n";
	} else {
		print "(this will also delete <EM>all</EM> members of the group\n";
		print "who are in no other groups)<BR>\n";
	}
	print $q->start_form(-name => 'DeleteGroup');
	print "Group";
	makeGroupPopup($q);
	print "&nbsp;";
	print $q->submit(action => 'Delete Group');
	print $q->end_form();

	print "<HR>\n";
}

#---------------------------------------------------------------
# MANAGE TYPES
#---------------------------------------------------------------

# <input class="color" id="bg" onchange="updateBg()" value="0066FF" size="10" />

sub jsColorScriptPath
{
	my $base = AbsenceConfig::fetch('js_dir_rel');
	my $path = "$base/jscolor.js";
}

sub includeJsColorScript
{
	my $path = jsColorScriptPath();
	print qq{<script src="$path" type="text/javascript"></script>\n};
}

sub mainAddTypeForm
{
	my $q = shift;

	includeJsRtCoincidenceScript();
	print $q->h1('Add a Reservation Type');
	print $q->start_form(-name => 'add_type');
	if ($MULTI_RES) { print "<TABLE><TR><TD VALIGN=TOP>\n"; }
	print "<TABLE>\n";
	print "<TR><TH ALIGN=right>Type Name</TH><TD>\n",
		$q->textfield(
			-name		=> 'name',
			-size		=> 20,
			-maxlength	=> 20,
		),
		"</TD></TR><TR><TH ALIGN=right>Description</TH><TD>\n",
		$q->textfield(
			-name		=> 'description',
			-size		=> 40,
			-maxlength	=> 80,
		),
		"</TD></TR><TR><TH ALIGN=right>Color</TH><TD>\n",
		$q->textfield(
			-name		=> 'color',
			-size		=> 8,
			-maxlength	=> 8,
			-class		=> 'color',
		);
	if ($MULTI_RES) {
		print "</TD></TR><TR><TH ALIGN=right>Transparency</TH><TD>\n",
			$q->popup_menu(
				-name		=> 'transparency',
				-values		=> [0..100],
				-default	=> 0,
			);
	}
	print "</TD></TR><TR><TH ALIGN=right>Default</TH><TD>\n",
		$q->checkbox(
			-name		=> 'default_type',
			-checked	=> 0,
			-value		=> 'on',
			-label		=> '',
		),
		"</TD></TR><TR><TH ALIGN=right>Skip Non Workdays</TH><TD>\n",
		$q->checkbox(
			-name		=> 'skip_non_workdays',
			-checked	=> 0,
			-value		=> 'on',
			-label		=> '',
		);
		if ($MULTI_RES) {
			print "</TD></TR><TR><TH ALIGN=right>Height</TH><TD>\n",
				makeTypeHeightPopup($q, 'document.add_type'),
				"</TD></TR><TR><TH ALIGN=right>Priority</TH><TD>\n",
				$q->popup_menu(
					-name		=> 'priority',
					-values		=> [reverse(-10..-1,1..10)],
					-default	=> 10,
				),
				"</TD></TR></TABLE>\n",
				"</TD><TD>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</TD><TD>\n";
			makeCoincidenceCheckboxes($q, undef, undef, 'block');
		}
	print "</TD></TR><TABLE>\n";
	print "&nbsp;";

	print $q->submit(action => 'Add Type');
	print $q->end_form();
}

sub makeTypeHeightPopup
{
	my ($q, $form, $default) = @_;

	$DEBUG && abslog("makeTypeHeightPopup: form=[$form]");

	if (!defined($default)) {
		$default = 'full';
	}

	my @values = ('quarter', 'half', 'full', 'block');

	return $q->popup_menu(
		-name   	=> 'height',
		-values		=> \@values,
		-default	=> $default,
		-onChange	=> "change_height($form);",
	);
}

sub makeCoincidenceCheckboxes
{
	my ($q, $type_id, $defaults, $disable) = @_;

	my @types = AbsenceDB::getTypes();

	my ($block_labels, $non_block_labels, @block_values, @non_block_values);
	my (@default_block, @default_non_block);
	my $myself;
	foreach my $type (sort { $a->{name} cmp $b->{name} } @types) {
		if (defined($type_id) && ($type->{id} eq $type_id)) {
			$myself = $type;
			next;
		}
		my $height = defined($type->{height}) ? $type->{height} : 'full';
		my $def_ref;
		if ($height eq 'block') {
			push(@block_values, $type->{id});
			$block_labels->{ $type->{id} } = $type->{name};
			$def_ref = \@default_block;
		} else {
			push(@non_block_values, $type->{id});
			$non_block_labels->{ $type->{id} } = ($height eq 'full')
				? $type->{name}
				: "$type->{name} ($height)";
			$def_ref = \@default_non_block;
		}
		if (inListN($type->{id}, $defaults)) {
			push(@{ $def_ref }, $type->{id});
		}
	}
	if ($DEBUG) {
		abslog("default_block=[".join(',', @default_block)."]");
		abslog("default_non_block=[".join(',', @default_non_block)."]");
	}

	print "<B>Reservation-Type can NOT be used on same day<BR>with the following types:</B><BR>\n";
	my $checked = 0;
	if (defined($type_id)) {
		$checked = AbsenceDB::checkVorbidTypeCoincidence($type_id, $type_id);
	}
	print "<TABLE RULES=all CELLSPACING=4 CELLPADDING=4 BORDER=1><TR><TD COLSPAN=2>\n";
	print $q->checkbox(
			-name		=> 'self_coincidence',
			-checked	=> $checked,
			-value		=> 'on',
			-label		=> defined($type_id)
				? "$myself->{name} (self)"
				: 'Self',
		), "<BR></TD></TR><TR><TD VALIGN=TOP>\n";

	if (@non_block_values) {
		my @nb_disable;
		if ($disable ne 'block') {
			@nb_disable = (-disabled => \@non_block_values);
		}
		print "<B>non-block reservation types</B><BR>\n",
			$q->checkbox_group(
				-name		=> 'rt_non_block',
				-values		=> \@non_block_values,
				-default	=> \@default_non_block,
				-labels		=> $non_block_labels,
				-linebreak	=> 'true',
				@nb_disable,
			);
	}
	print "</TD><TD VALIGN=TOP>\n";
	if (@block_values) {
		my @b_disable;
		if ($disable eq 'block') {
			@b_disable = (-disabled => \@block_values);
		}
		print "<B>block reservation types</B><BR>\n",
			$q->checkbox_group(
				-name		=> 'rt_block',
				-values		=> \@block_values,
				-default	=> \@default_non_block,
				-labels		=> $block_labels,
				-linebreak	=> 'true',
				@b_disable,
			), "<BR>\n";
	}
	print "</TD></TR></TABLE>\n";
}

#---------------------------------------------------------------
# MODIFY TYPE
#---------------------------------------------------------------

sub mainModifyTypeForm
{
	my $q = shift;

	print "<HR>\n";

	print $q->h1('Modify a Reservation Type');
	print $q->start_form(-name => 'ModifyType');
	print "Type";
	makeTypePopup($q);
	print "&nbsp;";
	print $q->submit(action => 'Modify Type');
	print $q->end_form();

	print "<HR>\n";
}

#---------------------------------------------------------------
# DEACTIVATE TYPE
#---------------------------------------------------------------

sub mainDeactivateTypeForm
{
	my $q = shift;

	print $q->h1('Deactivate a Reservation Type');
	print $q->start_form(-name => 'DeactivateType');
	print "Type";
	makeTypePopup($q);
	print "&nbsp;";
	print $q->submit(action => 'Deactivate Type');
	print $q->end_form();

	print "<HR>\n";
}

sub mainReactivateTypeForm
{
	my $q = shift;

	print $q->h1('Reactivate a Reservation Type');
	print $q->start_form(-name => 'ReactivateType');
	print "Type";
	makeTypePopup($q, '-reactivate');
	print "&nbsp;";
	print $q->submit(action => 'Reactivate Type');
	print $q->end_form();

	print "<HR>\n";
}

sub makeTypePopup
{
	my ($q, $arg) = @_;

	my @types;
	if ($arg eq '-reactivate') {
		@types = AbsenceDB::getTypes('invalid');
	} else {
		@types = AbsenceDB::getTypes();
	}

	my $labref;
	my @ids;
	foreach my $type_ref (@types) {
		$labref->{ $type_ref->{id} } = $type_ref->{name};
		push(@ids, $type_ref->{id});
	}

	my @sorted = sort {lc($labref->{ $a }) cmp lc($labref->{ $b })} @ids;
	print $q->popup_menu(
		-name   	=> 'type_id',
		-values		=> \@sorted,
		-labels		=> $labref,
	);
}

sub formColorToDbColor
{
	my $form_color = shift;

		my ($r, $g, $b) = ($form_color =~ /^(..)(..)(..)$/);
		$r = hex($r);
		$g = hex($g);
		$b = hex($b);

		return ($r, $g, $b);
}

sub dbColorToFormColor
{
	my $tref = shift;

	return sprintf("%02x%02x%02x",
		$tref->{color_red},
		$tref->{color_green},
		$tref->{color_blue},
	);
}

sub addType
{
	my $q = shift;

	dbg("- add type -");

	if (!$SUPER_USER) {
		abslog("attempt to circumvent security (addType)");
		reloadDisplay($q, 'Error adding type', 'security breach thwarted');
	}

	if (length($Q::name) == 0) {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding type', 'You must supply a name.');
	}

	if (defined(AbsenceDB::queryTypeName($Q::name))) {
		dbg("looks bad");
		reloadDisplay(
			$q,
			'Error modifying type',
			"name [$Q::name] already in use by an active type",
		);
	}

	if (defined(AbsenceDB::queryTypeName($Q::name, 'invalid'))) {
		dbg("looks bad");
		reloadDisplay(
			$q,
			'Error modifying type',
			"name [$Q::name] already in use by an inactive type",
		);
	}

	my $ref = { name => $Q::name };

	if (defined($Q::description)) {
		$ref->{description} = $Q::description;
	}

	if (length($Q::color) == 6) {
		($ref->{color_red}, $ref->{color_green}, $ref->{color_blue})
			= formColorToDbColor($Q::color);
	} else {
		$ref->{color_red}	= 0;
		$ref->{color_green}	= 0;
		$ref->{color_blue}	= 0;
	}

	$ref->{default_type} = ($Q::default_type eq 'on') ? 'TRUE' : 'FALSE';

	$ref->{skip_non_workdays}
		= ($Q::skip_non_workdays eq 'on') ? 'TRUE' : 'FALSE';

	if (defined($Q::height)) {
		$ref->{height} = $Q::height;
	}

	if (defined($Q::priority)) {
		$ref->{priority} = $Q::priority;
	}

	$ref->{transparency} = $Q::transparency;

	my $new_id = AbsenceDB::addType($ref);
	abslog("addType: \n".Dumper($ref));
	dbg("looks good");
	AbsenceDB::setTypeModTime();

	# now worry about coincidence rules
	if ($MULTI_RES) {
		my @no_coincidence = ($Q::height eq 'block')
			? @Q::rt_block
			: @Q::rt_non_block;
		($Q::self_coincidence eq 'on') && push(@no_coincidence, $new_id);
		foreach my $id (@no_coincidence) {
			AbsenceDB::vorbidTypeCoincidence($new_id, $id);
		}
	}

	reloadTop($q, 'Type Successfully Added', '_top',
		'Finished. Click on button below to reload top page');
	exit;
}

sub modifyTypeForm
{
	my $q = shift;

	my $tref = AbsenceDB::getType($Q::type_id);
	#abslog("dump of tref:\n".Dumper($tref)."\n--done--");

	if (!defined($tref)) {
		reloadTop($q, "Modify Type $Q::type_id", 'display',
			"Modify Type $Q::type_id", 'display',
			'Type disappeared');
	}

	my $color = dbColorToFormColor($tref);

	print $q->header(-expires => '-1d');
	print $q->start_html(
		-title		=> "Absence: Modify Type $tref->{name}",
		-target		=> 'display',
		-BGCOLOR	=> $COLOR_MANAGE,
		-script		=> $JS_COLOR,
	);

	includeJsRtCoincidenceScript();
	print "color=[$color]<BR>\n";
	print $q->h1('Modify a Reservation Type');
	print $q->start_form(-name => 'modify_type');
	print $q->hidden(-name => 'type_id', -default => $Q::type_id);
	if ($MULTI_RES) { print "<TABLE><TR><TD VALIGN=TOP>\n"; }
	print "<TABLE>\n";
	print "<TR><TH ALIGN=right>Type Name</TH>\n<TD>\n";
	print $q->textfield(
		-name		=> 'name',
		-size		=> 20,
		-maxlength	=> 20,
		-value		=> $tref->{name},
	);
	print "</TD></TR>\n";
	print "<TR><TH ALIGN=right>Description</TH>\n<TD>\n";
	print $q->textfield(
		-name		=> 'description',
		-size		=> 40,
		-maxlength	=> 60,
		-value		=> $tref->{description},
	);
	print "</TD></TR>\n";
	print "<TR><TH ALIGN=right>Color</TH>\n<TD>\n",
		$q->textfield(
			-name		=> 'color',
			-size		=> 8,
			-maxlength	=> 8,
			-class		=> 'color',
			-value		=> $color,
		);

	if ($MULTI_RES) {
		my $trans = defined($tref->{transparency})
			? $tref->{transparency}
			: 0;
		print "</TD></TR><TR><TH ALIGN=right>Transparency</TH><TD>\n",
			$q->popup_menu(
				-name		=> 'transparency',
				-values		=> [0..100],
				-default	=> $trans,
			);
	}

	print "</TD></TR>\n",
		"<TR><TH ALIGN=right>Default Type</TH>\n<TD>\n",
		$q->checkbox(
			-name		=> 'default_type',
			-checked	=> $tref->{default_type},
			-value		=> 'on',
			-label		=> 'Default Type',
		),
		"</TD></TR>\n",
		"<TR><TH ALIGN=right>Skip Non Workdays</TH>\n<TD>\n",
		$q->checkbox(
			-name		=> 'skip_non_workdays',
			-checked	=> $tref->{skip_non_workdays},
			-value		=> 'on',
			-label		=> 'Skip Non Workdays',
		);

	if ($MULTI_RES) {
		my $prio = defined($tref->{priority}) ? $tref->{priority} : 1;
		print "\n</TD></TR>\n<TR><TH ALIGN=right>Height</TH>\n<TD>\n",
			makeTypeHeightPopup($q, 'document.modify_type', $tref->{height}),
			"\n</TD></TR>\n<TR><TH ALIGN=right>Priority</TH>\n<TD>\n",
			$q->popup_menu(
				-name		=> 'priority',
				-values		=> [reverse(-10..-1,1..10)],
				-default	=> $prio,
			);
		print "</TD></TR>\n";
		print "</TD></TR></TABLE>\n";
		print "</TD><TD>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</TD>\n<TD>\n";
		my @no_coincidence = AbsenceDB::getTypeCoincidence($tref->{id});
		$DEBUG && abslog("no_coincidence=[".join('/', @no_coincidence)."]");
		my $disable_def = $tref->{height} eq 'block' ? 'full' : 'block';
		makeCoincidenceCheckboxes(
			$q,
			$Q::type_id,
			\@no_coincidence,
			$disable_def
		);
	}

	print "</TD></TR><TABLE>\n";

	print $q->submit(action => 'Submit Type Modification');
	print $q->end_form();
	exit;
}

sub modifyType
{
	my $q = shift;

	if (!$SUPER_USER) {
		abslog("attempt to circumvent security (modifyType)");
		reloadDisplay($q, 'Error modifying type', 'security breach thwarted');
	}

	dbg("- modify type -");
	if (length($Q::name) == 0) {
		dbg("looks bad");
		reloadDisplay($q, 'Error modifying type', 'You must supply a name.');
	}

	if ($Q::color !~ /^[0-9a-f]+$/i) {
		dbg("looks bad");
		reloadDisplay($q, 'Error modifying type', 'color-value invalid.');
	}

	if (defined(AbsenceDB::queryTypeName($Q::name, 'valid', $Q::type_id))) {
		dbg("looks bad");
		reloadDisplay(
			$q,
			'Error modifying type',
			"name [$Q::name] already in use by an active type",
		);
	}

	if (defined(AbsenceDB::queryTypeName($Q::name, 'invalid', $Q::type_id))) {
		dbg("looks bad");
		reloadDisplay(
			$q,
			'Error modifying type',
			"name [$Q::name] already in use by an inactive type",
		);
	}

	my $ref = { name => $Q::name, id => $Q::type_id };

	if (defined($Q::description)) {
		$ref->{description} = $Q::description;
	}

	if (length($Q::color) == 6) {
		($ref->{color_red}, $ref->{color_green}, $ref->{color_blue})
			= formColorToDbColor($Q::color);
	} else {
		$ref->{color_red}	= 0;
		$ref->{color_green}	= 0;
		$ref->{color_blue}	= 0;
	}

	$DEBUG && abslog("default_type=[$Q::default_type], skip=[$Q::skip_non_workdays]");

	$ref->{default_type}
		= ($Q::default_type eq 'on') ? 'TRUE' : 'FALSE';

	$ref->{skip_non_workdays}
		= ($Q::skip_non_workdays eq 'on') ? 'TRUE' : 'FALSE';

	if (defined($Q::height)) {
		$ref->{height} = $Q::height;
	}

	if (defined($Q::priority)) {
		$ref->{priority} = $Q::priority;
	}

	$ref->{transparency} = $Q::transparency;

	#dbg("--dump of ref about to be passed to modifyType:\n".Dumper($ref)."\n--end--\n");
	my $ret = AbsenceDB::modifyType($ref);
	abslog("Type Modified:\n".Dumper($ref));

	# now worry about coincidence rules
	if ($MULTI_RES) {
		AbsenceDB::purgeTypeCoincidenceEntries($Q::type_id);
		my @no_coincidence = ($Q::height eq 'block')
			? @Q::rt_block
			: @Q::rt_non_block;
		($Q::self_coincidence eq 'on') && push(@no_coincidence, $Q::type_id);
		foreach my $id (@no_coincidence) {
			AbsenceDB::vorbidTypeCoincidence($Q::type_id, $id);
		}
	}

	dbg("looks good");
	AbsenceDB::setTypeModTime();
	#AbsenceImage::removeLegendFile();
	#AbsenceImage::removeMonthImages();
	reloadTop($q, 'Type Successfully Modified', '_top',
		'Finished. Click on button below to reload top page');
	exit;
}

sub deactivateType
{
	my $q = shift;

	dbg("- deactivate type -");

	if (!$SUPER_USER) {
		abslog("attempt to circumvent security (deactivateType)");
		reloadDisplay($q, 'Error deactivating type', 'security breach thwarted');
	}

	my $ret = AbsenceDB::deactivateType($Q::type_id);
	abslog("Type Deactivated. ID=$Q::type_id");
	dbg("looks good");
	AbsenceDB::setTypeModTime();
	#AbsenceImage::removeLegendFile();
	#AbsenceImage::removeMonthImages();
	reloadTop($q, 'Type Successfully Deactivated', '_top',
		'Finished. Click on button below to reload top page');
	exit;
}

sub reactivateType
{
	my $q = shift;

	dbg("- reactivate type -");

	if (!$SUPER_USER) {
		abslog("attempt to circumvent security (reactivateType)");
		reloadDisplay($q, 'Error deactivating type', 'security breach thwarted');
	}

	# ROBBO
	my $res = validate_input([
		type		=> [ num => $Q::type_id ],
	]);
	if ($res) {
			reloadDisplay($q, 'Error reactivating type', $res);
	}

	my $ret = AbsenceDB::reactivateType($Q::type_id);
	abslog("Type Reactivated. ID=$Q::type_id");
	dbg("looks good");
	AbsenceDB::setTypeModTime();
	#AbsenceImage::removeLegendFile();
	#AbsenceImage::removeMonthImages();
	reloadTop($q, 'Type Successfully Deactivated', '_top',
		'Finished. Click on button below to reload top page');
	exit;
}

#---------------------------------------------------------------
# ADD USER
#---------------------------------------------------------------

sub mainAddUserForm
{
	my $q = shift;

	print $q->h1('Add a user');
	print qq[<P>A "user" is purely an authentication object, with no
		object associated with it.<P>];

	# ROBBO
	print $q->start_form(
		-name		=> 'add_user',
		-onSubmit	=> 'return validate(document.add_user);',
	);
	print "<TABLE>\n";

	#------------------------
	# Username
	#------------------------
	print "<TR><TH ALIGN=right>Username</TH><TD>\n";
	print $q->textfield(
		-name		=> 'username',
		-size		=> 20,
		-maxlength	=> 20,
	);
	print "</TD><TD>[This is the login-name]</TD></TR>\n";


	if ($MANAGE_PASSWORD) {
		#------------------------
		# password
		#------------------------
		if ($MANAGE_PASSWORD) {
			print "<TR><TH ALIGN=right>Password</TH><TD>\n";
			print $q->textfield(
				-name		=> 'password',
				-size		=> 30,
				-maxlength	=> 30,
			);
			print "</TD></TR>\n";
		}
	}

	aclForm($q, 'add_user');

	print "</TABLE>\n";
	print $q->submit(action => 'Add User');
	print $q->end_form();
	print "<HR>\n";
}

sub addUser
{
	my $q = shift;

	dbg("- add user -");

	if (!length($Q::username)) {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding person', 'You must specify a Username.');
	}

	if ($Q::username eq 'self') {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding person', 'Username must not be "self".');
	}

	if ($Q::username =~ /\s/) {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding person',
			'User-ID must not have any spaces or tabs in it.');
	}

	if ($MANAGE_PASSWORD && !length($Q::password)) {
		dbg("looks bad: no password supplied");
		reloadDisplay($q, 'Error adding person', 'You must specify a Password.');
	}

	my $res = validate_input([
		username	=> $Q::username,
		acl		=> \@Q::group_acl,
		acl		=> \@Q::person_acl,
	]);
	if ($res) {
			reloadDisplay($q, 'Error adding user', $res);
	}


	# ROBBO: big work here
	$DEBUG && abslog("from form: pacl=[".join('/', @Q::person_acl).
		"], gacl=[".join('/', @Q::group_acl).']');

	my ($pacl, $gacl) = convertAcl(\@Q::person_acl, \@Q::group_acl);
	if (!checkAcls($pacl, $gacl, 'add-user')) {
		dbg("looks bad: acl violation");
		reloadDisplay($q, 'Error adding person', 'Invalid person/groups');
	}

	my @params = ($Q::username, $Q::password, $pacl, $gacl);

	dbg("params=(".join(',',@params).")");
	my $ret = AbsenceDB::addUser(@params);
	dbg("ret = [$ret]");

	if ($ret eq 'ok') {
		dbg("looks good");
		print $q->redirect($MAIN_SCRIPT);
	} elsif ($ret =~ /^baddata:\s(.*)$/) {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding user',
			$1);
	} elsif ($ret eq 'duplicate') {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding user',
			"user-ID [$Q::username] already exists in system");
	}

	exit;
}

sub mainModifyUserForm
{
	my $q = shift;

	print $q->h1('Modify a user');
	print $q->start_form;
	print "User";
	makeUsersPopup($q, 'modify');
	print "&nbsp;";
	print $q->submit(action => 'Modify User');
	print $q->end_form();
	print "<HR>\n";
}

sub modifyUserForm
{
	my $q = shift;

	my $uref = AbsenceDB::getUser(id => $Q::user_id);

	if (!defined($uref)) {
		reloadDisplay($q, 'Error modifying user', 'User disappeared.');
	}

	print $q->header(-expires => '-1d');
	print $q->start_html(
		-title		=> "Absence: Modify $uref->{username}",
		-target		=> 'display',
		-BGCOLOR	=> $COLOR_MANAGE,
	);

	includeJsAclScript();
	print $q->h1("Modify $uref->{username}");

	print $q->start_form(
		-method		=> 'POST',
		-name		=> 'mod_user',
		-onSubmit	=> 'return validate(document.mod_user);',
	);

	print "<TABLE>\n";

	#---------------------------------------------
	# User-Name
	#---------------------------------------------
	print qq[<TR><TH ALIGN=right>User-ID</TH><TD BGCOLOR="#bbbbbb">\n];
	print $q->hidden(-name => 'uid', -default => $Q::user_id);
	print "$uref->{username}";
	print "</TD>\n";
	print "<TD>[login-name]</TD>\n";
	print "</TR>\n";

	#------------------------
	# password
	#------------------------
	if ($MANAGE_PASSWORD) {
		print "<TR><TH ALIGN=right>Password</TH><TD>\n";
		print $q->textfield(
			-name		=> 'password',
			-size		=> 30,
			-maxlength	=> 30,
			-value		=> $uref->{password},
		);
		print "</TD><TD>[to change password, clear and enter new password]</TR>\n";
	}

	aclForm($q, 'mod_user', $Q::user_id);

	print "</TABLE>\n";

	print $q->submit(action => 'Submit User Modification');
	print $q->end_form();
	print $q->end_html();
	exit;
}

sub modifyUser
{
	my $q = shift;

	dbg("- modify user -");

	if (!length($Q::uid)) {
		dbg("looks bad");
		reloadDisplay($q, 'Error modifying user',
			'User-ID field must not be blank.');
	}

	if ($MANAGE_PASSWORD && !length($Q::password)) {
		dbg("looks bad: no password supplied");
		reloadDisplay($q, 'Error adding user',
			'You must specify a Password.');
	}

	my $res = validate_input([
		uid		=> [ num	=> $Q::uid ],
		acl	=> \@Q::group_acl,
		acl	=> \@Q::person_acl,
	]);
	if ($res) {
			reloadDisplay($q, 'Error modifying user', $res);
	}

	# ROBBO: add check whether $AUTH_UID is allowed to modify
	# $Q::uid

	my $oldinfo = AbsenceDB::getUser(id => $Q::uid);

	if (!defined($oldinfo)) {
		reloadDisplay($q, 'Error modifying user',
			"user-id [$Q::uid] disappeared.");
	}

	dbg("pacl=[".join(',',@Q::person_acl)."]");
	dbg("gacl=[".join(',',@Q::group_acl)."]");
	my ($pacl, $gacl) = convertAcl(\@Q::person_acl, \@Q::group_acl);
	#dbg("(converted) pacl=[".join(',',@{$pacl})."]");
	#dbg("(converted) gacl=[".join(',',@{$gacl})."]");
	dumpAcl("modifyUser: uid=$Q::uid, after convert:",
		{p => $pacl, g => $gacl});
	if (!checkAcls($pacl, $gacl, 'mod-user', $Q::uid)) {
		dbg("looks bad: acl violation");
		reloadDisplay($q, 'Error adding user', 'Invalid person/groups');
	}

	recombineAcls($Q::uid, $pacl, $gacl);
	dumpAcl("modifyUser: uid=$Q::uid, after recombine:",
		{p => $pacl, g => $gacl});
	my @params = ($Q::uid, $Q::password, $pacl, $gacl);

	my $ret = AbsenceDB::modifyUser(@params);

	dbg("ret = [$ret]");

	if ($ret eq 'ok') {
		dbg("looks good");
		print $q->redirect($MAIN_SCRIPT);
		exit;
	} elsif ($ret =~ /^baddata:\s(.*)$/) {
		dbg("looks bad");
		reloadDisplay($q, 'Error modifying user',
			$1);
	} else {
		dbg("looks bad");
		abslog("error modifying user [$oldinfo->{username}] ret = [$ret]");
		reloadDisplay($q, 'Error modifying user',
			"modify-user [$oldinfo->{username}] failed.");
	}
}

sub mainDeleteUserForm
{
	my $q = shift;

	print $q->h1('Delete a user');
	print $q->start_form;
	print "User";
	makeUsersPopup($q, 'delete');

	print '&nbsp;';
	print $q->submit(action => 'Delete User');
	print $q->end_form();

	print "<HR>\n";
}

sub deleteUser
{
	my $q = shift;

	my $res = validate_input([
		user_id	=> [ num	=> $Q::user_id ],
	]);
	if ($res) {
			reloadDisplay($q, 'Error deleting user', $res);
	}

	my $u_ref = AbsenceDB::getUser(id => $Q::user_id);
	my $ret = AbsenceDB::deleteUser($Q::user_id);
	if ($ret eq 'ok') {
		abslog("deleteUser: username=[$u_ref->{username}]");
		print $q->redirect($MAIN_SCRIPT);
	} elsif ($ret eq 'bad-id') {
		reloadDisplay($q, 'Error deleting user',
			'User was deleted by someone else');
	}
	exit;
}

#=================================================================
#=================================================================
#=================================================================

sub makeUsersPopup
{
	my ($q, $op) = @_;

	my @users = AbsenceAuthorization::getAdminUsers($AUTH_UID, $op);

	my %labels = map { $_ => AbsenceDB::getUser(id => $_, 'username') } @users;

	print $q->popup_menu(
		-name   => 'user_id',
		-values => [sort { $labels{$a} cmp $labels{$b} } @users],
		-labels	=> \%labels,
	);
}

sub makePeoplePopup
{
	my ($q, $op) = @_;

	my @pids = AbsenceAuthorization::getAdminPeople($AUTH_UID, $op);
	#abslog("makePeoplePopup: pids=[".join(',', @pids)."]");

	my $labref;
	foreach my $pid (@pids) {
		$labref->{$pid} = AbsenceDB::getPerson($pid, 'name');
	}

	my @sorted = sort {lc($labref->{$a}) cmp lc($labref->{$b})} @pids;

	print $q->popup_menu(
		-name   => 'person_id',
		-values => \@sorted,
		-labels	=> $labref,
	);
}

sub removeMonthImages
{
	my (@list) = @_;

	my $dh;
	opendir($dh, $IMAGE_DIR) || die "opendir [$IMAGE_DIR]";
	
	my $it = AbsenceConfig::fetch('image_type');
	my $file;
	my $re = '('.join('|',@list).')';
	while($_ = readdir($dh)) {
		if ((@list && /^absence-${re}-.*\.$it$/) ||
			(!@list && /^absence-.*\.$it$/))
		{
			$file = "$IMAGE_DIR/$_";
			dbg("unlinking [$file]");
			unlink($file) || dbg("failed: $!");
			my $map = $file;
			$map =~ s/\.$it$/.map/;
			-f $map && unlink($map) || dbg("failed to unlink map file: $!");
		}
	}
	closedir($dh);
}

#=================================================================
# add a group
#=================================================================

sub addGroup
{
	my $q = shift;

	dbg("- add group -");

	if (!$SUPER_USER) {
		abslog("attempt to circumvent security (addGroup)");
		reloadDisplay($q, 'Error adding group', 'security breach thwarted');
	}

	if (length($Q::name) == 0) {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding group', 'You must supply a name.');
	}

	if ($Q::name =~ /^(all|self)$/i) {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding group', 'The name must not be "self" or "all".');
	}

	my $res = validate_input([
		name			=> $Q::name,
		description		=> $Q::description,
	]);
	if ($res) {
			reloadDisplay($q, 'Error adding group', $res);
	}

	my $ret = AbsenceDB::addGroup('new', $Q::name, $Q::pass, $Q::description);
	dbg("ret = [$ret]");

	if ($ret eq 'ok') {
		dbg("looks good");
		reloadTop($q, 'Group Successfully Added', '_top',
			'Finished. Click on button below to reload top page');
	} elsif ($ret eq 'duplicate') {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding group',
			"[$Q::name] already exists in system");
	} elsif ($ret =~ /^baddata:\s(.*)$/) {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding group',
			$1);
	} else {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding group',
			"[$Q::name] was deleted by someone else");
	}
	exit;
}

sub deleteGroup
{
	my $q = shift;

	dbg("- delete group -");

	if (!$SUPER_USER) {
		abslog("attempt to circumvent security (deleteGroup)");
		reloadDisplay($q, 'Error deleting group', 'security breach thwarted');
	}

	my $res = validate_input([
		group_id	=> [ num	=> $Q::group_id ],
	]);
	if ($res) {
			reloadDisplay($q, 'Error deleting group', $res);
	}

	my $ret = AbsenceDB::deleteGroup($Q::group_id);
	dbg("ret = [$ret]");

	if ($ret eq 'ok') {
		dbg("looks good");
		AbsenceImage::removeMonthImages($Q::group_id);
		reloadTop($q, 'Group Successfully Deleted', '_top',
			'Finished. Click on button below to reload top page');
	} else {
		dbg("looks bad");
		reloadTop($q, 'Error deleting group', 'display',
			'someone deleted the group before you.');
	}
	exit;
}

#=================================================================
# add a person
#=================================================================

sub addPerson
{
	my $q = shift;

	dbg("- add person -");

	if (length($Q::name) == 0) {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding person', 'You must supply a name.');
	}

	if ($Q::name eq 'self') {
		dbg("looks bad");
		reloadDisplay($q, 'Error adding person', 'Name must not be "self".');
	}

	if ($AUTH && $OAP) {
		if (!length($Q::username)) {
			dbg("looks bad");
			reloadDisplay($q, 'Error adding person', 'You must specify a Username.');
		}

		if ($MANAGE_PASSWORD && !length($Q::password)) {
			dbg("looks bad: no password supplied");
			reloadDisplay($q, 'Error adding person',
				'You must specify a Password.');
		}
	}

	my @grps = ($GROUP_POLICY eq 'single') ? ($Q::group_id) : @Q::group_ids;

	my $res = validate_input([
		name		=> $Q::name,
		username	=> $Q::username,
		email		=> $Q::email,
		group_id	=> [ num	=> \@grps ],
		acl		=> \@Q::group_acl,
		acl		=> \@Q::person_acl,
	]);
	if ($res) {
			reloadDisplay($q, 'Error adding person', $res);
	}

	if (!checkGroups(@grps)) {
		dbg("looks bad: acl violation");
		reloadDisplay($q, 'Error adding person', 'Invalid group membership');
	}

	my @params = ('new', $Q::name, $Q::username, $Q::email, \@grps);
	if ($OAP && $AUTH) {
		my ($pacl, $gacl) = convertAcl(\@Q::person_acl, \@Q::group_acl);
		if (!checkAcls($pacl, $gacl, 'add-person')) {
			dbg("looks bad: acl violation");
			reloadDisplay($q, 'Error adding person', 'Invalid person/groups');
		}
		push(@params, $Q::password, $pacl, $gacl);
	}

	dbg("params=(".join(',',@params).")");
	my $ret = AbsenceDB::addPerson(@params);
	dbg("ret = [$ret]");

	if ($ret eq 'ok') {
		dbg("looks good");
		print $q->redirect($MAIN_SCRIPT);
		foreach my $gid (@grps) {
			AbsenceImage::removeMonthImages($gid);
		}
	} elsif ($ret =~ /^duplicate/) {
		dbg("looks bad");
		if ($ret eq 'duplicate-name') {
			reloadDisplay($q, 'Error adding person',
				"[$Q::name] already exists in system");
		} else {
			reloadDisplay($q, 'Error adding person',
				"user-nameID [$Q::username] already exists in system");
		}
	} else {
		reloadDisplay($q, 'Error adding person', "unknown error occurred.");
	}
	exit;
}

# ROBBO: hmm
sub recombineAcls
{
	my ($uid, $pacl, $gacl) = @_;

	$SUPER_USER && return ($pacl, $gacl);

	my ($gvisref, $ginvref, $pvisref, $pinvref) = ([], [], [], []);
	my $uref = AbsenceDB::getUser(id => $uid);
	if (defined($uref)) {
		my $acls = AbsenceDB::getUserAcls($uid);
		($gvisref, $ginvref) = AbsenceAuthorization::divideGroupAcls(
			$AUTH_UID,
			$uid,
			$acls->{group},
		);
		($pvisref, $pinvref) = AbsenceAuthorization::dividePersonAcls(
			$AUTH_UID,
			$acls->{object},
		);
	}

	push(@{$pacl}, @{$pinvref});
	push(@{$gacl}, @{$ginvref});

}

sub checkGroups
{
	my @grps = @_;

	$AUTH || return 1;

	foreach my $grp (@grps) {
		if (!inListN($grp, \@ADMIN_GROUPS)) {
			return 0;
		}
	}

	return 1;
}

sub checkAcls
{
	$AUTH || return 1;
	$SUPER_USER && return 1;

	my ($paclr, $gaclr, $op, $uid) = @_;
	my @admpeople = AbsenceAuthorization::getAdminPeople($AUTH_UID);

	dbg("checkAcls: people");
	foreach my $acl (@{$paclr}) {
		dbg("acl: lev=[$acl->{level}], oid=[$acl->{target}]");
		if ($acl->{target} eq 'self') { next; }
		if (!inListN($acl->{target}, \@admpeople)) {
			dbg("did not find [$acl->{target}] in (".join(',', @admpeople).")");
			return 0;
		}
	}

	dbg("checkAcls: groups");
	foreach my $acl (@{$gaclr}) {
		dbg("acl: lev=[$acl->{level}], gid=[$acl->{target}]");
		if ($acl->{target} eq 'self') {
			($op eq 'mod-person') &&
				!AbsenceAuthorization::selfAllowed($AUTH_UID, $uid) &&
					return 0;
			next;
		}
		if ($acl->{target} eq 'all') {
			return 0;
		}
		if (!inListN($acl->{target}, \@ADMIN_GROUPS)) {
			dbg("did not find [$acl->{target}] in (".join(',', @ADMIN_GROUPS).")");
			return 0;
		}
	}

	return 1;
}

sub convertAcl
{
	my @res;
	for(my $ind = 0; $ind < 2; $ind++) {
		foreach my $acl (@{$_[$ind]}) {
			if ($acl =~ /^(\d+):(.*)$/) {
				my ($lev, $target) = ($1, $2);
				push(@{$res[$ind]}, { level => $lev, target => $target });
			} else {
				die "unable to match [$acl]";
			}
		}
	}

	@res;
}

sub modifyPerson
{
	my $q = shift;

	dbg("- modify person -");
	if (length($Q::name) == 0) {
		dbg("looks bad");
		reloadDisplay($q, 'Error modifying person', 'You must supply a name.');
	}

	if ($AUTH && $OAP && !length($Q::username)) {
		dbg("looks bad");
		reloadDisplay($q, 'Error modifying person',
			'Username field must not be blank.');

		if ($MANAGE_PASSWORD && !length($Q::password)) {
			dbg("looks bad: no password supplied");
			reloadDisplay($q, 'Error adding person',
				'You must specify a Password.');
		}
	}

	my @grps = ($GROUP_POLICY eq 'single') ? ($Q::group_id) : @Q::group_ids;

	#---------------------------------------------------------------
	# checkGroups() makes sure all groups in @grps are in fact
	# administratable by $AUTH_UID
	#---------------------------------------------------------------
	if (!checkGroups(@grps)) {
		dbg("looks bad: acl violation");
		reloadDisplay($q, 'Error adding person', 'Invalid group membership');
	}

	dbg("Q::group_ids=[".join(',', @grps)."]");
	my $oldinfo = AbsenceDB::getPerson($Q::person_id);
	if (!defined($oldinfo)) {
		reloadDisplay($q, 'Error modifying person',
			"person [$Q::name] disappeared.");
	}

	dbg("modifyPerson: modified group_ids=[".join(',', @grps)."]");

	my $uref;
	if ($OAP && $AUTH) {
		# recombine visible and invisible groups
		#my $uref = AbsenceDB::getUser(username => $Q::username);
		$uref = AbsenceDB::getUserForOid($Q::person_id);
		my $inr = [];
		if (defined($uref)) {
			$inr = (AbsenceAuthorization::divideGroups($AUTH_UID, $uref->{id}))[1];
		}
		dbg("modifyPerson: invisible group_ids=[".join(',', @{$inr})."]");
		push(@grps, @{$inr});
		dbg("modifyPerson: recombined group_ids=[".join(',', @grps)."]");
	}

	# username can be passed here even if $AUTH == 0, (it will be undef)
	# because it is ignored in AbsenceDB::addPerson();

	my @params = ($Q::person_id, $Q::name, $Q::username, $Q::email, \@grps);

	if ($OAP && $AUTH) {
		#my $uid = AbsenceDB::findUidForObject($Q::person_id);
		#dbg("pacl=[".join(',',@Q::person_acl)."]");
		#dbg("gacl=[".join(',',@Q::group_acl)."]");
		my ($pacl, $gacl) = convertAcl(\@Q::person_acl, \@Q::group_acl);
		dbg("(converted) pacl=[".join(',',@{$pacl})."]");
		dbg("(converted) gacl=[".join(',',@{$gacl})."]");
		if (!checkAcls($pacl, $gacl, 'mod-person', $uid)) {
			dbg("looks bad: acl violation");
			reloadDisplay($q, 'Error adding person', 'Invalid person/groups');
		}
		recombineAcls($uid, $pacl, $gacl);
		push(@params, $Q::password, $pacl, $gacl);
		dumpAcl("modifyPerson: uid=$Q::username, RECOMBINED:",
			{p => $pacl, g => $gacl});
	}

	my @old_groups = @{$oldinfo->{group}};
	my $ret = AbsenceDB::addPerson(@params);

	dbg("ret = [$ret]");

	if ($ret eq 'ok') {
		dbg("looks good");
		print $q->redirect($MAIN_SCRIPT);
		my (@new, @old);
		@old = @old_groups;
		foreach my $g (@old_groups) {
			if (inListN($g, \@grps)) {
				deleteElementN($g, \@old);
			} else {
				push(@new, $g);
			}
		}
		push(@new, @old);

		AbsenceImage::removeMonthImages(@new);

		exit;
	} elsif ($ret eq 'duplicate-username') {
		dbg("looks bad");
		abslog("error modifying person_id [$Q::person_id] ret = [$ret]");
		reloadDisplay($q,
			"Error modifying person, username [$Q::username] already in use.",
			"modify-person [$Q::name] failed.");
	} else {
		dbg("looks bad");
		abslog("error modifying person_id [$Q::person_id] ret = [$ret]");
		reloadDisplay($q,
			'Error modifying person',
			"modify-person [$Q::name] failed.");
	}
}

sub deleteElementN
{
	my($elem, $lref) = @_;

	my $count = 0;
	my $match = 0;
	foreach my $tmp (@$lref) {
		if ($tmp == $elem) {
			$match = 1;
			last;
		}
		$count++;
	}

	if ($match) {
		splice(@$lref, $count, 1);
	}
}

sub inListN
{
	my ($thing, $lref) = @_;

	foreach my $elem (@{$lref}) {
		($thing == $elem) && return 1;
	}

	return 0;
}

sub modifyPersonForm
{
	my $q = shift;

	my $pref = AbsenceDB::getPerson($Q::person_id);

	dbg("modifyPersonForm: person_id=[$Q::person_id]");
	print $q->header(-expires => '-1d');
	print $q->start_html(
		-title		=> "Absence: Modify $pref->{name}",
		-target		=> 'display',
		-BGCOLOR	=> $COLOR_MANAGE,
	);

	if (!defined($pref)) {
		reloadDisplay($q, 'Error modifying person', 'Person disappeared.');
	}

	if ($AUTH) {
		includeJsAclScript();
	}
	print $q->h1("Modify $pref->{name}");

	my @onsubmit = ($OAP && $AUTH)
		? (-onSubmit   => 'return validate(document.mod_person);') 
		: ();

	print $q->start_form(
		-method	=> 'POST',
		-name	=> 'mod_person',
		@onsubmit,
	);
	print $q->hidden(-name => 'person_id', -default => $Q::person_id);
	print "<TABLE>\n";

	#---------------------------------------------
	# name
	#---------------------------------------------
	print "<TR><TH ALIGN=right>Name (Last, First)</TH><TD>\n";
	print $q->textfield(
		-name		=> 'name',
		-size		=> 20,
		-maxlength	=> 20,
		-value		=> $pref->{name},
	);
	print "</TD><TD>[This will appear in the month-images]</TD></TR>\n";

	#---------------------------------------------
	# Username
	#---------------------------------------------
	my $uid;
	if ($AUTH && $OAP) {
		$uid = AbsenceDB::findUidForObject($Q::person_id);
		my ($username, $password);
		if (defined($uid)) {
			my $uref = AbsenceDB::getUser(id => $uid);
			$username = $uref->{username};
			$password = $uref->{password};
		}
		#------------------------
		# USERNAME
		#------------------------
		print "<TR><TH ALIGN=right>Username</TH><TD>\n";
		print $q->textfield(
			-name		=> 'username',
			-size		=> 20,
			-maxlength	=> 20,
			-value		=> $username,
		);
		print "</TD><TD>[This is the login-name</TD></TR>\n";
		#------------------------
		# PASSWORD
		#------------------------
		if ($MANAGE_PASSWORD) {
			print "<TR><TH ALIGN=right>Password</TH><TD>\n";
			print $q->textfield(
				-name		=> 'password',
				-size		=> 30,
				-maxlength	=> 30,
				-value		=> $password,
			);
			print "</TD><TD>[to change password, clear and enter new password]</TD></TR>\n";
		}
	}

	#---------------------------------------------
	# E-Mail address
	#---------------------------------------------
	print "<TR><TH ALIGN=right>E-Mail Address</TH><TD>\n";
	print $q->textfield(
		-name		=> 'email',
		-size		=> 30,
		-maxlength	=> 50,
		-value		=> exists($pref->{email}) ? $pref->{email} : '',
	);
	print "</TD></TR>\n";

	#---------------------------------------------
	# group
	#---------------------------------------------
	print "<TR><TH ALIGN=right VALIGN=top>Group</TH><TD>\n";

	#abslog("calling makeGroupPopup with gid=$gid");

	my @gids = AbsenceDB::getObjectGroups($Q::person_id);
	if ($GROUP_POLICY eq 'single') {
		makeGroupPopup($q, $gids[0]);
	} else {
		makeGroupCheckboxes($q, \@gids);
	}

	print "</TD><TD VALIGN=top>[select groups to which person should belong]</TD></TR>\n";

	($AUTH && $OAP) && aclForm($q, 'mod_person', $uid);
	print "</TABLE>\n";

	print $q->submit(action => 'Submit Modification');
	print $q->end_form();
	print $q->end_html();
	exit;
}

sub modifyGroup
{
	my $q = shift;

	if (!$SUPER_USER) {
		abslog("attempt to circumvent security (modifyGroup)");
		reloadDisplay($q, 'Error modifying group', 'security breach thwarted');
	}

	dbg("- modify group -");
	if (length($Q::name) == 0) {
		dbg("looks bad");
		reloadDisplay($q, 'Error modifying group', 'You must supply a name.');
	}
	my $ret = AbsenceDB::addGroup(
		$Q::group_id,
		$Q::name,
		$Q::pass,
		$Q::description,
	);
	dbg("ret = [$ret]");

	if ($ret eq 'ok') {
		dbg("looks good");
		reloadTop($q, 'Group Successfully Modified', '_top',
			'Finished. Click on button below to reload top page');
		#reloadDisplay($q, 'Group Successfully Modified',
		#	"new name is [$Q::name]");
	} else {
		dbg("looks bad");
		abslog("error modifying group_id [$Q::group_id] ret = [$ret]");
		reloadDisplay($q, 'Error modifying group',
			"modify-group [$Q::name] failed.");
	}
	exit;
}

sub modifyGroupForm
{
	my $q = shift;

	my $gref = AbsenceDB::getGroup($Q::group_id);

	if (!defined($gref->{name})) {
		reloadTop($q, "Modify Group $gref->{name}", 'display',
			"Modify Group $gref->{name}", 'display',
			'Group disappeared');
	}

	print $q->header(-expires => '-1d');
	print $q->start_html(
		-title		=> "Absence: Modify Group $gref->{name}",
		-target		=> 'display',
		-BGCOLOR	=> $COLOR_MANAGE,
	);

	print $q->h1('Modify a Group');
	print $q->start_form;
	print $q->hidden(-name => 'group_id', -default => $Q::group_id);
	print "<TABLE>\n";
	print "<TR><TH ALIGN=right>Group Name</TH><TD>\n";
	print $q->textfield(
		-name		=> 'name',
		-size		=> 20,
		-maxlength	=> 20,
		-value		=> $gref->{name},
	);
	print "</TD></TR>\n";
	print "<TR><TH ALIGN=right>Description</TH><TD>\n";
	print $q->textfield(
		-name		=> 'description',
		-size		=> 40,
		-maxlength	=> 60,
		-value		=> $gref->{description},
	);
	print "</TD></TR>\n";
	print "</TD></TR></TABLE>\n";

	print $q->submit(action => 'Submit Group Modification');
	print $q->end_form();
	exit;
}

sub makeGroupCheckboxes
{
	my ($q, $gref) = @_;
	my @gids = sort {$a <=> $b} @ADMIN_GROUPS;

	my $labref;
	my $name;
	foreach my $group_id (@gids) {
		$name = AbsenceDB::getGroup($group_id, 'name');
		$labref->{$group_id} = $name;
	}

	my @sorted = sort {lc($labref->{$a}) cmp lc($labref->{$b})} @gids;

	print $q->checkbox_group(
		-name		=> 'group_ids',
		-values		=> \@sorted,
		-default	=> $gref,
		-labels		=> $labref,
		-linebreak	=> 'true',
	);
}

sub makeGroupPopup
{
	my ($q, $default, $form, $action) = @_;

	my @addit;
	if ($form) {
		@addit = (-onChange => "JavaScript:document.${form}.submit('$action');"),
	}

	my $labref;
	foreach my $group_id (@ADMIN_GROUPS) {
		$labref->{$group_id} = AbsenceDB::getGroup($group_id, 'name');
	}

	my @sorted = sort {lc($labref->{$a}) cmp lc($labref->{$b})} @ADMIN_GROUPS;
	print $q->popup_menu(
		-name   	=> 'group_id',
		-values		=> \@sorted,
		-labels		=> $labref,
		-default	=> $default,
		@addit,
	);
}

sub deletePerson
{
	my $q = shift;

	# get gid first
	#my $gref = AbsenceDB::getPerson($Q::person_id, 'group');
	my @gids = AbsenceDB::getObjectGroups($Q::person_id);

	my $ret = AbsenceDB::deletePerson($Q::person_id);
	if ($ret eq 'ok') {
		print $q->redirect($MAIN_SCRIPT);
		foreach my $gid (@gids) {
			AbsenceImage::removeMonthImages($gid);
		}
	} elsif ($ret eq 'bad-id') {
		reloadDisplay($q, 'Error deleting person',
			'Person was deleted by someone else');
	}
	exit;
}

sub reloadTop
{
	my ($q, $title, $target, $msg) = @_;

	my $js = qq[
		function goto_top() { top.location = '$TOP_PAGE'; }
		setTimeout('goto_top()', 5000);
	];

	my @args = ($target eq '_top')
		? (-script => $js)
		: ();

	print $q->header(-expires => '-1d');
	print $q->start_html(
		-title		=> "Absence: $title",
		-target		=> $target,
		-BGCOLOR	=> $COLOR_MANAGE,
		@args,
	);

	print $q->h1($msg);
	#print $q->start_form(
	#	-action => $TOP_PAGE,
	#	-method => 'GET',
	#	-target	=> '_top',
	#	#-onSubmit	=> " top.location = '$TOP_PAGE';",
	#);
	print button(
		-name		=> 'reload_button',
		-value		=> 'Reload Calendar',
		-onClick	=> "parent.location.replace('$TOP_PAGE');",
	);
	#print $q->submit(action => 'Reload Calendar');
	#print $q->end_form();
	print $q->end_html();
	exit 0;
}

sub reloadDisplay
{
	my ($q, $title, $msg) = @_;

	print $q->header(-expires => '-1d');
	print $q->start_html(
		-title		=> "Absence: $title",
		-target		=> 'display',
		-BGCOLOR	=> $COLOR_MANAGE,
	);

	print $q->h1($msg);
	print $q->start_form(
		-action => $MAIN_SCRIPT,
		-method => 'GET',
	);
	print $q->submit(action => 'Reload Calendar');
	print $q->end_form();
	print $q->end_html();
	exit 0;
}

sub dbg
{
	my $msg = shift;
	$DEBUG || return;
	abslog("DEBUG: absence-manage: $msg");
}

sub old_dbg
{
	$DEBUG || return;

	my $msg = shift;
	my $fh = FileHandle->new(">>$DEBUG_OUT");
	defined($fh) || die "open $DEBUG_OUT for appending";
	print $fh "$msg\n";
	$fh->close;
}

sub dumpAcl
{
	my ($msg, $acl) = @_;

	dbg("dumpAcl: $msg");
	exists($acl->{p}) && dbg("  PERSON: "
		. join('/', map "lev=$_->{level},oid=$_->{target}", @{$acl->{p}}));
	exists($acl->{g}) && dbg("  GROUPS: "
		. join('/', map "lev=$_->{level},gid=$_->{target}", @{$acl->{g}}));
}

sub convertDefaultAcl
{
	my $acl = shift;

	$acl || return ();

	my %level_map = (
		r		=> 1,
		read	=> 1,
		w		=> 2,
		write	=> 2,
	);

	if ($acl =~ /^([wr]|write|read):(all|self)$/i) {
		return { level => $level_map{$1}, target => $2 };
	}

	die "unable to parse default-acl [$acl]";
}

sub aclForm
{
	my ($q, $form, $uid) = @_;

	$DEBUG && abslog("aclForm: form=[$form], uid=[$uid]");
	my $js_target = 'document.'.$form;

	my $thing = ($form =~/_person$/) ? 'person' : 'user';
	my $verb = ($form =~/add/) ? 'adding' : 'modifying';

	my ($pdref, $gdref);

	$pdref = {};
	$gdref = {};

	my ($lev, $person, $gid);

	my @levmap = (
		undef,
		'Read',
		'Write',
		undef,
		'Admin',
	);

	if ($form ne 'add_user') {
		my (@pdefs, @gdefs);

		if ($uid) {
			my $acl = AbsenceAuthorization::getAdminAcls($AUTH_UID, $uid);
			#if ($DEBUG) {
			#	my $realacl = AbsenceDB::getAcl($uid);
			#	dumpAcl("aclForm: form=$form, uid=$uid, REAL:", $realacl);
			#	dumpAcl("aclForm: form=$form, uid=$uid, PRUNED:", $acl);
			#}
			@pdefs = @{ $acl->{object} };
			@gdefs = @{ $acl->{group} };
			$DEBUG && abslog("dump of \@gdefs:\n".Dumper(\@gdefs)."\n--done--");
		} else {
			my $tmp = AbsenceConfig::fetch('pacl_default');
			@pdefs = (convertDefaultAcl($tmp));
			$tmp = AbsenceConfig::fetch('gacl_default');
			@gdefs = (convertDefaultAcl($tmp));
		}

		foreach my $acl (@pdefs) {
			($acl->{level} == 2) || die "nonsense pacl level [$acl->{level}]";
			if ($acl->{target} eq 'self') {
				$pdref->{'2:self'} = 'Write:self';
			} elsif ($acl->{target} =~ /^(\d+)$/) {
				my $name = AbsenceDB::getPerson($acl->{target}, 'name');
				$pdref->{"2:$acl->{target}"} = "Write:$name";
			} else {
				die "strange person rule: [$acl->{level}:$acl->{target}]";
			}
		}

		foreach my $acl (@gdefs) {
			if ($acl->{target} =~ /^(all|self)$/) {
				$gdref->{"$acl->{level}:$acl->{target}"}
					= "$levmap[$acl->{level}]:$acl->{target}";
			} elsif ($acl->{target} =~ /^(\d+)$/) {
				my $name = AbsenceDB::getGroup($acl->{target}, 'name');
				$gdref->{"$acl->{level}:$acl->{target}"}
					= "$levmap[$acl->{level}]:$name";
			} else {
				die "strange group rule: [$acl->{level}:$acl->{target}]";
			}
		}
	}

	print qq[<TR><TD></TD><TD COLSPAN=99>\n];
	javaScriptWarning();
	print qq[</TH></TR>\n];

	######################### GROUP START #########################
	#------------------------------------------------------------
	# GROUP stuff
	#------------------------------------------------------------
	print qq[<TR><TH ALIGN=right VALIGN=top>Group ACL<BR>Management</TH>
		<TD COLSPAN=2>\n];

	# TABLE wrapper
	print qq[<TABLE BORDER=1><TR><TD>\n];

	print qq[<TABLE>\n];

	#-------------------------------------------
	# instructions
	#-------------------------------------------
	if ($thing eq 'person') {
		print qq[<TR><TH COLSPAN=99 ALIGN=left><SMALL>
			Use Group ACLs to control what
			the person you are $verb can do with her own group or other
			groups.  Choose a permission (read/write/admin), choose a
			target group for that permission to act upon, then click on
			"Add" to add the 'rule' to the list on the right.  The group
			"self" refers to all groups to which the person belongs.<BR>
			Permissions: read: view group, write: add/change/delete
			absences, admin: add/change/delete people.
			<BR><HR>
			</SMALL></TH></TR>\n];
	} else {
		print qq[<TR><TH COLSPAN=99 ALIGN=left><SMALL>
			Use Group ACLs to control what
			the user you are $verb can do with
			groups.  Choose a permission (read/write/admin), choose a
			target group for that permission to act upon, then click on
			"Add" to add the 'rule' to the list on the right.
			<BR>
			Permissions: read: view group, write: add/change/delete
			absences, admin: add/change/delete people.
			<BR><HR>
			</SMALL></TH></TR>\n];
	}

	print qq[<TR>
		<TD VALIGN=top WIDTH="100">
		Permission:<BR>\n];

	my @perms = qw(read write admin);
	my $lref = {
		1	=> 'Read',
		2	=> 'Write',
		4	=> 'Admin',
	};

	print $q->radio_group(
			-name		=> 'groupverb',
			-values		=> [1, 2, 4],
			-linebreak	=> 'true',
			-labels		=> $lref,
		),
		qq[</TD>
		<TD VALIGN=top WIDTH="150">
		Group:<BR>\n];

	$lref = undef;
	foreach my $gid (@ADMIN_GROUPS) {
		$lref->{$gid} = AbsenceDB::getGroup($gid, 'name');
	}
	$lref->{self} = 'self';
	$lref->{all} = 'all';
	@list = ();
	if (($form eq 'add_person') ||
		(($form eq 'mod_person') &&
		 (AbsenceAuthorization::selfAllowed($AUTH_UID, $uid))))
	{
		@list = ('self');
	}
	AbsenceAuthorization::adminAll($AUTH_UID) && push(@list, 'all');
	push(@list, sort {lc($lref->{$a}) cmp lc($lref->{$b})} @ADMIN_GROUPS);
	print $q->popup_menu(
		-name		=> 'grouplist',
		-values		=> \@list,
		-labels		=> $lref,
		-size		=> 5,
		-multiple	=> undef,
	);
	print qq[</TD>
		<TD WIDTH="100">\n],
		$q->button(
			-name	=> 'add_group',
			-value	=> 'Add',
			-onClick	=> "add('group', $js_target)",
		), "<BR>\n",
		$q->button(
			-name	=> 'del_group',
			-value	=> 'Delete',
			-onClick	=> "del('group', $js_target)",
		),
		qq[</TD>
		<TD WIDTH="150">\nSelected Rules:<BR>\n],
		$q->popup_menu(
			-name		=> 'group_acl',
			-values		=> [keys(%{$gdref})],
			-labels		=> $gdref,
			-size		=> 5,
			-multiple	=> undef,
		),
		qq[</TD>\n</TR>\n];

	# HR
	#print qq[<TR><TD COLSPAN=4><HR></TD></TR>\n];

	print qq[</TABLE>\n];

	# TABLE wrapper
	print qq[</TD></TR></TABLE>\n];
	######################### GROUP END #########################

	print qq[</TD></TR>\n];

	######################### PERSON START #########################
	#------------------------------------------------------------
	# PERSON stuff
	#------------------------------------------------------------
	print qq[<TR><TH ALIGN=right VALIGN=top>Person ACL<BR>Management</TH>
		<TD COLSPAN=2>\n];

	# TABLE wrapper
	print qq[<TABLE BORDER=1><TR><TD>\n];

	print qq[<TABLE>\n];

	#-------------------------------------------
	# instructions
	#-------------------------------------------
	if ($thing eq 'person') {
		print qq[<TR><TH COLSPAN=99 ALIGN=left><SMALL>
			Use Person ACLs to control what
			the person you are $verb can do with herself or other
			people in groups which are visible to her (see above).
			The only permission available is "write". Choose a
			target person for whom "write" access is to be allowed,
			then click "Add" to add the 'rule' to the list on the right.
			The special person "self" refers to the person you are $verb.
			<BR>
			"write" access means adding, modifying, and deleting absences
			is allowed.
			<BR><HR>
			</SMALL></TH></TR>\n];
	} else {
		print qq[<TR><TH COLSPAN=99 ALIGN=left><SMALL>
			Use Person ACLs to control what
			the user you are $verb can do with
			people in groups which are visible to her (see above).
			The only permission available is "write". Choose a
			target person for whom "write" access is to be allowed,
			then click "Add" to add the 'rule' to the list on the right.
			The special person "self" refers to the person you are $verb.
			<BR>
			"write" access means adding, modifying, and deleting absences
			is allowed.
			<BR><HR>
			</SMALL></TH></TR>\n];
	}

	print qq[<TR>
		<TD VALIGN=top WIDTH="100">
		Permission:<BR>
		<INPUT TYPE=radio NAME="personverb" VALUE="write" CHECKED>Write
		</TD>
		<TD VALIGN=top WIDTH="150">
		Person:<BR>\n];

	$lref = {};
	my @people = AbsenceAuthorization::getAdminPeople($AUTH_UID);
	foreach my $pid (@people) {
		$lref->{$pid} = AbsenceDB::getPerson($pid, 'name');
	}
	$lref->{self} = 'self';
	my @list = ($form =~ /_person$/) ?  ('self') : ();
	push(@list, sort {lc($lref->{$a}) cmp lc($lref->{$b})} @people);

	print $q->popup_menu(
		-name		=> 'personlist',
		-values		=> \@list,
		-labels		=> $lref,
		-size		=> 5,
		-multiple	=> undef,
	);
		
	print qq[</TD>\n
		<TD WIDTH="100">\n],
		$q->button(
			-name	=> 'add_person',
			-value	=> 'Add',
			-onClick	=> "add('person', $js_target)",
		),
		"<BR>\n",
		$q->button(
			-name	=> 'del_person',
			-value	=> 'Delete',
			-onClick	=> "del('person', $js_target)",
		),
		qq[</TD>
		<TD WIDTH="150">\nSelected Rules:<BR>],
		$q->popup_menu(
			-name		=> 'person_acl',
			-values		=> [keys(%{$pdref})],
			-labels		=> $pdref,
			-size		=> 5,
			-multiple	=> undef,
		),
		qq[</TD></TR>\n];
		#print qq[<TR><TD COLSPAN=4 ALIGN="center">\n];
		#print qq[<SMALL>NOTE: you cannot add "write" permission for a
		#	person in a group for which the user does not have
		#	"read" permission.</SMALL>\n];
		#print qq[</TD></TR>\n];

		print qq[<TR>\n</TABLE>\n];

	# TABLE wrapper
	print qq[</TD></TR></TABLE>\n];
	######################### PERSON END #########################

	print qq[</TD></TR>\n];
}

sub includeJsAclScript
{
	my $path = AbsenceConfig::fetch('js_acl_script');
	#my $path = "$base/acl.js";
	#print qq[<SCRIPT LANGUAGE="JavaScript" SRC="$path"></SCRIPT>\n];
	print qq{<script src="$path" type="text/javascript"></script>\n};
}

sub includeJsRtCoincidenceScript
{
	my $base = AbsenceConfig::fetch('js_dir_rel');
	my $path = "$base/rt_coincidence.js";
	#print qq[<SCRIPT LANGUAGE="JavaScript" SRC="$path"></SCRIPT>\n];
	print qq{<script src="$path" type="text/javascript"></script>\n};
}

sub includeJsAclScript_old
{
	my $path = AbsenceConfig::fetch('js_acl_script');

	#print qq[<SCRIPT LANGUAGE="JavaScript" SRC="$path"></SCRIPT>\n];
	my $fh = FileHandle->new($path) || die "open [$path] to read";
	print qq[<SCRIPT LANGUAGE="JavaScript">\n<!--\n];
	while(<$fh>) {
		print $_;
	}
	$fh->close;
	print "-->\n</SCRIPT>\n";
}

sub javaScriptWarning
{
	print <<_EOF_;
<NOSCRIPT>
<FONT SIZE=4 COLOR="red">
You will not be able to manage ACLs because your browser<BR>
does not support JavaScript or you have not enabled it.
</FONT>
</NOSCRIPT>
_EOF_
}

__END__

=head1 NAME

absence-manage.pl - manage people and groups

=head1 DESCRIPTION

C<absence-manage.pl> is responsible for displaying some forms for
the management of people and groups, and handling the processing
after the user clicks on submit in one of the forms.

=cut
