#!/usr/bin/perl

# $Id: absence.pl 91 2009-07-14 15:58:06Z urban $

use CGI;
# uncomment the following if you are debugging
use CGI::Carp qw(fatalsToBrowser);

$CGI::POST_MAX = 300;
$CGI::DISABLE_UPLOADS = 1;

#----------------------------------------------------------------
# it is very important that this BEGIN block come before
# "use AbsenceConfig;" below.  The INSTANCE_NAME environment
# variable must be defined before AbsenceConfig is loaded.
#----------------------------------------------------------------
BEGIN {
	if ($ENV{QUERY_STRING} =~ /^instance=(\S+)$/) {
		# an instance has been specified
		$ENV{INSTANCE_NAME} = $1;
		$COOKIE_NAME = "absence-$1";
	} else {
		$COOKIE_NAME = 'absence';
	}
}

use AbsenceConfig;
use AbsenceImage;
use AbsenceLog;
use AbsenceAuthentication;
use AbsenceAuthorization;

my $IMAGE_DIR_ABS	= AbsenceConfig::fetch('image_dir_abs');
my $DATA_REL		= AbsenceConfig::fetch('data_dir_rel');
my $CGI_DIR			= AbsenceConfig::fetch('cgi_dir_rel');
my $MAP_TYPE		= AbsenceConfig::fetch('map_type');
my $COOKIE_DOMAIN	= AbsenceConfig::fetch('cookie_domain');
my $AUTH			= AbsenceConfig::fetch('authentication');
my $AUTH_TYPE		= AbsenceConfig::fetch('auth_type');
my $NPH_LO			= AbsenceConfig::fetch('nph_lo_script');
#my $MAIN_SCRIPT		= AbsenceConfig::fetch('main_script');
my $MARK_CM			= AbsenceConfig::fetch('mark_curr_month');
my $MARK_CD			= AbsenceConfig::fetch('mark_curr_day');
my $COLOR_MAIN		= AbsenceConfig::fetch('wp_main');
my $AUTH_UID;

my $DEBUG			= 0;
my $DEF_MONTHS		= 6;
my $VERSION			= '1.6';

$SIG{__DIE__} = \&handleErrors;

#=================================================================
# main
#=================================================================

my ($month, $year) = (localtime)[4,5];
$month++;
$year += 1900;

my $q = new CGI;
$q->import_names;
$q->charset('utf-8');

#-----------------------------------------------------------------
# first, authenticate if necessary
#-----------------------------------------------------------------
my $auth_cookie;
($AUTH_UID, $auth_cookie) = AbsenceAuthentication::authenticateUser($q);

my @cookies;
if ($auth_cookie) {
	push(@cookies, $auth_cookie);
}

#if ($AUTH && !defined($AUTH_UID)) {
#	$AUTH_UID = AbsenceAuthentication::authenticateUser($q);
#}

if (defined($Q::action) && ($Q::action eq 'Logout')) {
	logout($q);
}

#-----------------------------------------------------------------
# other stuff
#-----------------------------------------------------------------
my %cookie = $q->cookie($COOKIE_NAME);

my $chunk_mult	= 0;
my $chunksize	= $DEF_MONTHS;

my $cd = "cookie debugging...<BR>\n";

if (exists($cookie{mult})) {
	$chunk_mult = $cookie{mult};
	$cd .= "cookie: MULT = $chunk_mult<BR>\n";
}

if (exists($cookie{chunksize})) {
	$chunksize = $cookie{chunksize};
	$cd .= "cookie: CHUNKSIZE = $chunksize<BR>\n";
}

if (exists($cookie{month})) {
	$month = $cookie{month};
	$year = $cookie{year};
	$cd .= "cookie: MONTH = $month, YEAR = $year<BR>\n";
}

if (exists($cookie{group_id})) {
	$group_id = $cookie{group_id};
	$DEBUG && abslog("GROUP-ID=[$group_id], defined=[".defined($group_id)."], len=".length($group_id));
	if (defined($group_id) && length($group_id)) {
		if (!AbsenceDB::groupExists(id => $group_id)) {
			$cd .= "group_id [$group_id] disappeared, erasing.\n";
			$group_id = undef;
		}
		$cd .= "cookie: GROUP_ID = $group_id<BR>\n";
	} else {
		$group_id = undef;
	}
}

$DEBUG && abslog("action = [$Q::action]");

# if $Q::start_month is set, I know that a submit() took place

if (($Q::start_month) || ($Q::jump =~ /^(Next|Previous)/)) {
	$year		= $q->param('start_year');
	$month		= $q->param('start_month');
	$chunksize	= $q->param('chunksize');
	$group_id	= $q->param('group_id');
}

if ($Q::jump =~ /^Next/) {
	if (!%cookie) {
		print $q->header(-expires => 'now'),
			$q->start_html('Absence'),
			$q->h1('The <EM>Next</EM> button will not work without cookies.');
		print $q->end_html;
		exit;
	}
	$chunk_mult++;
	$cd .= "must increment. now mult = $chunk_mult<BR>\n";
} elsif ($Q::jump =~ /^Previous/) {
	if (!%cookie) {
		print $q->header(-expires => 'now'), $q->start_html('Absence');
		print $q->h1('The <EM>Previous</EM> button will not work without cookies.');
		print $q->end_html;
		exit;
	}
	$chunk_mult--;
	$cd .= "must decrement. now mult = $chunk_mult<BR>\n";
} elsif ($Q::start_month) {
	$chunk_mult = 0;
	$cookie{month} = $month;
	$cookie{year} = $year;
	$cd .= "block = \$Q::start_month<BR>\n";
} elsif ($Q::action eq '') {
	$cd .= "block = no action<BR>\n";
}

#-------------------------------------------------------------
# always set cookie
#-------------------------------------------------------------

$cookie{chunksize} = $chunksize;
$cookie{mult} = $chunk_mult;
$cookie{group_id} = $group_id;
$cookie = $q->cookie(
	-name	=> $COOKIE_NAME,
	-value	=> \%cookie,
	-path	=> '/',
	#-domain	=> $COOKIE_DOMAIN,
	-expires=> '+10y',
);
push(@cookies, $cookie);

# for debugging
#abslog("setting cookie:\n\tname=$COOKIE_NAME\n\tchunksize=$cookie{chunksize}\n\tmult=$cookie{mult}\n\tgid=$cookie{group_id}");

if ($chunk_mult != 0) {
	($month, $year) = moveByChunk($month, $year, $chunk_mult, $chunksize);
}

#my @cookies = (-cookie => $cookie);
#if ($AUTHCOOK) { push(@cookies, (-cookie => $AUTHCOOK)); }

print $q->header(-expires => '-1d', -cookie => \@cookies),
	$q->start_html(
		-title		=> 'Absence',
		-BGCOLOR	=> $COLOR_MAIN,
	);

$DEBUG && print "$cd";

my @groups = AbsenceAuthorization::getReadGroups($AUTH_UID);
if ($DEBUG) {
	print "\@groups = (".join(',', @groups).")<BR>\n";
}

if (!length($group_id) || !grep(/^${group_id}$/, @groups)) {
	if (@groups == 1) {
		$group_id = $groups[0];
	} else {
		print $q->h1('Select Group Above');
		print $q->end_html;
		exit;
	}
}

#print $q->h1('Absences for group FOOBAR');

# debug
if ($DEBUG) {
	print "DOCUMENT_ROOT = [$ENV{DOCUMENT_ROOT}]<BR>\n";
	my @keywords = $q->keywords;
	my @names = $q->param;

	print "Parameters:<BR>\n";
	foreach my $p (@names) {
		print " [$p] ".$q->param($p)."<BR>\n";
	}

	print "<P>Keywords:<BR>\n";
	foreach my $p (@keywords) {
		print " [$p]<BR>\n";
	}

	if ($cookie{chunksize}) {
		print "CHUNKSIZE: [$chunksize]<BR>\n";
	}
	print "chunk_mult = [$chunk_mult]<BR>\n";
	#print "<P>ENVIRONMENT:<BR><TABLE>\n";
	#foreach my $var (sort keys(%ENV)) {
	#	print "<TR><TH ALIGN=right>$var</TH><TD>$ENV{$var}</TD></TR>\n";
	#}
	#print "</TABLE>\n";
}

#my $legend = "$IMAGE_DIR_ABS/legend.$it";
#AbsenceImage::makeLegend($legend);

my $legend_rel = (AbsenceImage::legendFilePath())[0];
my $legend_mtime = AbsenceImage::makeLegend();

print qq[<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=0><TR><TD>\n];
print qq[<IMG BORDER=2 SRC="$legend_rel?$legend_mtime"></A>\n];
print qq[</TD><TD>\n];
print qq[&nbsp;(<A HREF="$DATA_REL/absence-help.html">instructions</A>)&nbsp;];

if (
	!$AUTH
	&& AbsenceConfig::fetch('partial_auth_hack')
	&& !AbsenceAuthentication::isAuthed()
)
{
	print qq[<BR>&nbsp;(<A HREF="$CGI_DIR/absence.pl?action=partial_auth">admin</A>)&nbsp;];

}
print "</TD>\n";

if ($AUTH) {
	print "<TD VALIGN=bottom><BR>\n";
	if ($AUTH_TYPE eq 'simple') {
		print $q->start_form(-method => 'POST'),
			$q->submit(action => 'Logout'),
			$q->end_form();
	} elsif ($AUTH_TYPE eq 'http') {
		print $q->start_form(-action => $NPH_LO, -method => 'GET'),
			$q->submit(action => 'Logout'),
			$q->end_form();
	}
	print "</TD>\n";
}

print "</TR></TABLE><BR>\n";

if ($DEBUG) {
	print "remote_user = [$ENV{REMOTE_USER}]<BR>\n";
	print "auth_uid = [$AUTH_UID]<BR>\n";
	print qq[<A HREF="/cgi-bin/del-auth-cookie.pl">delete auth cookie</A><BR>\n];
}

#---------------------
# more debug stuff...
#---------------------
#print "environment:<BR>\n<TABLE>\n";
#foreach my $key (sort keys(%ENV)) {
#	printf("<TR><TD>%s</TD><TD>%s</TD></TR>\n", $key, $ENV{$key});
#}
#print "</TABLE>\n";

my $count = 0;
my ($width, $height);
my $inst;
my %mmodtimes;
my ($modtime, $targ);

if (exists($ENV{INSTANCE_NAME})) { $inst = "/in=$ENV{INSTANCE_NAME}"; }

my $it = AbsenceConfig::fetch('image_type');

while($count < $chunksize) {
	my $image_fn = "absence-$group_id-$month-$year.$it";
	my $map_fn = "absence-$group_id-$month-$year.map";
	my $image_out = "$IMAGE_DIR_ABS/$image_fn";
	my $map_out = "$IMAGE_DIR_ABS/$map_fn";
	my $mapname	= "g${group_id}m${month}y$year";
	my $in = "${DATA_REL}/img/$image_fn";
	my $older = imageOlder($image_out, $group_id, $month, $year);
	if ($MARK_CM || $MARK_CD) {
		#abslog("loop: $month/$year");
		if (currentChanged($group_id, $month, $year)) {
			#abslog("loop2: currentChanged returned 1");
			$older = 1;
		}
	}
	if ((! -e $image_out) || $older) {
		if ($DEBUG) {
			if (! -e $image_out) { abslog("[$image_out] not found"); }
			if ($older) { abslog("older"); }
		}
		#print "-creating $image_out -<BR>\n";
		#print "<!-- creating [$image_out] -->\n";
		$DEBUG && abslog("creating [$image_out]");
		$mmodtimes{$year}->{$month} = time();
		AbsenceImage::create($month, $year, $group_id, $image_out, $map_out, $mapname);
	}

	$modtime = (stat($image_out))[9];
	($width, $height) = AbsenceDB::getImageDimensions($group_id, $month, $year);
	if ($MAP_TYPE eq 'server') {
		$targ = "${CGI_DIR}/absence-click.pl$inst/${group_id}-${month}-$year";
		print $q->a(
			{ -href => $targ },
			$q->img(
				{
					-src	=> "$in?$modtime",
					-border => 2,
					-ismap	=> undef,
					-width	=> $width,
					-height	=> $height,
				}
			),
		);
	} else {
		printMap($map_out);
		print $q->img(
				{
					-src	=> "$in?$modtime",
					-border => 2,
					-usemap	=> "#$mapname",
					-width	=> $width,
					-height	=> $height,
				}
			);
	}
	print "<BR>\n";

	if ($month == 12) {
		$year++;
		$month = 1;
	} else {
		$month++;
	}

	$count++;
}

if (%mmodtimes) {
	AbsenceDB::updateMonthModTimes($group_id, \%mmodtimes);
}

print $q->end_html;

#=====================================================================
# subs
#=====================================================================

sub printMap
{
	my $filename = shift;

	open(MAP, $filename) || die "could not open map [$filename]: $!";
	while(<MAP>) {
		print;
	}
	close(MAP);
}

sub handleErrors
{
	my $msg = shift;
	confess($msg);
}

#---------------------------------------------------------------------
# currentChanged()
# ----------------
# determine if the current month or current day has changed since
# the current month/day was last determined
#---------------------------------------------------------------------
sub currentChanged
{
	my ($gid, $month, $year) = @_;
	my $jm = $year * 12 + $month;

	#----------------------------------------------------
	# get last modification time from database
	#----------------------------------------------------
	my $mod_time = AbsenceDB::getModificationTime($gid, $month, $year);
	defined($mod_time) || return 1;

	my ($md, $mm, $my) = (localtime($mod_time))[3,4,5];
	$mm++;
	$my += 1900;
	my $mt_jm = $my * 12 + $mm;

	#----------------------------------------------------
	# get current values
	#----------------------------------------------------
	my ($d, $m, $y) = (localtime)[3,4,5];
	$m++;
	$y += 1900;
	my $curr_jm = $my * 12 + $mm;

	#----------------------------------------------------
	# is the image in question for the current month?
	#----------------------------------------------------
	if ($jm == $curr_jm) {
		if ($mt_jm != $jm) { return 1; }
		if ($MARK_CD) {
			if ($d == $md) {
				return 0;
			} else {
				return 1;
			}
		} else {
			# don't care about current day, so image is still ok
			return 0;
		}
	} else {
		if ($mt_jm == $jm) {
			# image was last generated when it was the current month
			return 1;
		}
		return 0;
	}
}

sub imageOlder
{
	my ($file, $gid, $month, $year) = @_;

	my $fmt = (stat($file))[9];
	my $smt = AbsenceDB::getModificationTime($gid, $month, $year);

	if (!defined($smt)) {
		$DEBUG && abslog("no modtime found for [$gid-$month/$year]");
		return 1;
	}

	print "<!-- fmt=[$fmt], smt=[$smt] -->\n";
	if ($fmt < $smt) {
		$DEBUG && abslog("$gid-$month/$year: older\nfmt=[$fmt], smt=[$smt]");
		return 1;
	}
	$DEBUG && abslog("$gid-$month/$year: current");
	return 0;
}

sub moveByChunk
{
	my ($month, $year, $mult, $chunksize) = @_;

	my $tmp = $year * 12 + $month;
	$tmp += ($mult * ($chunksize - 1));
	if ($chunksize != 1) { $tmp--; }

	my $m = $tmp % 12 + 1;
	my $y = int($tmp/12);

	($m, $y);
}

sub logout
{
	my $q = shift;
	AbsenceAuthentication::logout($q);
	exit 0;
}

__END__

=head1 NAME

absence.pl - heart of web-absence

=head1 DESCRIPTION

C<absence.pl> is responsible for displaying the month-images
in the main frame (below the control frame).  C<absence.pl> gets
the information about groups and months to display first from
C<absence-control.pl> vi CGI, and secondly via cookies (i.e.,
C<absence-control.pl> has precedence).

=cut
