#!/usr/bin/perl

#-------------------------------------------------------------------------
# install.pl
#
# If you already have a previous version of absence installed, you
# can use install.pl to update your configuration.  It will not
# overwrite the datafiles (database-file, hoiday-db, absence-types-db,
# log-file).
#
# This script is a nightmare.  I know.  It seems to work, though.
#-------------------------------------------------------------------------

use FileHandle;
use File::Basename;
use Config;

use lib qw{./utils};
use AbsenceInstall;
use AbsenceImport;
use AbsenceMigration;
use AbsenceUtils qw(query yesNo);

use strict;

my %VALUES;
my $CMDOUT;

# structure to hold information about parameters
my %PARMS = (
	data_dir_rel	=> {
		short	=> 'ddr',
		class	=> 'path',
		desc	=> 'relative path to data',
	},
	data_dir_abs	=> {
		short	=> 'dda',
		class	=> 'path',
		desc	=> 'absolute path to data',
		def		=> [
			'$DOCUMENT_ROOT',
			'data_dir_rel'
		],
	},
	cgi_dir_rel		=> {
		short	=> 'cdr',
		class	=> 'path',
		desc	=> 'relative path to cgi-bin',
		def		=> '/cgi-bin',
	},
	cgi_dir_abs		=> {
		short	=> 'cda',
		class	=> 'path',
		desc	=> 'absolute path to cgi-bin',
		def		=> [
			'$DOCUMENT_ROOT',
			'..',
			'cgi_dir_rel'
		],
	},
	image_dir_rel	=> {
		short	=> 'idr',
		class	=> 'path',
		desc	=> 'relative path to image-dir',
		def		=> [
			'data_dir_rel',
			'img',
		],
	},
	js_dir_rel	=> {
		short	=> 'jdr',
		class	=> 'path',
		desc	=> 'relative path to JavaScript-dir',
		def		=> [
			'data_dir_rel',
			'JavaScript',
		],
	},
	js_dir_abs	=> {
		short	=> 'jda',
		class	=> 'path',
		desc	=> 'absolute path to JavaScript-dir',
		def		=> [
			'data_dir_abs',
			'JavaScript'
		],
	},
	image_dir_abs	=> {
		short	=> 'ida',
		class	=> 'path',
		desc	=> 'absolute path to image-dir',
		def		=> [
			'data_dir_abs',
			'img',
		],
	},
	util_dir_abs	=> {
		short	=> 'uda',
		class	=> 'path',
		desc	=> 'absolute path to util-dir',
		def		=> [
			'data_dir_abs',
			'Utils',
		],
	},
	log_file		=> {
		short	=> 'lf',
		class	=> 'file',
		perm	=> 0660,
		desc	=> 'absolute path to log-file',
		name	=> 'log',
		def		=> [
			'data_dir_abs',
			'name',
		],
	},
	main_script		=> {
		short	=> 'ms',
		class	=> 'script',
		desc	=> 'relative path to main-script',
		name	=> 'absence.pl',
		def		=> [
			'cgi_dir_rel',
			'name',
		],
	},
	manage_script	=> {
		short	=> 'mgs',
		class	=> 'script',
		desc	=> 'relative path to manage-script',
		name	=> 'absence-manage.pl',
		def		=> [
			'cgi_dir_rel',
			'name',
		],
	},
	control_script	=> {
		short	=> 'cs',
		class	=> 'script',
		desc	=> 'relative path to control-script',
		name	=> 'absence-control.pl',
		def		=> [
			'cgi_dir_rel',
			'name',
		],
	},
	top_page		=> {
		short	=> 'tp',
		class	=> 'path',
		desc	=> 'relative path to starting-page',
		def		=> 'data_dir_rel',
	},
	legend_style	=> {
		short	=> 'ls',
		class	=> 'param',
		desc	=> '"fit" or "constant:XX"',
		def		=> 'fit',
		allow	=> '(fit|constant:\d+)',
	},
	image_type		=> {
		short	=> 'it',
		class	=> 'param',
		desc	=> '"png", "jpg", or "gif"',
		def		=> 'png',
		allow	=> ['png', 'jpg', 'gif'],
	},
	map_type		=> {
		short	=> 'mt',
		class	=> 'param',
		desc	=> "map-type (client or server)",
		def		=> 'client',
		allow	=> ['client', 'server'],
		extra	=> 'type "server" is not well tested',
	},
	cookie_domain	=> {
		short	=> 'cd',
		class	=> 'param-disuse',
		desc	=> 'domain to use for cookies',
	},
	holiday_label	=> {
		short	=> 'hl',
		class	=> 'param',
		desc	=> 'label to use for holidays',
		def		=> 'Holiday',
	},
	holiday_scheme	=> {
		short	=> 'hsch',
		class	=> 'param',
		desc	=> 'holiday scheme',
		extra	=> 'advanced: country/region-based holidays, basic: unified holidays',
		def		=> 'basic',
		allow	=> [ 'advanced', 'basic' ],
	},
	holiday_mark	=> {
		short	=> 'hmrk',
		class	=> 'param-holiday-basic',
		desc	=> 'holiday indicator',
		extra	=> 'small: show only in header, large: in entire column',
		def		=> 'small',
		allow	=> [ 'small', 'large' ],
	},
	show_holidays_in_header	=> {
		short	=> 'shih',
		class	=> 'param-holiday-advanced',
		desc	=> 'show regional/country holidays in header',
		def		=> 'yes',
		allow	=> [ 'yes', 'no' ],
	},
	header_holiday_country	=> {
		short	=> 'hhc',
		class	=> 'param-holiday-advanced',
		desc	=> 'show holidays from this country in header',
		def		=> undef,
	},
	header_holiday_region	=> {
		short	=> 'hhr',
		class	=> 'param-holiday-advanced',
		desc	=> 'show holidays from this region in header',
		def		=> undef,
	},
	skip_we_hol	=> {
		short	=> 'swh',
		class	=> 'param',
		desc	=> 'do not display reservations for sat/sun/holidays',
		def		=> 'no',
		allow	=> [ 'yes', 'no' ],
	},
	skip_mark_skipped	=> {
		short	=> 'sms',
		class	=> 'param-skip',
		desc	=> 'draw empty blocks for skipped days',
		def		=> 'no',
		allow	=> [ 'yes', 'no' ],
	},
	group_policy	=> {
		short	=> 'gp',
		class	=> 'param',
		desc	=> 'group-policy',
		def		=> 'single',
		allow	=> ['single', 'multiple'],
	},
	authentication	=> {
		short	=> 'auth',
		class	=> 'param',
		desc	=> 'Use authentication',
		def		=> 'no',
		allow	=> ['yes', 'no'],
	},
	auth_type	=> {
		short	=> 'at',
		class	=> 'param-auth',
		desc	=> 'authentication type ("http" or "simple")',
		def		=> 'simple',
		allow	=> ['http', 'simple'],
	},
	secret	=> {
		short	=> 'sec',
		class	=> 'param-auth-simple',
		desc	=> 'a short secret string for hashing',
		def		=> undef,
	},
	credential_src	=> {
		short	=> 'crs',
		class	=> 'param-auth',
		desc	=> 'source for credentials',
		def		=> 'absence',
		allow	=> ['absence', 'htaccess'],
	},
	session_timeout	=> {
		short	=> 'st',
		class	=> 'param-auth-simple',
		desc	=> 'session timeout',
		extra	=> 'values: 0 = none, 1m = 1 minute, 1d = 1 day, etc',
		def		=> 0,
		allow	=> '\d+[smhdwM]',
	},
	manage_password	=> {
		short	=> 'mp',
		class	=> 'param-auth',
		desc	=> 'absence manages passwords',
		def		=> 'yes',
		allow	=> ['yes', 'no'],
	},
	pw_hash_format	=> {
		short	=> 'phf',
		class	=> 'param-auth-simple',
		desc	=> 'how to store managed passwords',
		def		=> 'md5',
		allow	=> ['md5', 'cleartext'],
	},
	htaccess_path	=> {
		short	=> 'htp',
		class	=> 'param-auth-http',
		desc	=> 'absolute path to ".htaccess" file',
		def		=> undef,
	},
	auth_realm	=> {
		short	=> 'ar',
		class	=> 'param-auth-http',
		desc	=> 'realm name for http authentication',
		def		=> 'Absence',
	},
	pacl_default	=> {
		short	=> 'pacl',
		class	=> 'param-auth',
		desc	=> 'default person ACL',
		def		=> 'undef',
		allow	=> '(w:self)?',
		undef	=> 1,
	},
	gacl_default	=> {
		short	=> 'gacl',
		class	=> 'param-auth',
		desc	=> 'default group ACL',
		def		=> 'w:self',
		allow	=> '([wr]:(self|all)+)?',
		undef	=> 1,
	},
	objects_are_people	=> {
		short	=> 'oap',
		class	=> 'param-auth',
		desc	=> 'are managed objects people',
		def		=> undef,
		allow	=> ['yes', 'no'],
	},
	nph_lo_script	=> {
		short	=> 'nls',
		class	=> 'script',
		desc	=> 'path to logout script',
		name	=> 'nph-absence-logout.pl',
		def		=> [
			'cgi_dir_rel',
			'name',
		],
	},
	js_acl_script	=> {
		short	=> 'acla',
		class	=> 'script',
		desc	=> 'absolute path to JS ACL source',
		name	=> 'acl.js',
		def		=> [
			'js_dir_rel',
			'name',
		],
	},
	multi_res	=> {
		short	=> 'mr',
		class	=> 'param',
		desc	=> 'allow multiple reservations on same day',
		def		=> undef,
		allow	=> [ 'yes', 'no' ],
	},
	max_multi	=> {
		short	=> 'mmr',
		class	=> 'param-multi',
		desc	=> 'max num multi reservations',
		def		=> 2,
		allow	=> '\d+',
	},
	variable_height	=> {
		short	=> 'mrvh',
		class	=> 'param-multi',
		desc	=> 'allow height of person-row to vary',
		def		=> undef,
		allow	=> [ 'yes', 'no' ],
	},
	min_height	=> {
		short	=> 'mrmh',
		class	=> 'param-multi',
		desc	=> 'min height (in full reservations) for person-rows',
		def		=> 1,
		allow	=> '\d+',
	},
	alternate_colors	=> {
		short	=> 'altc',
		class	=> 'param',
		desc	=> 'alternate person-row bg colors',
		def		=> 'no',
		allow	=> [ 'yes', 'no' ],
	},
	vertical_grid	=> {
		short	=> 'vtg',
		class	=> 'param',
		desc	=> 'draw vertical grid lines',
		def		=> 'yes',
		allow	=> [ 'yes', 'no' ],
	},
	database_name	=> {
		short	=> 'dbn',
		class	=> 'param-db',
		desc	=> 'the database name',
		def		=> 'absence',
	},
	database_host	=> {
		short	=> 'dbh',
		class	=> 'param-db',
		desc	=> 'the database host',
		def		=> 'localhost',
	},
	database_port	=> {
		short	=> 'dbpt',
		class	=> 'param-db',
		desc	=> 'the database port number',
		def		=> undef,
		none	=> 1,
	},
	database_user	=> {
		short	=> 'dbu',
		class	=> 'param-db',
		desc	=> 'the database username',
		def		=> 'absence',
	},
	database_pass	=> {
		short	=> 'dbp',
		class	=> 'param-db',
		desc	=> 'the database password',
		def		=> 'absence',
	},
	database_adm_user	=> {
		short	=> 'dbau',
		class	=> 'param-db',
		desc	=> 'the database admin username',
		def		=> 'postgres',
	},
	database_adm_pass	=> {
		short	=> 'dbap',
		class	=> 'param-db',
		desc	=> 'the database admin password',
		def		=> 'undef',
		none	=> 1,
		undef	=> 1,
		extra	=> qq{use "undef" to indicate that no password is necessary,\nwhich may allow an admin connection to the server\nwithout providing a password},
	},
	migrate_data => {
		short	=> 'mig',
		class	=> 'install',
		desc	=> 'migrate existing v1.X data',
		def		=> 'no',
		allow	=> [ 'yes', 'no' ],
	},
);

my @PATHS = qw(
	cgi_dir_rel
	cgi_dir_abs
	data_dir_rel
	data_dir_abs
	image_dir_rel
	image_dir_abs
	js_dir_rel
	js_dir_abs
	top_page
);

my @cgi = qw(
	absence.pl
	absence-click.pl
	absence-control.pl
	absence-manage.pl
	nph-absence-logout.pl
	AbsenceConfig.pm
	AbsenceDate.pm
	AbsenceDB.pm
	AbsenceImage.pm
	AbsenceInput.pm
	AbsenceLog.pm
	AbsenceAuthentication.pm
	AbsenceAuthorization.pm
);

my @js = qw(
	acl.js
	rt_coincidence.js
);

my @misc = qw(
	INSTALL.txt
	README-AUTH.txt
	README-CONFIGURATION.txt
	README-INSTANCES.txt
	README.txt
);

my @html = qw(
	absence-help.html
);

#demo.html

my $VERBOSE = 1;
my $DEBUG = 0;

my %VALS;

#----------------------------------------------------------------------
# main
#----------------------------------------------------------------------

umask(0);

# generate short-name to long-name param map
my ($long, $desc);
my $force = 0;

my $WINDOZE = 0;
if ($Config{osname} eq 'MSWin32') {
	# blech, gasp, spit, hack, cough, wheeze, grunt, scream
	$WINDOZE = 1;
}

if (!$WINDOZE && ($> == 0)) {
	print "$0 should not be run as root.\n";
	print "please re-run as a normal user.\n\n";
	if (!yesNo("Are you SURE you want to continue", 0)) {
		print "bailing!\n";
		exit;
	}
}

#--------------------------------------------------------------------
# create hash to look up nicknames
#--------------------------------------------------------------------
my %SHORTMAP;
foreach (keys(%PARMS)) {
	$SHORTMAP{ $PARMS{$_}->{short} } = $_;
}

#--------------------------------------------------------------------
# parse script parameters
#--------------------------------------------------------------------
while($_ = shift) {
	s/^-//;
	if (exists($SHORTMAP{$_})) {
		$_ = $SHORTMAP{$_};
	}
	if (/^h$/) {
		usage();
		exit;
	} elsif (/^d$/) {
		print "setting DEBUG mode.\n";
		$DEBUG++;
	} elsif (/^f/) {
		$force = 1;
	} elsif (/^(dr|docroot)/) {
		$ENV{DOCUMENT_ROOT} = shift;
	} elsif (exists($PARMS{$_})) {
		my $val = shift;
		my $tmp = ($val =~ /^def(ault)?$/) ? default($_) : $val;
		print "setting $_ to [$tmp]\n";
		#$tmp = undef if ($tmp eq 'undef');
		$VALUES{$_} = $tmp;
	} elsif (exists($SHORTMAP{$_})) {
		my $val = shift;
		my $tmp = ($val =~ /^def(ault)?$/) ? default($_) : $val;
		print "setting $_ to [$tmp]\n";
		#$tmp = undef if ($tmp eq 'undef');
		$VALUES{$SHORTMAP{$_}} = $tmp;
	}
}

print <<_EOF_;

Starting installation of absence V2.0...

I may need to ask you some configuration details.

The descriptions of the parameters refer to "absolute" and "relative"
paths.  By "absolute", I mean a normal (filesystem) path as seen by a
script, for example "/users/apache/html".  By "relative", I mean the
path component of a URL, i.e., relative to DOCUMENT_ROOT.
For example, given the URL

	http://www.unix-wissen.de/apps/absence

The "relative path" in this case is "/apps/absence".

_EOF_

#--------------------------------------------------------------------
# ask general questions
#--------------------------------------------------------------------
my $use_dr = 0;
#foreach my $p (@PATHS, getClassKeys('param')) {
foreach my $p (getClassKeys('path', 'param')) {
	checkParam($p);
	if ($VALS{$p} =~ /\$DOCUMENT_ROOT/) {
		$use_dr = 1;
		if ($WINDOZE) {
			print "WARNING: Windoze/IIS does not support DOCUMENT_ROOT\n";
		}
	}
}

#--------------------------------------------------------------------
# ask database questions
#--------------------------------------------------------------------
print "\nNow I must get information about database.\n\n";
#$VALUES{db_admin_user} = query('What is the username of the DB administrator');
#$VALUES{db_admin_pass} = query('What is the password for the DB administrator');
foreach my $p (getClassKeys('param-db')) {
	checkParam($p);
}

my @db_params = (
	user	=> $VALUES{database_adm_user},
	pass	=> $VALUES{database_adm_pass},
);

if (defined($VALUES{database_host}) && ($VALUES{database_host} ne 'localhost'))
{
	push(@db_params, host => $VALUES{database_host});
}

if (!AbsenceInstall::testConnection(@db_params)) {
	die "connection to database as $VALUES{database_adm_user} failed\n";
}
print "successfully connected to database as user [$VALUES{database_adm_user}]\n";

#--------------------------------------------------------------------
# ask authentication/authorization questions
#--------------------------------------------------------------------
if ($VALUES{authentication} eq 'yes') {
	print "\nNow I must ask you some questions about authentication.\n\n";
	checkParam('auth_type');
	my $type = $VALUES{auth_type};

	foreach my $p (getClassKeys('param-auth', "param-auth-$type")) {
		checkParam($p);
	}
}

#--------------------------------------------------------------------
# ask multi-res questions
#--------------------------------------------------------------------
if ($VALUES{multi_res} eq 'yes') {
	print "\nNow I must ask some questions about multiple reservations.\n\n";

	foreach my $p (getClassKeys('param-multi')) {
		checkParam($p);
	}
}

#--------------------------------------------------------------------
# ask holiday questions
#--------------------------------------------------------------------
print "\nNow I must ask some questions about holidays.\n\n";
my $scheme = $VALUES{holiday_scheme};

foreach my $p (getClassKeys("param-skip-$scheme")) {
	checkParam($p);
}

#--------------------------------------------------------------------
# ask skip questions
#--------------------------------------------------------------------
if ($VALUES{skip_we_hol} eq 'yes') {
	print "\nNow I must ask some questions about skipping weekends/holidays.\n\n";

	foreach my $p (getClassKeys('param-skip')) {
		checkParam($p);
	}
}

#--------------------------------------------------------------------
# ask install questions
#--------------------------------------------------------------------
print "\nNow I must ask some questions about the installation.\n\n";

foreach my $p (getClassKeys('install')) {
	checkParam($p);
}

#--------------------------------------------------------------------
# sort out DOCUMENT_ROOT
#--------------------------------------------------------------------
my $dr;
if ($use_dr) {
	print "** USE DOCUMENT_ROOT **\n";
	if (!exists($ENV{DOCUMENT_ROOT})) {
		while(1) {
			$dr = query('enter the value of $DOCUMENT_ROOT');
			if (-d $dr) { last; }
			print "\n[$dr] is not a valid directory.\n";
		}
		$ENV{DOCUMENT_ROOT} = $dr;
	} else {
		$dr = $ENV{DOCUMENT_ROOT};
	}
}

#--------------------------------------------------------------------
# setup file paths
#--------------------------------------------------------------------
foreach my $p (getClassKeys('file')) {
	if (exists($VALUES{$p}) && $VALUES{$p}) {
		$VALS{$p} = expand($VALUES{$p});
	} else {
		my $v = default($p);
		$VALUES{$p} = $v;
		$VALS{$p} = expand($v);
		$VALS{$p} =~ s!//+!/!g;
		$DEBUG && print "got [$v]\n";
		$VALS{$p} =~ s/^\s+//;
	}
}

#--------------------------------------------------------------------
# expand $DOCUMENT_ROOT in all parameters
#--------------------------------------------------------------------
if ($use_dr) {
	foreach my $p (keys(%VALS)) {
		$DEBUG && print "[$p] VALS{$p} = [$VALS{$p}]\n";
		if ($VALS{$p} =~ /\$DOCUMENT_ROOT/) {
			$DEBUG && print "%% DOC_ROOT FOUND in [$p], val=[$VALS{$p}] %%\n";
			$VALS{$p} =~ s/\$DOCUMENT_ROOT/$dr/;
			$VALS{$p} =~ s!//!/!g;
			$DEBUG && print "%% After subst: [$VALS{$p}] %%\n";
		}
	}
}

#--------------------------------------------------------------------
# substitute "$ENV{DOCUMENT_ROOT}" for "$DOCUMENT_ROOT" in %values
#--------------------------------------------------------------------
foreach my $key (keys(%VALUES)) {
	if ($VALUES{$key} =~ /\$DOCUMENT_ROOT/) {
		$VALUES{$key} =~ s/\$DOCUMENT_ROOT/\$ENV{DOCUMENT_ROOT}/;
	}
}

if (($VALUES{map_type} ne 'client') && ($WINDOZE)) {
	print "A map_type of 'client' doesn't work on Windoze.\n";
	print "Automatically setting map_type to 'server'\n";
	$VALUES{map_type} = 'server';
}

#------------------------------------------------------------------
# I have collected all parameter values, so I can now start 
# creating directories (if necessary) and copying files
#------------------------------------------------------------------

my $migrate = ($VALS{migrate_data} eq 'yes') ? "\n- migrate data" : '';
print <<_EOF_;

I am about to perform the following steps:
- create [$VALS{data_dir_abs}] if necessary
- create [$VALS{cgi_dir_abs}] if necessary
- create [$VALS{js_dir_abs}] if necessary
- create [$VALS{util_dir_abs}] if necessary
- copy Perl modules and scripts to [$VALS{cgi_dir_abs}]
- copy HTML, README, and data files to [$VALS{data_dir_abs}]
- copy JavaScript files to [$VALS{js_dir_abs}]
- copy utility files to [$VALS{util_dir_abs}]
- create a new top-level HTML entry point (index.html on unix)
- create a new log-file in [$VALS{log_file}]
- create a new image directory [$VALS{image_dir_abs}]
- set parameters in AbsenceConfig.pm
- change permissions on datafiles and directories
- create a new empty database$migrate

_EOF_

if (!yesNo('are you ready', 1)) {
	print "installation aborted by user.\n";
	exit;
}

verbose('checking/creating directory structure');

# cgi-bin-directory
if (! -e $VALS{cgi_dir_abs}) {
	verbose("cgi-bin directory [$VALS{cgi_dir_abs}] does not exist. creating...");
	makePath($VALS{cgi_dir_abs}, 0755)
		|| die "mkdir for $VALS{cgi_dir_abs} failed. err=$CMDOUT\n";
}

# data-directory
if (! -e $VALS{data_dir_abs}) {
	verbose("data directory [$VALS{data_dir_abs}] does not exist. creating...");
	makePath($VALS{data_dir_abs}, 0755)
		|| die "mkdir for $VALS{data_dir_abs} failed. err=$CMDOUT\n";
}

# image-directory
if (! -d $VALS{image_dir_abs}) {
	verbose("image directory [$VALS{image_dir_abs}] does not exist. creating...");
	myMkdir($VALS{image_dir_abs}, 0770) || die "mkdir failed";
}

# javascript-directory
if (! -d $VALS{js_dir_abs}) {
	verbose("javascript directory [$VALS{js_dir_abs}] does not exist. creating...");
	makePath($VALS{js_dir_abs}, 0750) || die "mkdir failed";
}

# utils-directory
if (! -d $VALS{util_dir_abs}) {
	verbose("utility directory [$VALS{util_dir_abs}] does not exist. creating...");
	makePath($VALS{util_dir_abs}, 0750) || die "mkdir failed";
}

#------------------------------------------------------------------
# figure out where source files should be
#------------------------------------------------------------------
my $dn = dirname($0);
my $SRCDIR;
print "\n";
while(1) {
	$SRCDIR = query('where is the unpacked kit ("." is OK)', $dn);
	if (! -e "$SRCDIR/cgi-bin/absence.pl") {
		print "that doesn't seem to be right...\n";
	} else {
		last;
	}
}

#------------------------------------------------------------------
# copy files into place
#------------------------------------------------------------------
verbose("copying cgi-bin files");
foreach my $file (@cgi) {
	copyFile("$SRCDIR/cgi-bin/$file", $VALS{cgi_dir_abs})
		|| die "copy of cgi-bin/$file to $VALS{cgi_dir_abs} failed. err=$CMDOUT\n";
}

verbose("copying misc files");
foreach my $file (@misc) {
	copyFile("$SRCDIR/$file", $VALS{data_dir_abs})
		|| die "copy of $file to $VALS{data_dir_abs} failed. err=$CMDOUT\n";
}

verbose("copying html files");
foreach my $file (@html) {
	copyFile("$SRCDIR/html/$file", $VALS{data_dir_abs})
		|| die "copy of html/$file to $VALS{data_dir_abs} failed. err=$CMDOUT\n";
}

# javascript files
verbose("copying javascript files");
my @files = glob("$SRCDIR/javascript/*");
foreach my $file (@files) {
	copyFile($file, $VALS{js_dir_abs})
		|| die "copy of $file to $VALS{js_dir_abs} failed. err=$CMDOUT\n";
}

@files = glob("$SRCDIR/utils/*");
verbose("copying util files");
foreach my $file (@files) {
	copyFile($file, $VALS{util_dir_abs})
		|| die "copy of utils/$file to $VALS{util_dir_abs} failed. err=$CMDOUT\n";
}

#------------------------------------------------------------------
# create log file
#------------------------------------------------------------------
verbose("creating log-file");
touch($VALS{log_file}, 0660);

#------------------------------------------------------------------
# create HTML entry point
#------------------------------------------------------------------
verbose("creating index.html");
createIndex();

#------------------------------------------------------------------
# create database
#------------------------------------------------------------------
verbose("creating database");
my $rv = createDatabase($SRCDIR);

#------------------------------------------------------------------
# add countries to database
#------------------------------------------------------------------
if ($rv eq 'exists') {
	print "will not import country-list into existing DB.\n";
}
else {
	importCountries($SRCDIR);
}

#------------------------------------------------------------------
# edit AbsenceConfig.pm
#------------------------------------------------------------------
verbose("editing AbsenceConfig.pm");
editConfigFile("$VALS{cgi_dir_abs}/AbsenceConfig.pm", \%VALUES);

#------------------------------------------------------------------
# load AbsenceDB module, which is needed by AbsenceMigration and
# addAdministrators().
#------------------------------------------------------------------
AbsenceUtils::loadAbsenceDB($VALS{cgi_dir_abs});

#------------------------------------------------------------------
# migrate data, if requested
#------------------------------------------------------------------
if ($VALUES{migrate_data} eq 'yes') {
	eval {
		AbsenceMigration::migrateData(\%VALS);
	};
	if ($@) {
		print "\n** migration failed. see logfile. error:\n$@\n";
	}
	else {
		print "migration successful.\n";
	}
}

#------------------------------------------------------------------
# add admin users/people
#------------------------------------------------------------------
addAdministrators($rv);

#------------------------------------------------------------------
# worry about file and directory permissions
#------------------------------------------------------------------
if ($WINDOZE) {
	#--------------------------------------------------------------
	# Windoze
	#--------------------------------------------------------------
	print <<'	_EOF_';

	Your platform is windows, so I'm going to let you worry about file
	and directory permissions.  I can offer these hints:
	- $VALS{cgi_dir_abs} should be executable for scripts.
	- $VALS{data_dir_abs} whould be writable.

	You will probably need to create two virtual directories, one called
	"absence", pointing to {$VALS{data_dir_abs}", the
	other called "cgi-bin" pointing to "$VALS{cgi_dir_abs}".

	Also, remember that multiple instances won't work on windoze.
	_EOF_
	exit;
}

#--------------------------------------------------------------
# UNIX
#--------------------------------------------------------------
print <<_EOF_;

There are two ways of setting up file and directory permissions:
The quick-n-dirty way of the proper way.

The quick-n-dirty is not desirable because it causes various
files and directories to be set world-writable and world-executable.
This makes it insecure.  The Proper method uses group permissions
to allow the webserver process(es) access to the files in question.

Use the proper method if you can.

Quick-n-dirty:
  - set mode on data_dir_abs to		0777
  - set mode on image_dir_abs to	0777
  - set mode on log_file to		0666

Proper:
  - set group, mode on data_dir_abs to    <webserver>, 0770
  - set group, mode on image_dir_abs to   <webserver>, 0770
  - set group, mode on log_file to 	  <webserver>, 0660

The catch to the proper way is that some steps must be
performed as root.  To that end this script will generate
a small bourne-shell script to be executed by root.

Please select the method you want to use.

_EOF_

my $root_script;
my $ans = query('(q) quick-n-dirty / (p) proper', 'p', ['p', 'q']);
if (lc($ans) eq 'q') {
	quickPermissions();
	print "finished with installation\n";
} else {
	$root_script = properPermissions();
	print basename($0)." finished.  Please run '$root_script' as root\n";
}

if (!$WINDOZE) {
	print "\nThe author of absence is curious about who is using it,\n";
	print "and would like to send a mail to himself via this script.\n";
	print "answer no if you don't want to or don't have a direct connection\n";
	print "to the internet.\n";

	if (yesNo('Send a mail to author', 1)) {
		sendMail();
	}
}

if ($ans eq 'p') {
	my $str = "** Don't forget to run '$root_script' as root! **";
	my $len = length($str);
	print "\n", '*' x $len, "\n", $str, "\n", '*' x $len, "\n";
}

exit;

#======================================================================
# subroutines
#======================================================================

sub addAdministrators
{
	my $db_status = shift;

	if ($VALUES{authentication} ne 'yes') {
		return;
	}

	if ($db_status eq 'exists' && !yesNo('do you need to add an administrator?', 1))
	{
		return;
	}

	verbose("adding administrators");

	if ($DEBUG) {
		print "debug-mode, will not add administrators.\n";
		return;
	}

	print "\n** adding administrators **\n";
	print "\nYou must add at least one priviledged user to the system.\n";

	my $get_pw = ($VALUES{manage_password} eq 'yes');

	my $super_acl = {
		level	=> 4,
		target	=> 'all',
	};

	#----------------------------------------
	# loop until no more are desired
	#----------------------------------------
	while(1) {
		my $pw = '';
		my $user = query('Enter  Username', '');
		if ($get_pw) {
			$pw = query('Enter Password', '');
		}
		my $ret = AbsenceDB::addUser($user, $pw, [], [ $super_acl ]);
		if ($ret eq 'ok') {
			print "sucessfully added [$user]\n";
		} elsif ($ret eq 'duplicate') {
			print "[$user] already exists in database.\n";
		} else {
			print "error adding new user. (ret=[$ret])\n";
		}
		print "\n";
		if (yesNo('done adding admin users', 'yes')) {
			last;
		}
	}
	print "done adding users.\n";
}

sub checkParam
{
	my $p = shift;

	if (exists($VALUES{$p}) && $VALUES{$p}) {
		if ($VALUES{$p} eq 'undef') {
			$VALUES{$p} = undef;
		}
		$VALS{$p} = expand($VALUES{$p});
	} else {
		my $v = queryParam($p);
		if ($v eq 'undef') {
			$VALS{$p} = $VALUES{$p} = undef;
		} else {
			$VALUES{$p} = $v;
			$VALS{$p} = expand($v);
			$VALS{$p} =~ s!//+!/!g;			# remove multiple slashes
			$DEBUG && print "got [$v]\n";
			$VALS{$p} =~ s/^\s+//;			# remove all white-space
		}
	}
}

# not currently used
sub getAllClasses
{
	my %classes;
	foreach my $key (keys(%PARMS)) {
		my $class = $PARMS{$key}->{class};
		$classes{$class} = 1;
	}

	return keys(%classes);
}

sub getClassKeys
{
	my @classes = @_;

	if ($classes[0] eq 'all') {
		# I list them here to control the order
		@classes = qw(path file script param param-db param-multi param-holiday param-auth param-auth-simple param-auth-http param-skip install);
	}

	my @out;
	my @tmp;
	foreach my $class (@classes) {
		foreach my $p (keys(%PARMS)) {
			if ($PARMS{$p}->{class} eq $class) {
				push(@tmp, $p);
			}
		}
		push(@out, sort @tmp);
		@tmp = ();
	}

	@out;
}

sub sendMail
{
	# only know how to send mail
	my @list = glob('/usr/*bin/sendmail');
	if (!@list) {
		print "can't find sendmail, giving up.\n";
		return;
	}

	print "I hope you don't mind answering a few questions...\n";
	print "The default answer is 'n/a'.  Feel free to leave the default.\n";
	my $purpose = query('What do you intend to do with absence', 'n/a');
	my $company = query('What does your company do', 'n/a');
	my $name = query('What is your name?', 'n/a');
	my $comments = query('Would you like to make any comments', 'none');

	my $fh = FileHandle->new("|$list[0] urban\@unix-beratung.de");
	print $fh <<_EOF_;
To: urban\@unix-beratung.de
Subject: someone using absence

purpose: $purpose
company does: $company
user: $name
comments: $comments
_EOF_

	print "Thanks!\n";
}

sub docRoot
{
	$ENV{DOCUMENT_ROOT};
}

sub expand
{
	my $rhs = shift;
	defined($rhs) || return undef;

	#if ($rhs =~ /\$/) {
	#	$rhs =~ s/\${([^}]+)}/substitute($1)/ge;
	#}

	while($rhs =~ /\${([^}]+)}/) {
		$rhs =~ s/\${([^}]+)}/substitute($1)/ge;
	}
	return $rhs;
}

sub substitute
{
	my $vname = shift;

	if (exists($VALUES{$vname})) {
		return $VALUES{$vname};
	} else {
		die "no var found";
	}

}

sub quickPermissions
{
	my %PERMS = (
		$VALS{data_dir_abs}		=> 0777,
		$VALS{image_dir_abs}	=> 0777,
		$VALS{log_file}			=> 0666,
	);

	foreach my $file (keys(%PERMS)) {
		my $p = sprintf('0%o', $PERMS{$file});
		verbose("setting [$file] to [$p]");
		$DEBUG || chmod($PERMS{$file}, $file);
	}
}

sub properPermissions
{
	my %PERMS = (
		$VALS{data_dir_abs}		=> 0770,
		$VALS{image_dir_abs}	=> 0770,
		$VALS{js_dir_abs}		=> 0750,
		$VALS{log_file}			=> 0660,
	);

	# first try to figure out if there is a group for the webserver
	my @grps = findWebGroup();
	my $grp;
	my $def;
	if (@grps) {
		print "Please tell me what under what group the webserver runs.\n";
		print "I found the following possibilities: ".join(', ', @grps)."\n";
		$def = $grps[0];
	} else {
		print "I was unable to find any groups that the webserver might use\n";
		print "Please tell me what under what group the webserver runs.\n";
	}

	while(1) {
		$grp = query('Which group shall I use ("none" to give up)', $def);
		if ($grp eq 'none') {
			last;
		}
		if (getgrnam($grp)) { last; }
		print "\ngroup [$grp] does not exist.\n";
	}

	if ($grp eq 'none') {
		print <<_EOF_;
Manual instructions:
You must determine what group your web-server runs under and
than set the group and mode of the following files:

_EOF_
		printf("%-50s  %-11s %s\n", 'File', 'Group', 'Mode');
		foreach my $file (keys(%PERMS)) {
			printf("%-50s  %-11s  0%o\n", $file, 'web-group', $PERMS{$file});
		}
		return;
	}

	my $rs = "$SRCDIR/root-script.sh";
	my $fh = FileHandle->new;
	open($fh, ">$rs") || die "open [$rs] for writing";
	print $fh "#!/bin/sh\n\n";
	if (exists($ENV{DOCUMENT_ROOT})) {
		print $fh "DOCUMENT_ROOT=$ENV{DOCUMENT_ROOT}\n";
	}
	foreach my $file (keys(%PERMS)) {
		my $perms = $PERMS{$file};
		my $p = sprintf('0%o', $perms);
		verbose("(group,mode) for [$file] --> ($grp, $p)");
		print $fh "\nchgrp $grp $file\n";
		printf $fh "chmod 0%o $file\n", $perms;
	}
	$fh->close;
	chmod(0755, $rs) || die "chmod: $!";

	$rs;
}

# only used on U*X platforms
sub findWebGroup
{
	my $group;

	my @out;
	my $fh = FileHandle->new('/etc/group') || die 'open /etc/group to read';
	while(<$fh>) {
		$group = (split(':', $_))[0];
		if ($group =~ /(httpd|apach|web|www)/) {
			push(@out, $group);
		}
	}
	$fh->close;

	@out;
}

sub touch
{
	my ($file, $mode) = @_;

	if (-e $file) {
		verbose("touch: [$file] exists. do nothing.");
		return;
	}

	my $m = sprintf('0%o', $mode);
	verbose("touching [$file] with mode = ($m)");
	$DEBUG && return;

	my $fh = FileHandle->new;
	open($fh, ">$file") || die "touch on $file failed";
	$fh->close;
	chmod($mode, $file) || die "chmod: $!";
}

sub createIndex
{
	my $index_file = "$VALS{data_dir_abs}/"
		.($WINDOZE ? 'Default.htm' : 'index.html');

	if ($DEBUG) {
		$index_file = '/tmp/index.html';
		print "setting index file to /tmp/index.html\n";
	}

	my $fh = FileHandle->new;
	open($fh, ">$index_file") || die "open [$index_file] for writing";
	my $cgirel = $VALS{cgi_dir_rel};
	print $fh <<_EOF_;
<HTML>
<HEAD>
<TITLE>Absence</TITLE>
</HEAD>
<FRAMESET ROWS="60,*">
	<FRAME SRC="$cgirel/absence-control.pl" NAME=control FRAMEBORDER=0>
	<FRAME SRC="$cgirel/absence.pl" name=display>
</FRAMESET>
</HTML>
_EOF_
	$fh->close;
	chmod(0644, $index_file) || die "chmod: $!";
}

sub verbose
{
	my $str = shift;

	$VERBOSE && print "$str\n";
}

sub copyFile
{
	my ($src, $targ) = @_;

	my $cmd;
	if ($WINDOZE) {
		$src =~ s!/+!\\!g;
		$targ =~ s!/+!\\!g;
		$cmd = "copy $src $targ";
	} else {
		$cmd = "cp $src $targ";
	}

	myExec($cmd);
}

sub copyDir
{
	my ($src, $targ) = @_;
	my $cmd;
	$cmd = "cp -r $src $targ";
	myExec($cmd);
}


sub makePath
{
	my ($path, $mode) = @_;

	if (-d $path) { return 1; }

	my $cmd;
	if ($WINDOZE) {
		$path =~ s!/+!\\!g;
		$cmd = "mkdir $path";
	} else {
		$cmd = "mkdir -p $path";
	}

	myExec($cmd);

	if ($mode && !$DEBUG) {
		chmod($mode, $path) || die "chmod $mode $path failed: $!";
	}
}

sub myMkdir
{
	my ($file, $perms) = @_;

	my $m = sprintf('0%o', $perms);
	verbose("CMD: mkdir($file, $m)");
	$DEBUG && return 1;

	return mkdir($file, $perms);
}

sub myExec
{
	my $cmd = shift;

	verbose("execing cmd [$cmd]");
	$DEBUG && return 1;

	$CMDOUT = `$cmd 2>&1`;

	#print "return value: [$?]\n";

	die "cmd [$cmd] failed: $CMDOUT\n" if ($?);
	#return $? ? 0 : 1;
	return 1;
}

sub usage
{
	print "usage: $0 [options]\n options:\n";
	my ($short, $long, $desc, $both);
	printf(" %-15s  %s\n", '-h', 'print help (this)');
	#printf(" %-15s  %s\n", '-f', "(f)orce. be bold.");
	printf(" %-15s  %s\n\n", '-dr <DOCROOT>', "set \$DOCUMENT_ROOT to next arg");
	print " configuration parameters:\n";
	printf("%-5s %-18s %-27s %s\n",'short', 'long', 'description', '[default]');
	printf("%-5s %-18s %-27s %s\n",'-----', '----', '-----------', '--------');

	foreach (getClassKeys('all')) {
		$short = $PARMS{$_}->{short};
		$desc = $PARMS{$_}->{desc};
		$desc =~ s/relative/rel./;
		$desc =~ s/absolute/abs./;
		$short = "$short";
		$long = "$_";
		$both = "$short / $long";
		printf(" %-4s %-18s %-27s [%s]\n", $short, $long, $desc, default($_));
	}
}

sub inList
{
	my ($thing, $lref) = @_;

	foreach (@{$lref}) {
		if ($thing eq $_) {
			return 1;
		}
	}

	0;
}

sub queryParam
{
	my $p = shift;

	my $default = default($p);
	my $desc = $PARMS{$p}->{desc};
	my $allow = $PARMS{$p}->{allow};
	my $none = exists($PARMS{$p}->{none}) ? $PARMS{$p}->{none} : 0;
	my $undef = exists($PARMS{$p}->{undef}) ? $PARMS{$p}->{undef} : 0;
	if (($p eq 'data_dir_abs') || ($p eq 'cgi_dir_abs')) {
	  print "\n[You may want to use \$DOCUMENT_ROOT in the path below]\n";
	}

	if (exists($PARMS{$p}->{extra})) {
		print "$PARMS{$p}->{extra}\n\n";
	}
	return query($desc, $default, $allow, $none, $undef);
}

sub query_param_old
{
	my $p = shift;

	my $default = default($p);
	my $ans;
	my $desc = $PARMS{$p}->[1];
	my $def = ($default) ? " [$default]" : '';
	my $qs = "$desc${def}? ";
	while(!$ans) {
		print "$qs";
		chomp($ans = <STDIN>);
		if (!length($ans) && $default) {
			$ans = $default;
		}
	}

	$ans;
}

sub default
{
	my $p = shift;

	my $default;
	my $path = 0;
	my $def = $PARMS{$p}->{def};
	if ($def) {
		my @list = ref($def) ? @{$def} : $def;
		my @new;
		foreach (@list) {
			if (exists($PARMS{$_})) {
				$path = 1;
				push(@new, "\${$_}");
			} elsif ($_ eq 'name') {
				push(@new, $PARMS{$p}->{name});
			} else {
				push(@new, $_);
			}
		}
		#if ($path) {
			$default = join('/', @new);
		#} else {
		#	$default = $args[0];
		#}
	}

	$default;
}

sub editConfigFile
{
	my ($cf_path, $vref) = @_;

	$DEBUG && return;

	verbose("editing [$cf_path]");

	my $debug = 0;
	my %fixed;

	my $fh = FileHandle->new($cf_path) || die "file not found [$cf_path]\n";
	my @lines = <$fh>;
	$fh->close;

	my @new;
	my $looking = 1;
	foreach (@lines) {
		$debug && print ">> $_";
		chomp;
		if ($looking) {
			if (/^\s*#/) { push(@new, $_); next; }
			if (/^((?:my)?\s+\$?)(\S+)(\s*=>?\s*)(['"]?[^'";,]+['"]?)([;,]\s*(#.*)?)$/) {
				my ($pre, $varname, $mid, $val, $end) = ($1, $2, $3, $4, $5);
				$debug && print "  pre=[$pre] vn=[$varname] mid=[$mid] val=[$val] end=[$end]\n";
				if (exists($vref->{$varname}) && !exists($fixed{$varname})) {
					$debug && print "replacing [$val] with [$vref->{$varname}]\n";
					my $tmp = quote($vref->{$varname});
					push(@new, "$pre$varname$mid$tmp$end");
					$fixed{$varname} = 1;
				} else {
					push(@new, $_);
				}
			} else {
				push(@new, $_);
			}
		} else {
			push(@new, $_);
		}
		if (/%%END_MARKER%%/) { $looking = 0; }
	}

	my $dn = dirname($cf_path);
	my $tmpfile = "$dn/cf-$$";
	$fh = FileHandle->new;
	open($fh, ">$tmpfile") || die "open [$tmpfile] for writing";
	foreach (@new) {
		print $fh "$_\n";
	}
	$fh->close;
	chmod(0644, $tmpfile) || die "chmod: $!";

	if ($DEBUG) {
		print "leaving edited AbsenceConfig as [$tmpfile]\n";
		return;
	}

	rename($tmpfile, $cf_path) || die "rename failed";
}

sub quote
{
	my $val = shift;

	defined($val) || return 'undef';

	if ($val =~ /\$/) {
		return qq["$val"];
	} else {
		return qq['$val'];
	}
}

#-------------------------------------------------------------------
# Database stuff
#-------------------------------------------------------------------
sub createDatabase
{
	my $srcdir = shift;

	my $db_exists = 0;
	my $retval;

	my $sql_file = "$srcdir/db/create_tables.sql";

	-f $sql_file || die "sql-file [$sql_file] not found\n";

	if ($DEBUG) {
		print "debug-mode, will not create database.\n";
		return;
	}

	my %admin_params = (
		a_user		=> $VALUES{database_adm_user},
		a_pass		=> $VALUES{database_adm_pass},
	);

	if (defined($VALUES{database_host}) &&
		($VALUES{database_host} ne 'localhost'))
	{
		$admin_params{host} = $VALUES{database_host};
	}

	my %params = (
		db_user		=> $VALUES{database_user},
		db_pass		=> $VALUES{database_pass},
		create_db	=> 1,
	);

	print <<_EOF_;

	To create the database you need an absence-specific db-user which
	has permission to create a database (CREATEDB).  If you have access
	to the db-superuser (normally "postgres") you can create the
	absence db-user.  Otherwise, the absence db-user needs to exist.

	How shall I proceed?
		(create)	- create absence db-user
		(exists)	- absence db-user already exists
		(abort)		- abort installation

_EOF_

	
	my $ans = query('create/exists/abort', 'abort', [qw/create exists abort/]);

	if ($ans eq 'create') {
	    if (AbsenceInstall::adminCheckUser(
				%admin_params,
				user => $VALUES{database_user}
			)
		)
	    {
		    print "user already exists, re-using.\n";
	    } else {
			verbose("creating absence user [$VALUES{database_user}] ...");
		    AbsenceInstall::adminCreateAbsenceUser(%admin_params, %params);
	    }
	}
	elsif ($ans eq 'exists') {
		my $cu = AbsenceInstall::checkUser(%params);
		if ($cu eq 'not_exist') {
			print "db-user [$params{db_user}] does not exist. abort.\n";
			exit 0;
		}
		elsif ($cu eq 'no_createdb') {
			print "db-user [$params{db_user}] does not have CREATEDB right. abort.\n";
			exit 0;
		}
		else {
			print "proceeding with installation.\n";
		}
	}
	else {
		print "user aborted installation.\n";
		exit 0;
	}

	if (AbsenceInstall::checkDatabase(%params, db => $VALUES{database_name}))
	{
		print "\ndatabase [$VALUES{database_name}] exists.\n"
			. "you have three choices:\n"
			. "\t(drop)	- drop database and create new\n"
			. "\t(cont)	- continue with current database\n"
			. "\t(abort)	- abort installation\n\n";

		my $ans = query('What would you like to do?', 'abort', [qw/drop cont abort/]);
		if ($ans eq 'drop') {
			AbsenceInstall::dropDatabase(%admin_params,
				db => $VALUES{database_name});
			$retval = 'created';
		}
		elsif ($ans eq 'cont') {
			print "continuing installation with current database...\n";
			$db_exists = 1;
			$retval = 'exists';
		}
		elsif ($ans eq 'abort') {
			print "aborting installation!\n";
			exit 0;
		}
	}

	if (!$db_exists) {
		verbose("creating database [$VALUES{database_name}] ...");

		AbsenceInstall::createAbsenceDb(
			user		=> $VALUES{database_user},
			pass		=> $VALUES{database_pass},
			host		=> $VALUES{database_host},
			name		=> $VALUES{database_name},
		);

		verbose("populating database ...");

		AbsenceInstall::populateDatabase(
			user		=> $VALUES{database_user},
			pass		=> $VALUES{database_pass},
			name		=> $VALUES{database_name},
			host		=> $VALUES{database_host},
			file		=> $sql_file,
		);
	}

	return $retval;
}

sub importCountries
{
	my $srcdir = shift;

	AbsenceImport::init(
		$VALUES{database_user},
		$VALUES{database_pass}, 
		$VALUES{database_name}, 
		$VALUES{database_host}, 
	);

	verbose("loading country names and codes into database");

	my $input_file = "$srcdir/misc/iso3166-countrycodes.txt";
	my $fh = FileHandle->new($input_file);
	defined($fh)
		|| die "failed to open iso-country-code file [$input_file]: $!";
	my $parsing = 0;
	my @list;
	while(<$fh>) {
		chomp;
		next if (/^\s*$/);
		if ($parsing) {
			if (/^(\S+(?:\s+\S+)*?)\s+(\S{2})\s+(\S{3})\s+(\d{3})\s*$/) {
				my ($name, $two, $three, $num) = ($1, $2, $3, $4);
				push(@list, { code => $two, name => $name });
			}
		}
		elsif (/^-{50,}$/) {
			$parsing = 1;
		}
	}
	$fh->close;

	AbsenceImport::loadCountryTable(\@list);
}
