package AbsenceInstall;

use FileHandle;
use IPC::Open3;
use IO::Select;
use POSIX ":sys_wait_h";
use English;
use DBI;
use Exporter;

use AbsenceUtils;

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw();

#use AbsenceImport;

my $VERBOSE = 1;

#-------------------------------------------------------------------------
# popuateDatabase()
# 	reads an sql-file and runs commands
# params:
#	user	=> <admin-user>
#	pass	=> <admin-pass>
#	name	=> <database-name>
#	file	=> <sql-input-file>
#-------------------------------------------------------------------------
sub populateDatabase
{
	my %args = @_;

	my $dbh = connectDb($args{user}, $args{pass}, $args{name}, $args{host});

	my @statements = readSqlFile($args{file});

	foreach my $sql (@statements) {
		if (!executeSql($dbh, $sql)) {
			die "sql statment failed:\n$sql\n--end--.\n";
		}
	}
	$dbh->disconnect();
	return 1;
}

sub testConnection
{
	my %args = @_;

	my $dbh = connectDb($args{user}, $args{pass}, 'template1', $args{host});

	if (defined($dbh)) {
		$dbh->disconnect;
		return 1;
	}

	return 0;
}

#-------------------------------------------------------------------------
# adminCreateAbsenceUser
# creates the db-user and the db
# params:
#	user	=> <admin-user>
#	pass	=> <admin-pass>
#	host	=> <db-host>
#	db_user	=> <owner of absence-db>
#	db_pass	=> <owner password>
#	create_db	=> [0|1]	# give the user CREATEDB privilege
#-------------------------------------------------------------------------
sub adminCreateAbsenceUser
{
	my %args = @_;

	my $dbh = connectDb($args{a_user}, $args{a_pass}, 'template1', $args{host});
	if (!defined($dbh)) {
		die "unable to connect to host [$args{host}], user='$args{user}, pass='$args{pass}\n";
	}

	my $create_db = $args{create_db} ? 'CREATEDB' : 'NOCREATEDB';

	#-------------------------------------
	# create db owner user
	#-------------------------------------
	my $sql = qq{
		CREATE ROLE $args{db_user}
			NOSUPERUSER $create_db NOCREATEROLE INHERIT LOGIN;
	};

	$VERBOSE && print "creating db-user '$args{db_user}'\n";
	if (!executeSql($dbh, $sql)) {
		die "attempt to create absence user failed.\n";
	}

	#-------------------------------------
	# set password
	#-------------------------------------
	$sql = qq{
		ALTER USER $args{db_user} ENCRYPTED PASSWORD '$args{db_pass}';
	};
	$VERBOSE && print "setting password for db-user to '$args{db_pass}'\n";
	#print "SQL=[$sql]\n";
	if (!executeSql($dbh, $sql)) {
		die "attempt to set password for absence user failed.\n";
	}

	#-------------------------------------
	# disconnect as admin
	#-------------------------------------
	$dbh->disconnect();
}

sub adminCheckUser
{
	my %args = @_;

	my $dbh = connectDb($args{a_user}, $args{a_pass}, 'template1', $args{host});

	if (!defined($dbh)) {
		die "unable to connect to host [$args{host}], user='$args{a_user}, pass='$args{a_pass}\n";
	}

	my $sql = qq{
		SELECT * FROM pg_user WHERE usename = '$args{user}';
	};

	my $sth = $dbh->prepare($sql);
	defined($sth) || die "adminCheckUser: prepare for SQL [$sql] failed";

	$sth->execute || die "execute of SQL [$sql] failed";
	my $res = $sth->fetchrow_hashref;
	$sth->finish;
	$dbh->disconnect();

	return defined($res) ? 1 : 0;
}

sub checkUser
{
	my %args = @_;
	my $dbh = connectDb($args{db_user}, $args{db_pass}, 'template1', $args{host});
	if (!defined($dbh)) {
		return 'not_exist';
	}

	my $sql = "SELECT * FROM pg_roles WHERE rolname = '$args{db_user}';";

	my $sth = $dbh->prepare($sql);
	defined($sth) || die "checkUser: prepare for SQL [$sql] failed";

	$sth->execute || die "checkUser: execute of SQL [$sql] failed";
	my $res = $sth->fetchrow_hashref;
	$sth->finish;
	$dbh->disconnect();

	if (!$res->{rolcreatedb}) { return 'no_createdb'; }

	return 'ok';
}

sub checkDatabase
{
	my %args = @_;

	#my $dbh = connectDb($args{a_user}, $args{a_pass}, 'template1', $args{host});
	my $dbh = connectDb($args{db_user}, $args{db_pass}, 'template1', $args{host});

	if (!defined($dbh)) {
		die "unable to connect to host [$args{host}], user='$args{a_user}, pass='$args{a_pass}\n";
	}

	my $sql = qq{
		SELECT * FROM pg_database WHERE datname = '$args{db}';
	};

	my $sth = $dbh->prepare($sql);
	defined($sth) || die "checkDatabase: prepare for SQL [$sql] failed";

	$sth->execute || die "execute of SQL [$sql] failed";
	my $res = $sth->fetchrow_hashref;
	$sth->finish;
	$dbh->disconnect();

	return defined($res) ? 1 : 0;
}

sub dropDatabase
{
	my %args = @_;

	my $dbh = connectDb($args{a_user}, $args{a_pass}, 'template1', $args{host});

	if (!defined($dbh)) {
		die "unable to connect to host [$args{host}], user='$args{a_user}, pass='$args{a_pass}\n";
	}

	my $sql = qq{
		DROP DATABASE $args{db};
	};

	my $rv = $dbh->do($sql);
	$dbh->disconnect();
	defined($rv) || die "drop database failed";
}

sub dropUser
{
	my %args = @_;

	my $dbh = connectDb($args{a_user}, $args{a_pass}, 'template1', $args{host});

	if (!defined($dbh)) {
		die "unable to connect to host [$args{host}], user='$args{a_user}, pass='$args{a_pass}\n";
	}

	my $sql = qq{
		DROP ROLE $args{user};
	};

	my $rv = $dbh->do($sql);
	$dbh->disconnect();
	defined($rv) || die "drop role failed";
}

#-------------------------------------------------------------------------
# createAbsenceDb()
# params:
#	user		=> <user>	# connect with this user
#	pass		=> <pass>	# password for above user
#	owner		=> <owner>	# db-owner
#	name		=> <name>	# db-name
#	tablespace	=> <ts>		# default-tablespace
#-------------------------------------------------------------------------
sub createAbsenceDb
{
	my %args = @_;

	my $dbh = connectDb($args{user}, $args{pass}, 'template1', $args{host});
	if (!defined($dbh)) {
		die "unable to connect to host [$args{host}], user='$args{user}, pass='$args{pass}\n";
	}

	my $ts = exists($args{tablespace})
		? "TABLESPACE '$args{tablespace}'"
		: '';
	
	my $owner = '';
	if (exists($args{owner})) {
		$owner = "OWNER $args{owner}";
	}

	my $sql = qq{
		CREATE DATABASE $args{name} WITH ENCODING 'UTF8' ${ts} $owner
	};
	$VERBOSE && print "creating DB '$args{name}'\n";
	if (!executeSql($dbh, $sql)) {
		die "create DB '$args{name} failed.\n";
	}

	#-------------------------------------
	# disconnect 
	#-------------------------------------
	$dbh->disconnect();
}

sub readSqlFile
{
	my $file = shift;

	my $comment = 0;
	my $debug = 0;

	my $fh = FileHandle->new($file);
	defined($fh) || die "unable to open [$file] for reading: $!";

	my @sql;
	my $sql = '';
	my $quoted = 0;

	while(<$fh>) {
		$debug && print "LINE: $_";
		if ($comment) {
			if (m!\*/(.*)$!) {
				$debug && print "  comment = 0\n";
				$comment = 0;
				if (length($1)) {
					$_ = $1;
					redo;
				}
			}
		} else {
			if (/^\s*$/) { next; }
			if (m!^\s*/\*(.*)$!) {
				$debug && print "  comment = 1\n";
				$comment = 1;
				my $rest = $1;
				if ($rest =~ m!\*/(.*)$!) {
					$comment = 0;
					if (length($1)) {
						$_ = $1;
						redo;
					}
				}
				next;
			}
			$sql .= $_;
			if ($quoted) {
				if (m!^'!) {
					$quoted = 0;
					$debug && print "  quoted = 0\n";
				}
			} else {
				if (m!'$!) {
					$quoted = 1;
					$debug && print "  quoted = 1\n";
					next;
				}
			}
			if (!$quoted) {
				if ($sql =~ /;\s*$/) {
					push(@sql, $sql);
					$sql = '';
					$debug && print "  -end of sql-\n";
				}
			}
		}
	}
	$fh->close;

	return @sql;
}

1;
