#!/usr/bin/perl

use Getopt::Long;
use Data::Dumper;
use File::Basename;
use Encode;

#Getopt::Long::Configure('debug');
my $DEBUG = 1;
my ($import_file, $region, $country);

$Data::Dumper::Indent = 1;

my ($db_user, $db_pass, $db_host, $db_name);

$db_host = 'localhost';

#my $PATH_TO_CGI_BIN_DIR = '';

my $mode;
my $delete_list;
my $list_thing;
my ($new_region, $name, $desc);
my $import_format;
my $convert_file;
my $date;

my $res = GetOptions(
	'convert|conv=s'	=> sub { $mode = 'convert'; $convert_file = $_[1]; },
	'import|i=s'		=> sub { $mode = 'import'; $import_format = $_[1]; },
	'list|l=s'			=> sub { $mode = 'list'; $list_thing = $_[1]; },
	'add-region|ar'		=> sub { $mode = 'add-region'; },
	'add-holiday|ah'	=> sub { $mode = 'add-holiday'; },
	'delete-holiday|dh=s'	=> sub { $mode = 'delete-h'; $delete_list = $_[1]; },
	'delete-region|dr=s'	=> sub { $mode = 'delete-r'; $delete_list = $_[1]; },
	'file|f=s'			=> \$import_file,
	'country|c=s'		=> \$country,
	'region|r=s'		=> \$region,
	'user|u=s' 			=> \$db_user,
	'pass|p=s' 			=> \$db_pass,
	'host=s' 			=> \$db_host,
	'db-name|dbn=s' 	=> \$db_name,
	'help|h'			=> \&usage,
	'name|n=s'			=> \$name,
	'desc|d=s'			=> \$desc,
	'date|dt=s'			=> \$date,
);

if (!defined($mode)) {
	die "a mode must be specified, one of -convert/-import/-list/-delete*/-add-region\n";
}

#defined($import_file) || die "usage: $0 -file <input-file>\n";
($mode ne 'convert')
	&& (!defined($db_user) || !defined($db_pass) || !defined($db_name))
	&& die "DB username/password/name missing\n";

#----------------------------------------------------------------
# set up INC and require a few modules
#----------------------------------------------------------------
my $dn = dirname($0);
push(@INC, $dn);
foreach my $mod (qw(AbsenceUtils AbsenceMigration AbsenceImport)) {
	eval {
		require "$mod.pm";
	};
	if ($@) {
		die "failed to load module [$mod.pm]: $@\n";
	}
	
}

#----------------------------------------------------------------
# get AbsenceConfig (and AbsenceDB)
#----------------------------------------------------------------
#AbsenceUtils::loadAbsenceDB($PATH_TO_CGI_BIN_DIR);

($mode ne 'convert')
	&& AbsenceImport::init($db_user, $db_pass, $db_name, $db_host);

if ($mode eq 'list') {
	list($list_thing);
}
elsif ($mode eq 'delete-h') {
	delete_hol($delete_list);
}
elsif ($mode eq 'delete-r') {
	delete_reg($delete_list);
}
elsif ($mode eq 'add-region') {
	add_region($name, $desc);
}
elsif ($mode eq 'add-holiday') {
	add_holiday($date, $desc, $country, $region);
}
elsif ($mode eq 'convert') {
	convert_ical($convert_file);
}
else {
	import_file($import_format, $import_file, $country, $region);
}
exit;

#=======================================================================
# subroutines
#=======================================================================

sub list
{
	my $thing = shift;

	if ($thing =~ /^h.*/i) {
		list_holidays();
	}
	elsif ($thing =~ /^c.*/i) {
		list_countries();
	}
	elsif ($thing =~ /^r.*/i) {
		list_regions();
	}
	else {
		die "don't know how to list [$thing]\n";
	}
}

sub list_holidays
{
	binmode(*STDOUT, ":utf8");
	my @hols = AbsenceImport::getDbHolidays();
	print "\n";
	printf("%3s %-25s %-12s %-18s %-17s\n",
		'ID', 'NAME', 'DAY', 'COUNTRY', 'REGION');

	my (@fixed, @variable);

	foreach my $hol (@hols) {
		$hol->{description} = decode_utf8($hol->{description});
		if ($hol->{day} =~ /^0001-/) {
			push(@fixed, $hol);
		}
		else {
			push(@variable, $hol);
		}
	}

	print "# fixed holidays\n";
	foreach my $hol (@fixed) {
		my $country = defined($hol->{country_id})
			? AbsenceImport::getCountry($hol->{country_id})->{name}
			: '';
		my $region = defined($hol->{region_id})
			? AbsenceImport::getRegion($hol->{region_id})->{name}
			: '';
		printf("%3d %-25s %-12s %-18s %-17s\n",
			$hol->{id}, $hol->{description}, $hol->{day}, $country, $region);
	}

	print "\n# variable holidays\n";
	my $prev_year;
	foreach my $hol (sort { $a->{day} <=> $b->{day} } @variable) {
		my ($y,$m,$d) = ($hol->{day} =~ /^(\d{4})-(\d{2})-(\d{2})$/);
		if ($prev_year != $y) {
			print "# $y\n";
		}
		my $country = defined($hol->{country_id})
			? AbsenceImport::getCountry($hol->{country_id})->{name}
			: '';
		my $region = defined($hol->{region_id})
			? AbsenceImport::getRegion($hol->{region_id})->{name}
			: '';
		printf("%3d %-25s %-12s %-18s %-17s\n",
			$hol->{id}, $hol->{description}, $hol->{day}, $country, $region);
		$prev_year = $y;
	}
}

sub list_countries
{
	my @list = AbsenceImport::getCountries();
	print "\n";
	printf("%3s %3s %s\n", 'ID', 'CC', 'NAME');
	foreach my $c (@list) {
		printf("%3d %3s %s\n", $c->{id}, $c->{code}, $c->{name});
	}
	print "done.\n";
}

sub list_regions
{
	my @list = AbsenceImport::getRegions();
	print "\n";
	if (!@list) {
		print "no regions defined\n";
		return;
	}
	printf("%3s %-30s %-40.40s\n", 'ID', 'NAME', 'DESCRIPTION');
	foreach my $r (@list) {
		printf("%3s %-30s %-40.40s\n", $r->{id}, $r->{name}, $r->{description});
	}
	print "done.\n";
}

sub add_region
{
	my ($name, $desc) = @_;

	if (!defined($name) || !length($name)) {
		die qq{you must specify the region-name using "-rn"\n};
	}

	my $name_utf8 = AbsenceUtils::convertString($name);
	my $desc_utf8 = AbsenceUtils::convertString($desc);

	print "creating region [$name_utf8] with description [$desc_utf8]\n";

	my $id = AbsenceImport::createRegion($name_utf8, $desc_utf8);
	print "region created with id=$id\n";
}

sub add_holiday
{
	my ($date, $desc, $country, $region) = @_;
	if ($date !~ /^(\d{4}|\*)-(\d{1,2})-(\d{1,2})$/) {
		die "add-holiday: date not specified or not in format YYYY-MM-DD\n";
	}
	my ($year, $mon, $day) = ($1, $2, $3);
	if (!defined($desc) || length($desc) == 0) {
		die "add-holiday: no description provided (use '-desc <desc>')\n";
	}
	my $desc_utf8 = AbsenceUtils::convertString($desc);

	my $additional = handle_country_region($country, $region);

	my $y	= ($year eq '*') ? '0001' : $year;
	my $new_date = "$y-$mon-$day";
	my $ref = {
		day			=> $new_date,
		description	=> $desc_utf8,
		@{ $additional },
	};
	my $id = AbsenceImport::loadHoliday($ref);
	print "added holiday [$desc_utf8] with id=$id\n";
}

sub delete_hol
{
	my $delete_list = shift;

	my @list = split(',', $delete_list);

	if ($list[0] eq 'all') {
		print "deleting all holidays\n";
		AbsenceImport::deleteAllHolidays();
		print "done.\n";
		return;
	}

	foreach my $n (@list) {
		if ($n !~ /^\d+$/) {
			die "ID in delete-list is not number: [$n]\n";
		}
	}
	foreach my $n (@list) {
		print "deleting holiday with id = $n\n";
		AbsenceImport::deleteHoliday($n);
	}
	print "done.\n";
}

sub delete_reg
{
	my $delete_list = shift;

	my @list = split(',', $delete_list);
	foreach my $n (@list) {
		if ($n !~ /^\d+$/) {
			die "ID in delete-list is not number: [$n]\n";
		}
	}
	foreach my $n (@list) {
		print "deleting region with id = $n\n";
		AbsenceImport::deleteRegion($n);
	}
	print "done.\n";
}

sub import_file
{
	my ($format, $file, $country, $region) = @_;

	if (defined($country) && defined($region)) {
		die "cannot specify both -country and -region.\n";
	}

	my $additional = handle_country_region($country, $region);

	if ($format eq 'ical') {
		import_ical($file, $additional);
	} elsif ($format eq 'legacy') {
		import_legacy($file, $additional);
	} else {
		die "format [$format] unknown. must be 'ical' or 'legacy'\n";
	}
}

sub handle_country_region
{
	my ($country, $region) = @_;

	my $additional;

	if (defined($country)) {
		if ($country =~ /^\d+$/) {
			$additional = [country_id => $country];
		} elsif ($country =~ /^[a-z]{2}$/i) {
			my $id = AbsenceImport::lookupCountry($country);
			defined($id) || die "country-code [$country] unknown\n";
			$additional = [country_id => $id];
		} else {
			die "country must be an id (numeric) or a 2 letter country-code\n";
		}
	} elsif (defined($region)) {
		if ($region =~ /^\d+$/) {
			$additional = [region_id => $region];
		} else {
			my $id = AbsenceImport::lookupRegion($region);
			defined($id) || die "region [$region] unknown\n";
			$additional = [region_id => $id];
		}
	}

	return $additional;
}

sub import_legacy
{
	my ($file, $additional_ref) = @_;

	my $hdata = AbsenceMigration::readHolidays($file);

	import($hdata, $additional_ref);
}

sub import_ical
{
	my ($file, $additional_ref) = @_;

	my $hdata = AbsenceImport::parseIcal($file);

	import($hdata, $additional_ref);
}

sub import
{
	my ($hdata, $additional_ref) = @_;

	my $debug = 0;

	foreach my $year (keys(%{ $hdata })) {
		$debug && print "year=[$year]\n";
		foreach my $month (keys(%{ $hdata->{$year} })) {
			$debug && print "  month=[$month]\n";
			foreach my $day (keys(%{ $hdata->{$year}->{$month} })) {
				$debug && print "    day=[$day]\n";
				my $y		= ($year eq '*') ? '0001' : $year;
				my $d		= "$y-$month-$day";
				my $id = AbsenceImport::lookupHoliday($day, $month, $y, $additional_ref);
				if (defined($id)) {
					print "holiday for [$d] already exists with id=$id. skip.\n";
					next;
				}
				my $desc	= $hdata->{$year}->{$month}->{$day};
				my $ref = {
					day			=> $d,
					description => encode_utf8($desc),
					@{ $additional_ref },
				};
				$id = AbsenceImport::loadHoliday($ref);
				print "added holiday [$desc] with id=$id\n";
			}
		}
	}
}

sub convert_ical
{
	my $file = shift;

	defined($file) || die "usage: $0 <file>\n";

	binmode(*STDOUT, ":utf8");

	my $data = AbsenceImport::parseIcal($file);

	if (exists($data->{'*'})) {
		my $fref = $data->{'*'};
		print "# fixed holidays\n";
		foreach my $month (sort { $a <=> $b } keys(%{ $fref })) {
			foreach my $day (sort { $a <=> $b } keys(%{ $fref->{$month} })) {
				printf("%d.%d.%s\t%s\n",
					$day, $month, '*', $fref->{$month}->{$day});
					#$day, $month, '*', encode_utf8($fref->{$month}->{$day}));
			}
		}
	}

	print "\n# varying holidays\n";
	foreach my $year (sort {$a <=> $b} keys(%{ $data })) {
		next if ($year eq '*');
		print "# $year\n";
		foreach my $month (sort {$a <=> $b} keys(%{ $data->{$year} })) {
			foreach my $day (sort {$a <=> $b} keys(%{ $data->{$year}->{$month} })) {
				printf("%d.%d.%s\t%s\n",
					$day, $month, $year, $data->{$year}->{$month}->{$day});
					#$day, $month, $year, encode_utf8($data->{$year}->{$month}->{$day}));
			}
		}
	}
}

sub usage
{
	print <<_EOF_;
usage: $0 MODE DB-OPTIONS [ IMPORT-OPTIONS | ADD-OPTIONS ]

MODE is one of:				(abbreviation)
  -list    <holidays|countries|regions>	(-l)
  -import		<type>		(-i) <type> can be "ical" or "legacy"
  -convert		<file>		(-conv)
  -add-region				(-ar)
  -add-holiday				(-ah)
  -delete-holiday	<list|all>	(-dh)
  -delete-region	<list>		(-dr)

	<list> consists of a list of IDs as returned by -list,
	comma-separated, with no white-space. exampe: 2,66,81,93
	for holidays list can be 'all', which deletes all holidays and
	resets the ID sequence to 1.

  modes:
  -import		imports a list of holidays from a file
			into database
  -convert		converts a list of holidays from an ics file to
			legacy format on standard output
  -add-region		adds a region to database
  -add-holiday		adds a holiday to database
  -list			lists holidays in database
  -delete-holiday	deletes a list of holidays from database
  -delete-region	deletes a list of regions from database
  

DB-OPTIONS:
	-db-name DBNAME			(-dbn)
	-user    DBUSER			(-u)
	-pass    DBPASS			(-p)
	-host    DBHOST				(default: "localhost")

ADD-REGION-OPTIONS:
	-name <name>	 		(-n) (mandatory)
	-desc <description>		(-d) (optional)

ADD-HOLIDAY-OPTIONS:
	-date <day>	 		(-dt) (mandatory, format: YYYY-MM-DD)
	-desc <description>		(-d)  (mandatory)
	-country <country-id>		(-c)  (optional)
	-region <region-id>		(-r)  (optional)

IMPORT-OPTIONS:
	-file <filename>		(-f) file to import from
	-country <country>		(-c) assign country to imported holidays
	-region <region>		(-r) assign region to imported holidays

NOTE:
-country and -region are mutally exclusive.
To add a fixed holiday, use '*' instead of the four-digit year. example:
	-add-holiday -date '*-1-1' -desc "New Years Day"
_EOF_
exit;
}
